#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Utility classes of common use to all rendering backends. May be used
by non-backend packages as well.
"""
from rayon import RayonException

class LayoutUnits:
    def __init__(self, val, orig_spec=""):
        self.val = val
        self.orig_spec = orig_spec

# Note: when outputting to PDF, "pixels" are points. 
class Pixels(LayoutUnits):
    def to_pixels(self, ctx, total):
        return self.val
    def __repr__(self):
        return("Pixels(%d)" % self.val)

class Percent(LayoutUnits):
    def to_pixels(self, ctx, total):
        return (self.val / 100) * total
    def __repr__(self):
        return("Percent(%d)" % self.val)

class CallableUnits(LayoutUnits):
    def to_pixels(self, ctx, total):
        return self.val(ctx, total)


unit_map = {
    "px" : Pixels,
    "pt" : Pixels,
}
def units_from_spec(spec):
    if isinstance(spec, LayoutUnits):
        return spec
    if callable(spec):
        return CallableUnits(spec)
    try:
        floatspec = float(spec)
    except ValueError:
        # Pass, try to treat it as string
        pass
    else:
        # It's a bare number. Treat as a proportion
        if floatspec < 0.0 or floatspec > 1.0:
            raise ValueError("Invalid unit specification: %s" % str(spec))
        return Percent(floatspec * 100)
    # It's (hopefully) a valid string spec
    try:
        return unit_map[spec[-2:]](float(spec[:-2]), spec)
    except:
        # Oops
        raise ValueError("Invalid unit specification: %s" % str(spec))
    

class Context(object):
    """
    Rayon's 2-d drawing interface. Rayon rendering backends implement
    subclasses of :class:`Context` to give Rayon plot developers a
    consistent drawing interface.

    It should never be necessary to create this class directly; it
    will be created by Rayon when it renders a chart.

    If a renderer does not provide a given method on this object, it
    will raise a :exc:`NotImplementedError`.

    All color values in the :class:`Context` object should be 4-tuples
    of integers ``(R, G, B, A)``, giving the values of the red, green,
    blue and alpha channels, respectively. Each integer should be
    greater than or equal to 0 and less than or equal to 255. (For
    alpha, 0 is totally transparent and 255 is totally opaque.)
    """
    def __init__(self, *args, **kargs):
        if type(self) == Context:
            raise RayonException(
                "The backend.Context object should never be created")

    def canvas_dimensions(self):
        """
        Returns the dimensions of the drawable area in device units
        (points or pixels).
        """
        raise NotImplementedError("canvas_dimensions")

    def draw_bitmap(self, bitmap, x, y, width, height, fill):
        """
        Draws a :class:`~rayon.backgrounds.Bitmap` object onto the
        drawing surface. *x* and *y* are the Cartesian coordinates on
        the canvas where the top left corner of the box on the canvas
        where the bitmap should be drawn. *width* and *height* are the
        width and height of the box on the canvas where *bitmap* will
        be drawn.

        *fill* determines how the image will be drawn if *width* and
        *height* do not correspond to the width and height of
        *bitmap*. *fill* must be one of the following strings::

        ``scale``
            Scale *bitmap* to fit *width* and *height*.
            
        ``none``
            Draw *bitmap* as-is, ignoring *width* and *height*.
        """
        # Future values for fill: 'tile', 'center'
        pass

    def draw_image(self, filename, x, y, width, height, fill):
        """
        Draws a PNG image onto the drawing surface. *filename* is the
        name of the file containing the image to draw. *x* and *y* are
        the Cartesian coordinates on the canvas where the top left
        corner of the box on the canvas where the image should be
        drawn. *width* and *height* are the width and height of the
        box on the canvas where the image will be drawn.

        *fill* determines how the image will be drawn if *width* and
        *height* do not correspond to the width and height of the
        image. *fill* must be one of the following strings:

        ``scale``
            Scale the image to fit *width* and *height*.
            
        ``none``
            Draw the image as-is, ignoring *width* and *height*.
        """
        pass

    def draw_ellipse(self, x, y, width, height, stroke_color,
                     fill_color, stroke_width=1):
        """
        Draws an ellipse. *x* and *y* are the Cartesian coordinates on
        the canvas of the center of the ellipse. *width* and *height*
        are the ellipse's width at it's widest point and the height of
        the ellipse at it's highest point. (Put another way, they are
        the lengths, respectively, of the horizontal and vertical line
        segments passing through the center of the ellipse and bounded
        by its edges.)

        *stroke_color* is the color of the edge of the
        ellipse. *fill_color* is the color of the interior of the
        ellipse. *stroke_width* is the width of the edge of the
        ellipse, in device units (points or pixels).
        """
        pass

    def draw_line(self, x1, y1, x2, y2, color, width, fixup=False,
                  dashes=None, dash_offset=0):
        """
        Draws a single line segment on the canvas, from ``(x1, y1)``
        to ``(x2, y2)``. *color* is the color of the line; *width* is
        it's width in device units (points or pixels).

        If *fixup* is ``True`` and the line segment is perfectly
        vertical or horizontal, the rendering backend may attempt to
        shift the position of the segment by not more than .5
        pixels/points to fall cleanly along pixel boundaries. This
        adjustment will only be made when rendering to a raster format
        such as PNG.

        If specified, *dashes* is an array of integers describing a
        repeating pattern of dashes. Each integer specifies a length
        in device units (pixels or points) of a portion of the
        line. The line will be drawn for the length in the first
        element, not drawn for the length in the second element,
        etc.. When the last element is reached, the array will be
        revisited starting with the first element.  If *dash_offset*
        is not zero, the pattern of dashes in *dashes* will begin at
        the specified array index instead of at the beginning of the
        array.
        """
        pass

    def draw_lines(self, segments, color, width, fixup=False,
                   dashes=None, dash_offset=0):
        """
        Draws a series of line segments on the canvas. *segments* is a
        4-tuple ``(x1, y1, x2, y2)``, representing the start and end
        points of a given line segment in Cartesian
        coordinates. *color* is the color of the lines; *width* is
        their width in device units (points or pixels).

        If *fixup* is ``True`` and a line segment is perfectly
        vertical or horizontal, the rendering backend may attempt to
        shift the position of the segment by not more than .5
        pixels/points to fall cleanly along pixel boundaries. This
        adjustment will only be made when rendering to a raster format
        such as PNG.

        If specified, *dashes* is an array of integers describing a
        repeating pattern of dashes. Each integer specifies a length
        in device units (pixels or points) of a portion of the
        line. The line will be drawn for the length in the first
        element, not drawn for the length in the second element,
        etc.. When the last element is reached, the array will be
        revisited starting with the first element.  If *dash_offset*
        is not zero, the pattern of dashes in *dashes* will begin at
        the specified array index instead of at the beginning of the
        array. The pattern in *dashes* will be applied to every
        segment in *segments*, always starting at *dash_offset*.
        """
        pass

    def draw_polygon(self, vertices, stroke_width, stroke_color, fill_color,
                     fixup=False):
        """
        Draws a polygon. *vertices* is a sequence of ``(x, y)`` pairs
        representing the vertices of the polygon in Cartesian
        coordinates. If *vertices* does not describe a closed polygon,
        a final line segment will be drawn between the last point in
        *vertices* and the first.

        *stroke_width* is the width of the edges of the polygon in
        device units (points or pixels); *stroke_color* is the color
        of the edge. *fill_color* is the color of the interior of the
        polygon.
        """
        pass

    def draw_rectangle(self, x, y, width, height,
                       stroke_width, stroke_color, fill_color,
                       fixup=False):
        """
        Draws a rectangle or square. *x* and *y* describe the top left
        corner of the rectangle in Cartesian coordinates. *width* and
        *height* are the width and height of the rectangle in device
        units (points or pixels).
        
        *stroke_width* is the width of the edges of the rectangle in
        device units (points or pixels); *stroke_color* is the color
        of the edge. *fill_color* is the color of the interior of the
        polygon.
        """
        pass

    def draw_text(self, text, bbox, size, family, style, weight,
                  color=(0, 0, 0), angle=None):
        """
        Renders the string in *text* to the canvas. *bbox* is a
        4-tuple, each of whose elements are ``(x, y)`` pairs
        describing the Cartesian coordinates of the label bounding
        box's corners (top left, top right, bottom right and bottom
        left).

        *size* is the size of the font in points.

        *family* can be one of the strings ``sans-serif`` and
        ``serif``. *style* can be one of the strings ``normal``,
        ``slant`` and ``italic``. *weight* can be one of the strings
        ``normal``, ``light`` or ``bold``. The rendering backend may
        use these to select an appropriate font.

        *color* is the color that will be used to draw the
        text.

        By default, text will be rendered horizontally, from left to
        right. If specified, *angle* is an angle of rotation
        counterclockwise from this base. For example, an angle of
        ``90`` indicates the text should be rendered vertically, with
        the text running bottom to top. The text will rotate around
        the top left corner of the bounding box.
        """
        pass

    def get_text_dimensions(self, text, size, family, style, weight):
        """
        Returns a 2-tuple ``(width, height)`` representing the width
        and height of the rectangle required to render the string
        *text* using :meth:`~Context.draw_text` with the specified
        *size*, *family*, *style* and *weight* arguments.
        """
        pass

    def push_state(self):
        """
        Pushes the state of the rendering backend's transformation
        matrix onto a stack for later retrieval. It's usually best to
        call this method before changing the origin of the coordinate
        system with :meth:`~Context.translate` so the previous origin
        can be restored using :meth:`~Context.pop_state`.
        """
        pass

    def pop_state(self):
        """
        Restores the state of the rendering backend's transformation
        matrix from an internal stack. This method is usually called
        after calls to :meth:`~Context.translate` in order to restore
        the previous origin, which was saved with a call to
        :meth:`~Context.pop_state`.
        """
        pass

    def translate(self, dx, dy):
        """
        Moves the origin of the coordinate system *dx* device units
        (pixels or points) to the right and *dy* device units down. If
        *dx* or *dy* are negative, the origin is moved left and/or up,
        respectively.

        Calls to this method should generally be preceded by a call to
        :meth:`~Context.push_state` and followed by a call to
        :meth:`~Context.pop_state`, so that the state of the
        coordinate system doesn't change unpredictably from the
        standpoint of other users of this :class:`Context` object.
        """
        pass


class Page(object):
    def write(self, chart):
        """
        Output *chart* to the canvas.
        """
        pass
