#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Objects representing different types of backgrounds (colors, images,
gradients, etc.)
"""

from __future__ import division

import base
import miscutils

import array
import sys

class Background(base.Panel):
    """
    Base class for objects added to :class:`~rayon.charts.Chart`
    objects using :meth:`~rayon.charts.Chart.add_chart_background`,
    :meth:`~rayon.charts.Chart.add_chart_foreground`,
    :meth:`~rayon.charts.Chart.add_plot_background`, and
    :meth:`~rayon.charts.Chart.add_plot_foreground`.
    """
    pass


def normalize_color(clr):
    if isinstance(clr, str):
        clr = miscutils.parse_color_string(clr)

    if len(clr) == 3:
        clr = list(clr)
        clr.append(255)
    return clr

class ColorBackground(Background):
    """
    A :class:`Background` object representing a background of a
    *fill_color*, with a thin border of *stroke_color*. *fill_color*
    and *stroke_color* are both color specifications.
    """
    
    def __init__(self, fill_color=(255, 255, 255, 255),
                 stroke_color=(255,255,255,255)):
        Background.__init__(self)
        self.fill_color = normalize_color(fill_color)
        self.stroke_color = normalize_color(stroke_color)
    def draw(self, ctx, width, height):
        Background.draw(self, ctx, width, height)
        ctx.draw_rectangle(0, height, width, height,
                           stroke_width=1,
                           stroke_color=self.stroke_color,
                           fill_color=self.fill_color)

def make_funky(r, g, b, a, format, premul, for_png, debug=False):
    # Convert data in unpremultiplied RGBA to given format
    global msg_counter
    def msg(m):
        if debug:
            print m
    if premul:
        r = int(r * a/255)
        g = int(g * a/255)
        b = int(b * a/255)
        
    if format == "argb":
        if for_png:
            # TODO: Verify this
            bytes = g, r, a, b
        else:
            bytes = a, r, g, b
    elif format == "rgba":
        if for_png:
            bytes = b, g, r, a
        else:
            bytes = r, g, b, a
    msg("r(%d), g(%d), b(%d), A(%d) -> %s" % (r, g, b, a, str(bytes)))
    return bytes


def make_unfunky(bytes, format, premul, for_png):
    # Convert data in given format to unpremultiplied RGBA
    if format == "argb":
        if for_png:
            r, a, b, g = bytes
        else:
            a, r, g, b = bytes
    elif format == "rgba":
        if for_png:
            g, r, a, b = bytes
        else:
            r, g, b, a = bytes
    if premul:
        def unpremul(c, a):
            if a == 0:
                return 0
            else:
                return int(c / (a/255))
        r, g, b = tuple(unpremul(c, a) for c in (r, g, b))
    return r, g, b, a


class Bitmap(object):
    """
    An array of pixels representing an image. The colors of individual
    pixels can be set with the :meth:`~Bitmap.set` method.

    *width* and *height* are the width and height of the image in pixels.

    *format* can be either of the strings ``"rgba"`` or ``"argb"``,
    describing the format in which to store the pixels.

    The next two arguments concern how the data will be returned from
    :meth:`~Bitmap.get_data`. If *premultiply_alpha* is ``True``,
    :meth:`~Bitmap.get_data` will return the data with `premultiplied
    alpha
    <http://en.wikipedia.org/wiki/Alpha_compositing#Description>`_.
    Otherwise, the alpha value will be returned unpremultiplied.

    If *for_png* is ``True``, the data will be returned from
    :meth:`~Bitmap.get_data` in network byte order. Otherwise, the
    data will be returned in host byte order.
    """
    def __init__(self, width, height, format="rgba",
                 premultiply_alpha=False, for_png=False):
        if format not in ("rgba", "argb"):
            raise RuntimeError("Illegal format '%s'" % format)
        self.format = format
        self.width = width
        self.height = height
        self.premul = premultiply_alpha
        self.for_png = for_png
        self.buff = array.array("B", (0 for _ in xrange(width * height * 4)))
    def set(self, index, r, g, b, a=255):
        return self.set_pixel(index, r, g, b, a)
    def set_pixel(self, index, r, g, b, a=255):
        """
        Sets the pixel at *index* to the value specified by *r*, *g*,
        *b* and *a*.
        """
        index = index * 4
        bytes = make_funky(r, g, b, a,
                           self.format, self.premul, self.for_png)
        for offset, val in enumerate(bytes):
            self.buff[index+offset] = val
    def get_pixel(self, index):
        """
        Returns a tuple ``(r, g, b, a)`` containing the RGBA value of
        the pixel at *index*.
        """
        index = index * 4
        bytes = self.buff[index:index+4]
        return make_unfunky(bytes,
                            self.format, self.premul, self.for_png)
    def get(self, index):
        return self.get_pixel(index)
    def pprint(self, out=None):
        if out is None:
            out = sys.stdout
        out.write("[")
        for row in (x * self.width for x in xrange(self.height)):
            if row != 0:
                out.write(" ")
            out.write("[")
            for col in xrange(self.width):
                start = (row + col) * 4
                out.write("(%.2X %.2X %.2X %.2X)" %
                          (self.buff[start], self.buff[start+1],
                          self.buff[start+2], self.buff[start+3]))
                if col != self.width -1:
                    out.write(", ")
            if row != (self.width * self.height) - col - 1:
                out.write("],\n")
            else:
                out.write("]")
        out.write("]\n")
    def get_width(self):
        """
        Returns the width of the bitmap in pixels.
        """
        return self.width
    
    def get_height(self):
        """
        Returns the height of the bitmap in pixels.
        """
        return self.height
    def get_format(self):
        """
        Returns a tuple ``(color_format, premul, for_png)`` describing
        the format of the bitmap.  ``color_format`` is either
        ``"rgba"`` or ``"argb"``, and describes the order in which the
        color values are stored. ``premul`` is ``True`` if the alpha
        channel is premultiplied into the color values, otherwise
        ``False``. ``for_png`` is ``True`` if the data is stored in
        network byte order, ``False`` if it is stored in host byte
        order.
        """
        return self.format
    def set_format(self, format=None,
                   premultiply_alpha=None, for_png=None):
        """
        Modifies the format information of the stored data. *format* is either 
        ``"rgba"`` or ``"argb"``, and describes the order in which the
        color values are stored. ``premul`` is ``True`` if the alpha
        channel is premultiplied into the color values, otherwise
        ``False``. ``for_png`` is ``True`` if the data is stored in
        network byte order, ``False`` if it is stored in host byte
        order.
        """
        
        # This is expensive. Don't do it if no change is requested
        if ((format is None and
             premultiply_alpha is None and
             for_png is None) or
            (format == self.format and
             premultiply_alpha == self.premul and
             for_png == self.for_png)):
            pass
        if format is None:
            format = self.format
        if premultiply_alpha is None:
            premultiply_alpha = self.premul
        if for_png is None:
            for_png = self.for_png
        for index in xrange(int(len(self.buff) / 4)):
            raw_index = index * 4
            old_bytes = self.buff[raw_index:raw_index+4]
            r, g, b, a = make_unfunky(old_bytes,
                                      self.format, self.premul,
                                      self.for_png)
            new_bytes = make_funky(r, g, b, a,
                                   format, premultiply_alpha, for_png)
            for offset, val in enumerate(new_bytes):
                self.buff[raw_index+offset] = val
    def get_data(self):
        """
        Returns the raw bytes composing the bitmap.
        """
        return self.buff
        


class PNGBackground(Background):
    """
    A :class:`Background` object which draws the PNG file referenced
    by *filename* as a chart background/foreground. If *fill* is the
    string ``scale``, the image will be stretched/compressed to fit
    the dimensions of the area for which it is providing a
    background/foreground. Otherwise, the image will be rendered at
    its normal resolution, which may underfill or overfill the
    allotted space.
    """
    def __init__(self, filename, fill="scale"):
        Background.__init__(self)
        self.filename = filename
        self.fill = fill
    def draw(self, ctx, width, height):
        Background.draw(self, ctx, width, height)
        ctx.draw_image(self.filename, 0, 0, width, height, self.fill)


class BitmapBackground(Background):
    """
    A :class:`Background` object which renders a *bitmap* (a
    :class:`Bitmap` object) as a chart background/foreground. If
    *fill* is the string ``scale``, the image will be
    stretched/compressed to fit the dimensions of the area for which
    it is providing a background/foreground. Otherwise, the image will
    be rendered at its normal resolution, which may underfill or
    overfill the allotted space.
    """
    def __init__(self, bitmap, fill="none"):
        # For the future: fill exists if the target width and height
        # is larger than the bitmap width and height. Values should be
        # "none" (dont try to fill, just blit at the origin and leave
        # the rest alone), "scale" (scale the bitmap to the target
        # width/height) and "tile" (tile the bitmap over the target
        # width and height).
        #
        # Right now, only "none" is implemented. :)
        Background.__init__(self)
        self.bitmap = bitmap
        self.fill = fill
    def draw(self, ctx, width, height):
        Background.draw(self, ctx, width, height)
        ctx.draw_bitmap(self.bitmap, 0, 0, width, height, self.fill)


class CompositeBackground(Background):
    def __init__(self):
        Background.__init__(self)
        self.backgrounds = []
    def add(self, background):
        self.backgrounds.append(background)
    def draw(self, ctx, width, height):
        Background.draw(self, ctx, width, height)
        for bg in self.backgrounds:
            bg.draw(ctx, width, height)



def get_default_background_image(image):
    return __file__
