#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from backend import Pixels

class Component(object):
    """Base class for Rayon objects that manage space on a
    canvas. Components have boxes within the canvas for which they are
    responsible, and have a notion of minimum size."""
    def layout(self, ctx, width=None, height=None):
        """Administer a region that is I{width} pixels wide and
        I{height} pixels high. Point C{(0, 0)} is presumed to be the
        top-left corner of the region to administer. The subclass may
        delegate administration to children, or may choose to draw in
        the area itself (or leave it blank, as desired.)"""
        assert self.__class__.__name__ != "Component", \
               "Component base class should not be instantiated"
    def components_from_xy(self, x, y):
        """Return a list of the components ultimately responsible for
        anything drawn at point I{(x, y)}.

        @type  x: number
        @param x: x-axis coordinate of the point in user-space units,
            relative to the top-left of layout's bounding box.
        
        @type  y: number
        @param y: y-axis coordinate of the point in user-space units,
            relative to the top-left of layout's bounding box.
        
        @rtype:  list of components
        @return: A list of components underneath this point, or an
            empty list if no such components exist.
        """
        return []
    def minimum_size(self):
        """Return the minimum necessary size for the component to be
        drawn."""
        return (Pixels(0), Pixels(0))


class Panel(Component):
    """Base class for Rayon objects that manage a region of the canvas
    by drawing on it using L{Stamp} objects."""
    def layout(self, ctx, width=None, height=None):
        """In a C{Panel}, this method finalizes the width and height
        of the box and calls C{self.draw}. Subclasses, therefore,
        should probably leave this method alone and define a custom
        C{draw} method instead."""
        if width is None:
            width = ctx.canvas_dimensions()[0]
        if height is None:
            height = ctx.canvas_dimensions()[1]
        self.draw(ctx, width, height)
    def draw(self, ctx, width, height):
        """Draw within the box prescribed by I{width} and I{height},
        with point C{(0,0)} being the top-left corner."""
        pass
        # Useful for debugging:
        # ctx.draw_polygon(((0, 0), (width, 0), (width, height), (0, height)),
        #                  stroke_width=1, stroke_color=(255,0,0),
        #                  fill_color=(0,0,0,0))
    


class Stamp(object):
    """Base class for Rayon objects that draw on a canvas. Stamps are
    used by L{Component} objects to draw pictures in the spaces they
    manage."""
    pass
