#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import sys

import base
from layout.core import BoxSpacing, Layout, point_add
# XXX: circular dependency (layouts.py contains squarelayout, which
# uses borders...)
# TODO: Remove circular dependency
from layout.layouts import GridLayout

import scales
import lines
import text
import ticks
import tickmarkers

NO_PADDING = BoxSpacing(0, 0, 0, 0)

def merge_dictionaries(d1, d2):
    # Merge d1 with d2. d1 may be None. If an option exists in d1
    # already, don't overwrite it. Return a new, merged dictionary.
    if d1 is None:
        d1 = {}
    else:
        d1 = dict(d1)
    for k, v in d2.items():
        if k not in d1:
            d1[k] = v
    return d1

# The defaults for labels, tickmarkers, etc. are tuned assuming that
# vertical stuff is going to the left, and horizontal stuff is going
# on the bottom, mostly. To plot stuff on the right/top requires some
# tweaking.
vertical_right_params = dict(
    tickmarker_args = dict(
        label_position="e",
        halign='left',
    ),
)
    
horizontal_top_params = dict(
    tickmarker_args = dict(
        label_position="n"
    ),
)
    
    

class Border(Layout):
    """
    Base class for Rayon chart borders.
    """
    def __init__(self, **padding):
        Layout.__init__(self, BoxSpacing(**padding))
        # Subclass should set this
        self.child = None

    def layout_leaf(self, ctx, padded_topleft, padded_bottomright,
                    width, height):
        spc = self.spacing
        ctx.translate(spc.left_padding(ctx, width),
                      spc.top_padding(ctx, height))
        self.child.draw(ctx,
                        spc.inner_width(ctx, width),
                        spc.inner_height(ctx, height))
        # No children, so...
        return [], (padded_topleft, padded_bottomright), True
    def layout_branch(self, ctx, padded_topleft, padded_bottomright,
                      width, height):
        spc = self.spacing
        ctx.translate(spc.left_padding(ctx, width),
                      spc.top_padding(ctx, height))
        kids, kid_bbox, is_leaf = self.child.layout(
            ctx, padded_topleft,
            spc.inner_width(ctx, width),
            spc.inner_height(ctx, height))
        return kids, (padded_topleft, padded_bottomright), False
        
    def layout(self, ctx, topleft, width=None, height=None, call_depth=0):
        Layout.layout(self, ctx, topleft, width, height, call_depth)
        width, height = ctx.normalize_dimensions(width, height)

        spc = self.spacing
        padded_topleft = spc.padded_topleft(ctx, topleft, width, height)
        padded_bottomright = spc.padded_bottomright(
            ctx, point_add(topleft, (width, height)), width, height)

        ctx.push_state()
        try:
            # xXx debug
            #ctx.draw_rectangle(0, 0, width, -height,
            #                   1, (255, 0, 0), (0, 0, 0, 10), fixup=False)
            if isinstance(self.child, Layout):
                return self.layout_branch(ctx,
                                          padded_topleft, padded_bottomright,
                                          width, height)
            else:
                return self.layout_leaf(ctx,
                                        padded_topleft, padded_bottomright,
                                        width, height)
        finally:
            ctx.pop_state()


class NoBorder(Border):
    """
    A border which does nothing. Can be used as a placeholder where a
    border is expected.

    See :ref:`chart-spacing` for more information on padding.
    """
    def __init__(self, **padding):
        Border.__init__(self, **padding)
        self.child = base.Panel()

                
        

# Ticks/lines with or without lines/ticks


class TickSetPanel(base.Panel):
    def __init__(self, ticksets=None, line=None):
        if ticksets is None:
            self.ticksets = []
        elif isinstance(ticksets, ticks.Ticks):
            self.ticksets = [ticksets]
        elif isinstance(ticksets, list):
            self.ticksets = list(ticksets)
        else:
            raise RuntimeError("ticksets must be None, a Tickset "
                               "or a list of Ticksets")
        self.line = line
    def add_tickset(self, ticks):
        self.ticksets.append(ticks)
    def num_ticksets(self):
        # TODO: deprecated
        return len(self.ticksets)
    def get_tickset(self, idx):
        return self.ticksets[idx]
    def iter_tick_positions(self, idx):
        # TODO: deprecated
        return self.ticksets[idx].iter_tick_positions()
    def iter_tick_labels(self, idx):
        # TODO: deprecated
        return self.ticksets[idx].iter_tick_labels()
    def iter_ticksets(self):
        return (t for t in self.ticksets)
    # Subclass implements draw(...)

class HorizontalTickSetPanel(TickSetPanel):
    def __init__(self, ticksets=None, line=None, y_offset=0):
        TickSetPanel.__init__(self, ticksets, line)
        if y_offset < 0: y_offset = 0
        if y_offset > 1: y_offset = 1
        self.y_offset = y_offset
    
    def draw(self, ctx, width, height):
        TickSetPanel.draw(self, ctx, width, height)
        y = 0 + (self.y_offset * height)
        if self.line is not None:
            self.line.draw(ctx, 0, y, width, y)
        for ticks in self.ticksets:
            ticks.draw_horizontally(ctx, width, height, y=y)

        # Debugging - draw box around border
        #ctx.draw_rectangle(0, 0, width, -height, 1, (0,0,0), (0,0,0,0))

class VerticalTickSetPanel(TickSetPanel):
    def __init__(self, ticksets=None, line=None, x_offset=0):
        TickSetPanel.__init__(self, ticksets, line)
        if x_offset < 0: x_offset = 0
        if x_offset > 1: x_offset = 1
        self.x_offset = x_offset
    def draw(self, ctx, width, height):
        TickSetPanel.draw(self, ctx, width, height)
        x = width - (self.x_offset * width)
        
        if self.line is not None:
            self.line.draw(ctx, x, 0, x, height)
        for ticks in self.ticksets:
            ticks.draw_vertically(ctx, width, height, x=x)

        # Debugging - draw box around border
        #ctx.draw_rectangle(0, 0, width, -height, 1, (0,0,0), (0,0,0,0))

class TickSetBorder(Border):
    # Abstract class -- should not instantiate
    def add_tickset(self, tickset):
        """
        Adds *tickset* to the tick sets that will be drawn on this
        border. Tick sets will be drawn in the order in which they are
        added.
        """
        return self.child.add_tickset(tickset)
    def iter_tick_positions(self, idx):
        """
        Returns an iterator over the positions of the tick sets stored
        at *idx*.
        """
        # TODO: Deprecated. Remove
        return self.child.iter_tick_positions(idx)
    def iter_tick_labels(self, idx):
        """
        Returns an iterator over the tick labels of the tick sets
        stored at *idx*.
        """
        # TODO: Deprecated. Remove
        return self.child.iter_tick_labels(idx)
    def iter_ticksets(self):
        """
        Returns an iterator over the tick sets stored in this object.
        """
        return self.child.iter_ticksets()
    def num_ticksets(self):
        """
        Returns the number of tick sets stored in this object.
        """
        # TODO: Deprecated. Remove
        return self.child.num_ticksets()
    def get_tickset(self, idx):
        """
        Returns the tick set stored at *idx*.
        """
        return self.child.get_tickset(idx)

class HorizontalBorder(TickSetBorder):
    """
    Draws a simple horizontal border with no decoration.

    If supplied, *ticksets* is a an iterable of :ref:`tick sets
    <ref-ticks>` that will be drawn along the border. Tick sets may be
    added after object creation with the :meth:`add_tickset` method.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, ticksets=None, **padding):
        TickSetBorder.__init__(self, **padding)
        self.child = HorizontalTickSetPanel(ticksets=ticksets)

class VerticalBorder(TickSetBorder):
    """
    Draws a simple vertical border with no decoration.

    If supplied, *ticksets* is a an iterable of :ref:`tick sets
    <ref-ticks>` that will be drawn along the border. Tick sets may be
    added after object creation with the :meth:`add_tickset` method.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, ticksets=None, **padding):
        TickSetBorder.__init__(self, **padding)
        self.child = VerticalTickSetPanel(ticksets=ticksets)

class HorizontalLineBorder(TickSetBorder):
    """
    Draws a horizontal line on the border.
    
    If supplied, *ticksets* is a an iterable of :ref:`tick sets
    <ref-ticks>` that will be drawn along the border. Tick sets may be
    added after object creation with the :meth:`add_tickset` method.

    If supplied, *line* is is a :ref:`line <ref-lines>` that will be
    drawn horizontally across the width of the border.

    If supplied, *y_offset* describes how far from the top of the
    border region the line will be drawn. With a *y_offset* of 0, the
    line will be drawn on the top edge of the border region; with a
    *y_offset* of 1, it will be drawn at the bottom.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, ticksets=None, line=lines.create("solid", fixup=True),
                 y_offset=0, **padding):
        TickSetBorder.__init__(self, **padding)
        self.child = HorizontalTickSetPanel(
            ticksets=ticksets, line=line, y_offset=y_offset)

class HorizontalTopLineBorder(HorizontalLineBorder):
    # This class is Marked For Death in Rayon 2.0
    def __init__(self, ticksets=None, line=lines.create("solid", fixup=True),
                 y_offset=1, **padding):
        HorizontalLineBorder.__init__(
            self, ticksets, line, y_offset=y_offset, **padding)

    
class VerticalLineBorder(TickSetBorder):
    """
    Draws a vertical line on the border.
    
    If supplied, *ticksets* is a an iterable of :ref:`tick sets
    <ref-ticks>` that will be drawn along the border. Tick sets may be
    added after object creation with the :meth:`add_tickset` method.

    If supplied, *line* is is a :ref:`line <ref-lines>` that will be drawn
    vertically across the height of the border.

    If supplied, *x_offset* describes how far from the right edge of the border
    region the line will be drawn. With a *x_offset* of 0, the line
    will be drawn on the far right edge of the border region; with a
    *x_offset* of 1, it will be drawn on the far left edge.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, ticksets=None, line=lines.create("solid", fixup=True),
                 x_offset=0, **padding):
        TickSetBorder.__init__(self, **padding)
        self.child = VerticalTickSetPanel(
            ticksets=ticksets, line=line, x_offset=x_offset)

class VerticalRightLineBorder(VerticalLineBorder):
    """
    A :class:`VerticalLineBorder` optimized for display on the right
    side of a chart.

    This class's default parameters are optimized for displaying to
    the right of the plotting area on a chart; to display a border to
    the left of the plotting area, override the defaults as
    appropriate or use a :class:`VerticalLineBorder` object.

    If supplied, *ticksets* is a sequence of
    :class:`~rayon.ticks.TickSet` objects that will be drawn. More can
    be added after object creation with the
    :meth:`~TickSetBorder.add_tickset` method.

    *line* is is a :class:`rayon.lines.Line` object that will be drawn
    horizontally across the width of the border. *x_offset* describes
    how far from the left of the border region the line will be
    drawn. With a *x_offset* of 0, the line will be drawn to the far
    left; with a *y_offset* of 1, it will be drawn at the far right.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    def __init__(self, ticksets=None, line=lines.create("solid", fixup=True),
                 x_offset=1, **padding):
        VerticalLineBorder.__init__(
            self, ticksets, line, x_offset=x_offset, **padding)

# Ticked Borders

def nice_formatted_ticks(scale, tick_formatter, num_ticks, inside):
    if tick_formatter is None:
        tf = lambda _: None
    elif type(tick_formatter) == str:
        tf = lambda x: tick_formatter % x
    elif callable(tick_formatter):
        tf = tick_formatter
    else:
        raise RayonException(
            "tick formatter must be None, format string or function")
    return ((i, tf(i)) for i in scale.get_nice_tick_positions(num_ticks, inside))


def border_stuff_from_scale(scale, tick_tuples,
                            line, line_type, line_args,
                            tickmarker, tickmarker_type, tickmarker_args):

    tick_tuples = tuple(tick_tuples)

    if tickmarker is None:
        tickmarker = tickmarkers.create(tickmarker_type, **tickmarker_args)
        
    if line is None and line_type is not None:
        line = lines.create(line_type, **line_args)

    if isinstance(scale, scales.RangeScale):
        tickset = ticks.RangeTickSet(tick_tuples, scale, tickmarker)
    else:
        tickset = ticks.TickSet(tick_tuples, scale, tickmarker)

    return tickset, line


class HorizontalNiceTickedBorder(HorizontalLineBorder):
    """
    Draws a horizontal line on the border with automatically-generated
    tick marks.

    *scale* is the :ref:`scale <ref-scales>` for which tick marks are
    desired.

    *num_ticks* is the desired number of tick marks.

    If *inside* is ``True``, Rayon will try to find tick marks that
    fall within the range of *scale*. Otherwise, Rayon will try to
    find tick marks such that the upper and lower bounds of the scale
    fall within the range bounded by the tick marks. Rayon may ignore
    either or both of these parameters if they do not make sense with
    the specified scale.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *y_offset* describes how far from the top of the border region the
    line will be drawn. With a *y_offset* of 0, the line will be drawn
    on the top edge of the border region; with a *y_offset* of 1, it
    will be drawn at the bottom.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, scale, num_ticks=5, inside=False,
                 line=None, line_type="solid", line_args={'fixup': True},
                 y_offset=0,
                 tickmarker=None, tickmarker_type="horizontal",
                 tickmarker_args={}, tick_formatter=None,
                 **padding):
        tick_tuples = nice_formatted_ticks(scale, tick_formatter,
                                           num_ticks, inside)
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        HorizontalLineBorder.__init__(self, [tickset], line, y_offset=y_offset,
                                      **padding)


class HorizontalTopNiceTickedBorder(HorizontalLineBorder):
    """
    A :class:`HorizontalLineBorder` that automatically generates tick
    marks from a :class:`~rayon.scales.Scale` object and user
    preferences.

    This class's default parameters are optimized for displaying above
    the plotting area on a chart; to display a border above the
    plotting area, override the defaults as appropriate or use a
    :class:`HorizontalNiceTickedBorder` object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *num_ticks* is the desired number of tick marks. If *inside* is
    ``True``, Rayon will try to find tick marks that fall within the
    range of the :class:`~rayon.scales.Scale` object. Otherwise, Rayon
    will try to find tick marks such that the upper and lower bounds
    of the scale fall within the range bounded by the tick
    marks. Rayon may ignore either or both of these parameters if they
    do not make sense with the specified :class:`~rayon.scales.Scale`
    object.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *y_offset* is a number between 0 and 1 specifying the height at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn at the top; with a value of 1, it will be drawn
    at the bottom.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    def __init__(self, scale, num_ticks=5, inside=False,
                 line=None, line_type="solid", line_args={'fixup': True},
                 y_offset=1,
                 tickmarker=None, tickmarker_type="horizontal",
                 tickmarker_args=None, tick_formatter=None,
                 **padding):
        tickmarker_args = merge_dictionaries(
            tickmarker_args,
            horizontal_top_params['tickmarker_args'])
            
        tick_tuples = nice_formatted_ticks(scale, tick_formatter,
                                           num_ticks, inside)
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        HorizontalLineBorder.__init__(
            self, [tickset], line, y_offset=y_offset, **padding)
        

class VerticalNiceTickedBorder(VerticalLineBorder):
    """
    A :class:`VerticalLineBorder` that automatically generates tick
    marks from a :class:`~rayon.scales.Scale` object and user
    preferences.

    This class's default parameters are optimized for displaying to
    the left of the plotting area on a chart; to display a border to
    the right of the plotting area, override the defaults as
    appropriate or use a :class:`VerticalRightNiceTickedBorder`
    object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *num_ticks* is the desired number of tick marks. If *inside* is
    ``True``, Rayon will try to find tick marks that fall within the
    range of the :class:`~rayon.scales.Scale` object. Otherwise, Rayon
    will try to find tick marks such that the upper and lower bounds
    of the scale fall within the range bounded by the tick
    marks. Rayon may ignore either or both of these parameters if they
    do not make sense with the specified :class:`~rayon.scales.Scale`
    object.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *x_offset* is a number between 0 and 1 specifying the width at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn at the left; with a value of 1, it will be drawn
    at the right.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A vertical nice-ticked border optimized for display on the left
    def __init__(self, scale, num_ticks=5, inside=False,
                 line=None, line_type="solid", line_args={'fixup': True},
                 x_offset=0,
                 tickmarker=None, tickmarker_type="vertical",
                 tickmarker_args={}, tick_formatter=None,
                 **padding):
        tick_tuples = nice_formatted_ticks(scale, tick_formatter,
                                           num_ticks, inside)
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        VerticalLineBorder.__init__(self, [tickset], line, x_offset=x_offset,
                                    **padding)

class VerticalRightNiceTickedBorder(VerticalLineBorder):
    """
    A :class:`VerticalLineBorder` that automatically generates tick
    marks from a :class:`~rayon.scales.Scale` object and user
    preferences.

    This class's default parameters are optimized for displaying to
    the right of the plotting area on a chart; to display a border to
    the left of the plotting area, override the defaults as
    appropriate or use a :class:`VerticalNiceTickedBorder`
    object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *num_ticks* is the desired number of tick marks. If *inside* is
    ``True``, Rayon will try to find tick marks that fall within the
    range of the :class:`~rayon.scales.Scale` object. Otherwise, Rayon
    will try to find tick marks such that the upper and lower bounds
    of the scale fall within the range bounded by the tick
    marks. Rayon may ignore either or both of these parameters if they
    do not make sense with the specified :class:`~rayon.scales.Scale`
    object.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *x_offset* is a number between 0 and 1 specifying the width at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn at the left; with a value of 1, it will be drawn
    at the right.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A vertical nice-ticked border optimized for display on the right
    def __init__(self, scale, num_ticks=5, inside=False,
                 line=None, line_type="solid", line_args={'fixup': True},
                 x_offset=1,
                 tickmarker=None, tickmarker_type="vertical",
                 tickmarker_args=None, tick_formatter=None,
                 **padding):
        tickmarker_args = merge_dictionaries(
            tickmarker_args,
            vertical_right_params['tickmarker_args'])
            
        tick_tuples = nice_formatted_ticks(scale, tick_formatter,
                                           num_ticks, inside)
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        VerticalLineBorder.__init__(
            self, [tickset], line, x_offset=x_offset, **padding)
    
        


class HorizontalTickedBorder(HorizontalLineBorder):
    """
    A :class:`HorizontalLineBorder` that displays tick
    marks as supplied by a user.

    This class's default parameters are optimized for displaying below
    the plotting area on a chart; to display a border above the
    plotting area, override the defaults as appropriate or use a
    :class:`HorizontalTopTickedBorder` object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *tick_tuples* is a list of 2-tuples. The first item in the tuple
    is data that will be passed to *scale* to obtain a tick mark. The
    second is a string that will be used to label the tick mark. If
    the second item is ``None``, the string representation of the
    first element will be used to label the second element.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *y_offset* is a number between 0 and 1 specifying the height at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn at the top; with a value of 1, it will be drawn
    at the bottom.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A horizontal ticked border optimized for display on the bottom
    def __init__(self, scale, tick_tuples,
                 line=None, line_type="solid", line_args={'fixup': True},
                 y_offset=0,
                 tickmarker=None, tickmarker_type="horizontal",
                 tickmarker_args={}, **padding):
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        HorizontalLineBorder.__init__(self, [tickset], line,
                                      y_offset=y_offset, **padding)

class HorizontalTopTickedBorder(HorizontalLineBorder):
    """
    A :class:`HorizontalLineBorder` that displays tick
    marks as supplied by a user.

    This class's default parameters are optimized for displaying above
    the plotting area on a chart; to display a border below the
    plotting area, override the defaults as appropriate or use a
    :class:`HorizontalTickedBorder` object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *tick_tuples* is a list of 2-tuples. The first item in the tuple
    is data that will be passed to *scale* to obtain a tick mark. The
    second is a string that will be used to label the tick mark. If
    the second item is ``None``, the string representation of the
    first element will be used to label the second element.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *y_offset* is a number between 0 and 1 specifying the height at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn at the top; with a value of 1, it will be drawn
    at the bottom.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A horizontal ticked border optimized for display on the top
    def __init__(self, scale, tick_tuples,
                 line=None, line_type="solid", line_args={'fixup': True},
                 y_offset=1,
                 tickmarker=None, tickmarker_type="horizontal",
                 tickmarker_args=None, **padding):
        tickmarker_args = merge_dictionaries(
            tickmarker_args,
            horizontal_top_params['tickmarker_args'])
            
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        HorizontalLineBorder.__init__(
            self, [tickset], line, y_offset=y_offset, **padding)
        


class VerticalTickedBorder(VerticalLineBorder):
    """
    A :class:`VerticalLineBorder` that displays tick
    marks as supplied by a user.

    This class's default parameters are optimized for displaying to
    the left the plotting area on a chart; to display a border to the
    right of the plotting area, override the defaults as appropriate or
    use a :class:`VerticalRightTickedBorder` object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *tick_tuples* is a list of 2-tuples. The first item in the tuple
    is data that will be passed to *scale* to obtain a tick mark. The
    second is a string that will be used to label the tick mark. If
    the second item is ``None``, the string representation of the
    first element will be used to label the second element.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *x_offset* is a number between 0 and 1 specifying the height at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn to the left; with a value of 1, it will be
    drawn to the right.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A vertical ticked border optimized for display on the left
    def __init__(self, scale, tick_tuples,
                 line=None, line_type="solid", line_args={'fixup': True},
                 x_offset=0,
                 tickmarker=None, tickmarker_type="vertical",
                 tickmarker_args={}, **padding):
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        VerticalLineBorder.__init__(self, [tickset], line, x_offset=x_offset,
                                    **padding)

        
class VerticalRightTickedBorder(VerticalLineBorder):
    """
    A :class:`VerticalLineBorder` that displays tick
    marks as supplied by a user.

    This class's default parameters are optimized for displaying to
    the right the plotting area on a chart; to display a border to the
    left of the plotting area, override the defaults as appropriate or
    use a :class:`VerticalTickedBorder` object.

    *scale* is the :class:`~rayon.scales.Scale` object for which tick
    marks are desired.

    *tick_tuples* is a list of 2-tuples. The first item in the tuple
    is data that will be passed to *scale* to obtain a tick mark. The
    second is a string that will be used to label the tick mark. If
    the second item is ``None``, the string representation of the
    first element will be used to label the second element.

    If supplied, *line* specifies a :class:`~rayon.lines.Line` object
    to use to draw a horizontal line across the border region. If
    *line* is not specified, Rayon will create a
    :class:`~rayon.lines.Line` object based on *line_type* and
    *line_args*.

    *line_type* is a string specifying the style of line to create (as
    it would be used in
    :meth:`~rayon.tools.Toolbox.new_line`). *line_args* is a
    dictionary that will be passed to the constructor for the
    :class:`~rayon.lines.Line` object as keyword arguments. (I.e.,
    Equivalently to calling ``toolbox.new_line(line_type,
    **line_args)``)

    *x_offset* is a number between 0 and 1 specifying the height at
    which the horizontal line will be drawn. With a value of 0, the
    line will be drawn to the left; with a value of 1, it will be
    drawn to the right.

    If supplied, *tickmarker* is a
    :class:`~rayon.tickmarkers.TickMarker` object to use to draw tick
    marks. If *tickmarker* is not specified, Rayon will create a
    :class:`~rayon.tickmarkers.TickMarker` object based on
    *tickmarker_type* and *tickmarker_args*.

    *tickmarker_type* is a string specifying the style of tick marker
    to create (as it would be used in
    :meth:`~rayon.tools.Toolbox.new_tickmarker`). *tickmarker_args* is
    a dictionary that will be passed to the constructor for the
    :class:`~rayon.tickmarkers.TickMarker` object as keyword
    arguments. (I.e., equivalently to calling
    ``tools.new_tickmarker(tickmarker_type, **tickmarker_args)``)

    *tick_formatter* is a format string or a function that takes a
    single argument (the data corresponding to a tick) and returns a
    string that will be used to label the tick mark. The default
    format string is ``\"%s\"``.

    See :ref:`chart-spacing` for more information on spacing
    arguments.
    """
    # A vertical ticked border optimized for display on the right
    def __init__(self, scale, tick_tuples,
                 line=None, line_type="solid", line_args={'fixup': True},
                 x_offset=1,
                 tickmarker=None, tickmarker_type="vertical",
                 tickmarker_args=None, **padding):
        tickmarker_args = merge_dictionaries(
            tickmarker_args,
            vertical_right_params['tickmarker_args'])
            
        tickset, line = border_stuff_from_scale(
            scale, tick_tuples,
            line, line_type, line_args,
            tickmarker, tickmarker_type, tickmarker_args)
        VerticalLineBorder.__init__(
            self, [tickset], line, x_offset=x_offset, **padding)
        


# Labels


class LabeledPanel(base.Panel):
    def __init__(self, label, angle, font_size,
                 font_family, font_style, font_weight,
                 halign, valign, color):
        self.label = text.StaticText(text=label,
                                     font_size=font_size,
                                     font_family=font_family,
                                     font_style=font_style,
                                     font_weight=font_weight,
                                     color=color,
                                     halign=halign, valign=valign,
                                     angle=angle)
        self.angle = angle
    def get_labeler(self): return self.label

class HorizontalLabeledPanel(LabeledPanel):
    def __init__(self, label="", angle=0, font_size="normal",
                 font_family="default", font_style="normal",
                 font_weight="normal", halign="center",
                 valign="center", color=(0, 0, 0, 255)):
        LabeledPanel.__init__(self, label, angle, font_size, font_family,
                              font_style, font_weight, halign,
                              valign, color)

    halign_map = {
        'left'  : lambda w: 0,
        'right' : lambda w: w,
        'center': lambda w: w/2
    }
    valign_map = {
        'top'   : lambda h: 0,
        'bottom': lambda h: h,
        'center': lambda h: h/2
    }
    def draw(self, ctx, width, height):
        LabeledPanel.draw(self, ctx, width, height)
        try:
            x = self.halign_map[self.label.get_horizontal_alignment()](width)
        except KeyError, e:
            raise SyntaxError("Invalid halign '%s'" % e)
        try:
            y = self.valign_map[self.label.get_vertical_alignment()](height)
        except KeyError, e:
            raise SyntaxError("Invalid valign '%s'" % e)
        
        self.label.draw(ctx, x, y, width, height)

        # Debugging - show origin (should be top left of region)
        # import markers
        # m = markers.Labeledmarker(markers.Cross(color=(255,0,0)),
        #                           text.Labeler(color=(255, 0, 0)))
        # m.draw(ctx, 0, 0, width, height, "Origin (%s)" % self.label.content)
        # Debugging - draw box around border
        # ctx.draw_rectangle(0, 0, width, -height, 1, (0,0,0), (0,0,0,0))


class VerticalLabeledPanel(LabeledPanel):
    def __init__(self, label="", angle=0, font_size="normal",
                 font_family="default", font_style="normal",
                 font_weight="normal", halign="center",
                 valign="center", color=(0, 0, 0, 255)):
        LabeledPanel.__init__(self, label, 90 + angle,
                              font_size, font_family,
                              font_style, font_weight, halign,
                              valign, color)

    # Since everything's rotated 90 degrees counterclockwise, the
    # statictext's horizontal alignment actually determines the
    # panel's _vertical_alignment (and vice versa). A horizontal
    # alignment of "left" in the statictext == a vertical alignment of
    # "bottom" here.
    halign_map = {
        'left'  : lambda h: h,
        'right' : lambda h: 0,
        'center': lambda h: h/2
    }
    valign_map = {
        'top'   : lambda w: 0,
        'bottom': lambda w: w,
        'center': lambda w: w/2
    }
    def draw(self, ctx, width, height):
        LabeledPanel.draw(self, ctx, width, height)
        try:
            x = self.valign_map[self.label.get_vertical_alignment()](width)
        except KeyError, e:
            raise SyntaxError("Invalid valign '%s'" % e)
        try:
            y = self.halign_map[self.label.get_horizontal_alignment()](height)
        except KeyError, e:
            raise SyntaxError("Invalid halign '%s'" % e)

        self.label.draw(ctx, x, y, width, height)

        # xXx debug - draw box around border
        # ctx.draw_rectangle(0, 0, width, -height, 1, (0,0,0), (0,0,0,0))


class LabeledBorder(Border):
    """
    Base class for :class:`Border` objects that contain text labels.
    """
    def get_labeler(self):
        return self.child.get_labeler()

class HorizontalLabeledBorder(LabeledBorder):
    """
    Draws a single static text label (e.g., a title or caption)
    horizontally in a border.

    *label* is the text label to apply to the border.

    *font_size*, *font_family* and *font_style* take the values
    described in :ref:`specifying-text-properties`.

    *angle* is an optional angle of rotation, relative to the label's
    default horizontal, left-to-right orientation.

    *halign* and *valign* specify, respectively, the horizontal and
    vertical alignment of the label within the border region.

    *color* is the color in which to draw the text.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, label, font_size="normal",
                 font_family="default", font_style="normal",
                 font_weight="normal", halign="center",
                 valign="center", color=(0, 0, 0, 255),
                 angle=0,
                 **padding):
        # TODO: angle should really be the third argument (after
        # label), but this way preserves backward
        # compatibility. Consider moving this for 2.0
        Border.__init__(self, **padding)
        self.child = HorizontalLabeledPanel(label, angle,
                                            font_size, font_family,
                                            font_style, font_weight,
                                            halign, valign, color)

class VerticalLabeledBorder(LabeledBorder):
    """
    Draws a single static text label (e.g., a title or caption)
    vertically in a border.

    *label* is the text label to apply to the border.

    *font_size*, *font_family* and *font_style* take the values
    described in :ref:`specifying-text-properties`.

    *angle* is an optional angle of rotation, relative to the label's
    default vertical, bottom-to-top orientation.

    *halign* and *valign* specify, respectively, the horizontal and
    vertical alignment of the label within the border region.

    *color* is the color in which to draw the text.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, label, font_size="normal",
                 font_family="default", font_style="normal",
                 font_weight="normal", halign="center",
                 valign="center", color=(0, 0, 0, 255),
                 angle=0,
                 **padding):
        # TODO: angle should really be the third argument (after
        # label), but this way preserves backward
        # compatibility. Consider moving this for 2.0
        Border.__init__(self, **padding)
        self.child = VerticalLabeledPanel(label, angle,
                                          font_size, font_family,
                                          font_style, font_weight,
                                          halign, valign, color)





# Other

class SplitBorder(Border):
    """
    Base class for borders that are subdivided into multiple subcells.
    """
    def __init__(self, children, split="vertical", **padding):
        Border.__init__(self, **padding)        
        # Vertical split:
        #
        #  +-----+-----+
        #  |     |     |
        #  |     |     |
        #  +-----+-----+
        #
        # Horizontal split:
        #
        #  +-----------+
        #  |           |
        #  +-----------+
        #  |           |
        #  +-----------+
        if split == "vertical":
            col_weights, row_weights = ([1], [1] * len(children))
        elif split == "horizontal":
            col_weights, row_weights = ([1] * len(children), [1])
        grid = GridLayout(self.spacing, col_weights, row_weights, split)
        for c in children:
            grid.add_layout(c)
        self.child = grid


class VerticalSplitBorder(SplitBorder):
    """
    Draws a vertical border, split into cells arranged from left to
    right. Each contains a child border, which is drawn on the
    portion of the region allocated to it.

    *children* is a list of borders which will be drawn in the
    subregions, from left to right. There will be as many subregions
    as there are children.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, children, **padding):
        SplitBorder.__init__(self, children, split='vertical', **padding)
    
class HorizontalSplitBorder(SplitBorder):
    """
    Draws a horizontal border, split into cells arranged from top to
    bottom. Each contains a child border, which is drawn on
    the portion of the region allocated to it.

    *children* is a list of borders which will be drawn in the
    subregions, from left to right. There will be as many subregions
    as there are children.

    *padding* represents zero or more padding arguments. (See
    :ref:`ref-charts-padding` for more information on padding.)
    """
    def __init__(self, children, **padding):
        SplitBorder.__init__(self, children, split='horizontal', **padding)



#
# Functional interface
#
            

empty_border = NoBorder()


_border_classes = {
    'none'         : NoBorder,
    'horizontal'   : HorizontalBorder,
    'vertical'     : VerticalBorder,
    'hline'        : HorizontalLineBorder,
    'vline'        : VerticalLineBorder,
    'hlabel'       : HorizontalLabeledBorder,
    'vlabel'       : VerticalLabeledBorder,
    'vsplit'       : VerticalSplitBorder,
    'hsplit'       : HorizontalSplitBorder
}

# These are used elsewhere in Rayon and some supporting code, but
# their future is questionable. On the one hand, it's a PITA to
# configure every border for top/bottom/left/right. OTOH, it makes the
# documentation ridiculously complicated.
_alternative_border_classes = {
    'h'            : HorizontalBorder,
    'v'            : VerticalBorder,
    'hline_bottom' : HorizontalLineBorder,
    'hline_top'    : HorizontalTopLineBorder,
    'vline_left'   : VerticalLineBorder,
    'vline_right'  : VerticalRightLineBorder,
    'hnice_bottom' : HorizontalNiceTickedBorder,
    'hnice_top'    : HorizontalTopNiceTickedBorder,
    'hnice'        : HorizontalNiceTickedBorder,
    'vnice'        : VerticalNiceTickedBorder,
    'htick'        : HorizontalTickedBorder,
    'vtick'        : VerticalTickedBorder,
    'vnice_left'   : VerticalNiceTickedBorder,
    'vnice_right'  : VerticalRightNiceTickedBorder,
    'htick_bottom' : HorizontalTickedBorder,
    'htick_top'    : HorizontalTopTickedBorder,
    'vtick_left'   : VerticalTickedBorder,
    'vtick_right'  : VerticalRightTickedBorder,
}

_vertical_border_classes = {
    'line'    : VerticalLineBorder,
    'none'    : VerticalBorder,
    'nice'    : VerticalNiceTickedBorder,
    'tick'    : VerticalTickedBorder,
    'labeled' : VerticalLabeledBorder,
}
_horizontal_border_classes = {
    'line'    : HorizontalLineBorder,
    'none'    : HorizontalBorder,
    'nice'    : HorizontalNiceTickedBorder,
    'tick'    : HorizontalTickedBorder,
    'labeled' : HorizontalLabeledBorder,
}

def create(border_style, ticksets=None, **border_params):
    try:
        border_cls = _border_classes[border_style]
    except KeyError, e:
        try:
            border_cls = _alternative_border_classes[border_style]
        except KeyError, i:
            raise SyntaxError("No such border style %s" % str(e))
    return border_cls(**border_params)

# NOTE: create_vertical and create_horizontal are used by the rytools
# so users can just say "line" and "labeled" and it will do what they
# mean. They are quite inelegant, and though they're not scheduled for
# demolition, civic leaders eye them warily and lock their doors every
# time they pass by, so I wouldn't count on them living forever. --pg
def create_vertical(border_style, **border_params):
    try:
        return _vertical_border_classes[border_style](**border_params)
    except KeyError, e:
        raise SyntaxError("No such vertical border style %s" % str(e))

def create_horizontal(border_style, **border_params):
    try:
        return _horizontal_border_classes[border_style](**border_params)
    except KeyError, e:
        raise SyntaxError("No such horizontal border style %s" % str(e))

def list_border_types():
    return tuple((k, v) for k, v in _border_classes.items())

def list_styles():
    return sorted(_border_classes.keys())


