#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
"""
from __future__ import division
import miscutils

import sys

import borders
import backgrounds
import plots
from rayon import RayonException
from layout.core import (
    Layout,
    LayoutNode,
    BoxSpacing,
)
from layout.layouts import (
    GridLayout,
    LayeredLayout,
    SimpleLayout,
    SquareLayout,
)
from backend import units_from_spec

class Chart(Layout):
    def __init__(self, backend):
        Layout.__init__(self, BoxSpacing(0, 0, 0, 0))
        self.backend = backend
        self.parent = None
        # When the chart is laid out, node_tree is the LayoutNode tree
        # describing how space in the chart's area is used. We use
        # this in event handling to figure out what components are at
        # a given position.
        self.node_tree = None
        # A registry of event adapters created, keyed to the rendering
        # surface to which the adaptor is bound.
        self.event_adapters = {}

        # The layout tree looks like this, from bottom to top:
        #
        # main (LayeredLayout)
        # +--chart_backgrounds (LayeredLayout)
        # +--outer_square (SquareLayout)
        #    +--inner_square (SquareLayout)
        # +--chart_foregrounds (LayeredLayout)

        self.main = LayeredLayout(BoxSpacing())
        
        self.chart_backgrounds = LayeredLayout(BoxSpacing())
        self.main.add_layout(self.chart_backgrounds, "backgrounds")

        self.outer_square = SquareLayout(BoxSpacing())
        self.main.add_layout(self.outer_square, "outer_square")
        
        self.inner_square = SquareLayout(BoxSpacing())
        self.inner_square.set_nw_corner(LayeredLayout(BoxSpacing()))
        self.inner_square.set_ne_corner(LayeredLayout(BoxSpacing()))
        self.inner_square.set_se_corner(LayeredLayout(BoxSpacing()))
        self.inner_square.set_sw_corner(LayeredLayout(BoxSpacing()))
        self.outer_square.set_center(self.inner_square)

        self.chart_foregrounds = LayeredLayout(BoxSpacing())
        self.main.add_layout(self.chart_foregrounds, "foregrounds")

    def set_padding(self, allpad=None, tpad=0, bpad=0, lpad=0, rpad=0):
        """
        Sets the amount of padding around the edges of the
        chart.

        If specified, *allpad* is a single size specification (see
        :ref:`specifying-sizes`). The amount of padding specified by
        *allpad* will be applied to all four edges of the canvas. The
        remaining arguments will be ignored.

        *tpad*, *bpad*, *lpad* and *rpad* are size specifications
        indicating the amound of padding that should be applied to the
        top, bottom, left and right edges of the canvas, respectively.
        """
        if allpad is None:
            self.outer_square.set_spacing(BoxSpacing(tpad=tpad, bpad=bpad,
                                                     lpad=lpad, rpad=rpad))
        else:
            self.outer_square.set_spacing(BoxSpacing(tpad=allpad, bpad=allpad,
                                                     lpad=allpad, rpad=allpad))

    def clear_padding(self):
        """
        Remove all padding from the chart. This does not remove the
        padding from the interior components of a chart, only its
        outer edges.
        """
        self.set_padding()
    

    def _box_component(self, c, **padding):
        # Used by subclasses to wrap the Plot/Chart/other Layout in
        # something predictable that has spacing arguments
        if isinstance(c, Layout):
            # A little hackish -- use a LayeredLayout with one layer
            # as a wrapper for the Layout
            box = LayeredLayout(BoxSpacing(**padding))
            box.add_layout(c, "only_layout")
        else:
            box = SimpleLayout(c, BoxSpacing(**padding))
        return box

    def _unbox_component(self, c):
        # Used by subclasses to remove the wrapper for c before
        # returning it.
        if isinstance(c, LayeredLayout):
            return c.get_layout("only_layout")
        else:
            return c.get_component()
            
    def add_plot(self, plot, name=None, **padding):
        # Documented in subclasses
        if type(self) == Chart:
            raise NotImplemented("'add_plot' must be implemented by subclasses")

    def add_chart(self, chart, name=None, **padding):
        """
        Adds a chart object to the plotting area.

        *chart* is the chart to add.

        If supplied, *name* is a string that can be used later to
        retrieve *chart* from this object using :meth:`get_chart`.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        return self.add_plot(chart, name, **padding)

    def get_plot_element(self, name=None):
        """
        Returns a plot or chart associated with this chart in the
        plotting area. 

        *name* is the name given to the plot or chart when it was
        added using :meth:`add_plot`.  If no plot or chart
        corresponding to *name* is associated with this chart, this
        method returns ``None``.
        """
        if type(self) == Chart:
            raise NotImplemented(
                "'get_plot_element' must be implemented by subclasses")

    def get_plot(self, name=None):
        """
        Returns a plot associated with this chart in the plotting
        area.

        *name* is the name given to the plot or chart when it was
        added using :meth:`add_plot`.  If no plot corresponding to
        *name* is associated with this chart, this method returns
        ``None``.
        """
        rc = self.get_plot_element(name)
        if isinstance(rc, plots.Plot):
            return rc
        else:
            return None
        
    def get_chart(self, name=None):
        """
        Returns a chart associated with this chart in the plotting
        area.

        *name* is the name given to the associated chart when it was
        added using :meth:`add_chart`.  If no chart corresponding to
        *name* is associated with this chart, this method returns
        ``None``.
        """
        rc = self.get_plot_element(name)
        if isinstance(rc, Chart):
            return rc
        else:
            return None

    def get_plot_elements(self):
        """
        Returns a list of all chart and plot objects associated with
        the plotting area.
        """
        if type(self) == Chart:
            raise NotImplemented(
                "'get_plot_elements' must be implemented by subclasses")
        
    def get_plots(self):
        """
        Returns a list of all plot objects associated with the
        plotting area.
        """
        return [c for c in self.get_plots() if isinstance(c, plots.Plot)]
        
    def get_charts(self):
        """
        Gets all chart objects associated with this chart's plotting
        area.
        """
        return [c for c in self.get_plots() if isinstance(c, Chart)]

        

    # Borders
    #
    def add_left_border(self, border, width):
        """
        Adds a border to the left edge of the chart.

        *border* is a Rayon border object. (See :ref:`ref-borders`.)

        *width* is a size specification (see :ref:`specifying-sizes`)
        of the border width.

        This border runs from the top edge of the plot area to the
        bottom edge. To add a border that spans the whole height of the
        chart, see :meth:`add_left_title`.
        """
        self.inner_square.add_left_border(border, units_from_spec(width))
            
    def add_right_border(self, border, width):
        """
        Adds a border to the right edge of the chart.

        *border* is a Rayon border object. (See :ref:`ref-borders`.)

        *width* is a size specification (see :ref:`specifying-sizes`)
        of the border width.

        This border runs from the top edge of the plot area to the
        bottom edge. To add a border that spans the whole height of the
        chart, see :meth:`add_right_title`.
        """
        self.inner_square.add_right_border(border, units_from_spec(width))
        
    def add_top_border(self, border, height):
        """
        Adds a border to the top edge of the chart.

        *border* is a Rayon border object. (See :ref:`ref-borders`.)

        *height* is a size specification (see :ref:`specifying-sizes`)
        of the border height.

        This border runs from the left edge of the plot area to the
        right edge. To add a border that spans the whole width of the
        chart, see :meth:`add_top_title`.
        """
        self.inner_square.add_top_border(border, units_from_spec(height))
        
    def add_bottom_border(self, border, height):
        """
        Adds a border to the bottom edge of the chart.

        *border* is a Rayon border object. (See :ref:`ref-borders`.)

        *height* is a size specification (see
        :ref:`specifying-sizes`) of the border height.

        This border runs from the left edge of the plot area to the
        right edge. To add a border that spans the whole width of the
        chart, see :meth:`add_bottom_title`.
        """
        self.inner_square.add_bottom_border(border,
                                            units_from_spec(height))

    # Titles
    #
    # Note: titles are like borders, but they span the left and right
    # borders (for top/bottom titles) or top and bottom borders (for
    # left/right titles) of the chart. However, top/bottom titles do
    # not span over left/right titles, and vice versa. (TODO: this may
    # change)
    def add_top_title(self, title_border, height):
        """
        Adds a border to the top edge of the chart. Unlike
        :meth:`add_top_border`, this border runs the entire width
        of the chart, and over any left or right borders the chart may
        have.

        *title_border* is a Rayon border object. (See :ref:`ref-borders`.)

        *height* is a size specification (see
        :ref:`specifying-sizes`) of the border height.
        """
        self.outer_square.add_top_border(title_border,
                                         units_from_spec(height))

    def add_bottom_title(self, title_border, height):
        """
        Adds a border to the bottom edge of the chart. Unlike
        :meth:`add_bottom_border`, this border runs the entire width
        of the chart, and under any left or right borders the chart may
        have.

        *title_border* is a Rayon border object. (See :ref:`ref-borders`.)

        *height* is a size specification (see
        :ref:`specifying-sizes`) of the border height.
        """
        self.outer_square.add_bottom_border(title_border,
                                            units_from_spec(height))

    def add_left_title(self, title_border, width):
        """
        Adds a border to the left edge of the chart. Unlike
        :meth:`add_left_border`, this border runs the entire height of
        the chart, and to the left of any top or bottom borders the
        chart may have.

        *title_border* is a Rayon border object. (See :ref:`ref-borders`.)

        *width* is a size specification (see :ref:`specifying-sizes`)
        of the border width.
        """
        self.outer_square.add_left_border(title_border,
                                          units_from_spec(width))

    def add_right_title(self, title_border, width):
        """
        Adds a border to the right edge of the chart. Unlike
        :meth:`add_right_border`, this border runs the entire height of
        the chart, and to the right of any top or bottom borders the
        chart may have.

        *title_border* is a Rayon border object. (See :ref:`ref-borders`.)

        *width* is a size specification (see :ref:`specifying-sizes`)
        of the border width.
        """
        self.outer_square.add_right_border(title_border,
                                           units_from_spec(width))


    def add_padding(self, directions, size):
        """
        Adds one or more empty borders with a width/height of *size*,
        in order to pad the edges of a chart.

        *directions* is a comma-separated list of one or more of the
        literals ``top``, ``bottom``, ``left`` or ``right``,
        representing the sides on which to add padding.

        *size* is a is a size specification (see
        :ref:`layout-size-specs`) of the amount of padding to add.

        This method is deprecated: use :meth:`Chart.set_padding` instead.
        """
        if directions == 'all':
            directions = 'left,right,top,bottom'
        directions = directions.split(',')
        for d in directions:
            if d not in ('left', 'right', 'top', 'bottom'):
                raise RayonException("illegal direction spec '%s'" % directions)
        for d in directions:
            border = borders.create('none')
            if d == 'left':
                self.add_left_title(border, width=size)
            elif d == 'right':
                self.add_right_title(border, width=size)
            elif d =='top':
                self.add_top_title(border, height=size)
            elif d =='bottom':
                self.add_bottom_title(border, height=size)

    def add_custom_padding(self, left=0, right=0, top=0, bottom=0):
        # TODO: deprecated -- get rid of this
        b = borders.create('none')
        if left != 0:
            self.add_left_border(b, width=left)
        if right != 0:
            self.add_right_border(b, width=right)
        if top != 0:
            self.add_top_border(b, height=top)
        if bottom != 0:
            self.add_bottom_border(b, height=bottom)

    # Corners
    def __add_corner(self, corner, component, **padding):
        if not isinstance(component, Layout):
            component = SimpleLayout(component, BoxSpacing(**padding))
        corner.add_layout(component)
    def add_nw_corner(self, element, **padding):
        """
        Adds an element to the northwest corner of the chart.

        *element* is a Rayon plot or chart.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        self.__add_corner(self.inner_square.get_nw_corner(), element,
                          **padding)
    def add_ne_corner(self, element, **padding):
        """
        Adds an element to the northeast corner of the chart.

        *element* is a Rayon plot or chart.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        self.__add_corner(self.inner_square.get_ne_corner(), element,
                          **padding)
    def add_se_corner(self, element, **padding):
        """
        Adds *element* to the southeast corner of the chart.

        *element* is a Rayon plot or chart.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        self.__add_corner(self.inner_square.get_se_corner(), element,
                          **padding)
    def add_sw_corner(self, element, **padding):
        """
        Adds *element* to the southwest corner of the chart.

        *element* is a Rayon plot or chart.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        self.__add_corner(self.inner_square.get_sw_corner(), element,
                          **padding)

    # Chart fore/backgrounds
    def set_chart_foreground(self, foreground, **padding):
        """
        Defines a forground image or color for the entire chart. This
        image or color will be drawn in front of any other element,
        and will cover the entire drawable space of the chart,
        including borders and titles. Alpha transparency is observed
        in both images and colors.

        *foreground* is either a Rayon :ref:`background
        <ref-background>` object or a color specification. (See
        :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(foreground, backgrounds.Background):
            foreground = backgrounds.ColorBackground(foreground)
        self.chart_foregrounds.add_layout(
            SimpleLayout(foreground, BoxSpacing(**padding)))

    def set_chart_background(self, background, **padding):
        """
        Defines the background image or color for the entire
        chart. This image or color will be drawn behind any other
        element, and will cover the entire drawable space for the
        chart, including borders and titles.

        *background* is either a Rayon :ref:`background
        <ref-background>` object or a color specification. (See
        :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(background, backgrounds.Background):
            background = backgrounds.ColorBackground(background)
        self.chart_backgrounds.add_layout(
            SimpleLayout(background, BoxSpacing(**padding)))

    # Plot background
    def set_plot_background(self, background, **padding):
        # Documented in subclasses
        if type(self) == Chart:
            raise NotImplemented(
                "'set_plot_background' must be implemented in subclass")

    # Plot foreground
    def set_plot_foreground(self, foreground, **padding):
        # Documented in subclasses
        if type(self) == Chart:
            raise NotImplemented(
                "'set_plot_foreground' must be implemented in subclass")
        


    # Event handling
    def set_parent(self, parent):
        """
        Sets the parent of a chart to *parent* for the purposes of
        event handling in an interactive rendering backend. Does
        nothing if the backend is non-interactive.
        """
        if self.parent is not None:
            raise RayonException("Cannot change parents of a chart.")
        self.parent = parent

    def get_event_source(self, gui_event_source, activate=True):
        """
        Creates and returns a Rayon event source that maps this chart
        to *gui_event_source*. See :ref:`events` for more information
        on event handling in Rayon.

        .. note:: Event handling is only implemented in interactive rendering backends. Non-interactive backends provide this function for consistency, but it does nothing.

        *gui_event_source* is a GUI window or panel which will receive and
        dispatch UI events. If the backend is non-interactive, the
        type of *event_source* is unimportant.

        If *activate* is ``True``, the Rayon event source will be
        created in an activated state -- i.e., it will immediately
        begin dispatching events from *event_source* to the chart. If
        *activate* is ``False``, the user must activate the event
        source using the source's :meth:`~events.EventSource.activate`
        method.
        """
        # TODO: check against event code, make sure this is accurate.
        if not self.event_adapters.has_key(gui_event_source):
            self.event_adapters[gui_event_source] = (
                self.backend.EventSource(self, gui_event_source))
        adapter = self.event_adapters[gui_event_source]
        if activate:
            adapter.activate()
        return adapter


    def layout(self, ctx, topleft=(0,0), width=None, height=None, call_depth=0):
        Layout.layout(self, ctx, topleft, width, height, call_depth)
        grandkids, kid_bbox, is_leaf = self.main.layout(
            ctx, topleft, width, height, call_depth+1)
        topnode = LayoutNode(grandkids, kid_bbox[0], kid_bbox[1], 0,
                               self.main, is_leaf)
        #return [topnode], ((0, 0), (width, height)), False
        return [topnode], (topleft, (width, height)), False
        
        

    def __call__(self, ctx, use_dimensions=None):
        """Called by users to lay out the chart. (L{layout} is used
        internally.)"""
        if use_dimensions is None:
            width, height = ctx.canvas_dimensions()
        else:
            width, height = use_dimensions
        topnode, kid_bbox, is_leaf = self.layout(ctx, (0, 0), width, height)
        self.node_tree = topnode[0]


    def xy_to_components(self, x, y):
        # Given a spatial position in underlying coordinate space,
        # return the component(s) drawn at that point in space. If the
        # top-level component is composite, return the top-level
        # component
        #
        # @type  x: number
        # @param x: x-axis coordinate of the point
        #
        # @type  y: number
        # @param y: y-axis coordinate of the point
        #
        # @rtype:  tuple
        # @return: triple of the component, and the x and y coordinates,
        #     respectively, of the top left corner of the component's
        #    bounding box
        subnodes = []
        def callback(subnode, _x, _y, topleft):
            if subnode.is_leaf():
                subnodes.append((subnode.get_component().child,
                                 topleft[0], topleft[1]))
        if self.node_tree is not None:
            self.node_tree.search(x, y, callback)
        return subnodes
        
        




class SquareChart(Chart):
    """
    Renders each plot in a series of layers, each of which has the
    full dimension of the plotting area.  Plots will be drawn in the
    order in which they were added, with more recently-added ones
    drawn \"on top\".
    """
    
    def __init__(self, backend):
        # Note: Toolbox.new_chart() doesn't take a backend parameter
        # (it passes along the one it has), which is why the
        # documentation above doesn't mention it.
        Chart.__init__(self, backend)

        # The center
        self.center = LayeredLayout(BoxSpacing())

        # Assemble the center layers, from the bottom up
        
        self.plot_backgrounds = LayeredLayout(BoxSpacing())
        self.center.add_layout(self.plot_backgrounds, "center.plot_backgrounds")
        
        self.rear_gridlines = LayeredLayout(BoxSpacing())
        self.center.add_layout(self.rear_gridlines, "center.rear_gridlines")

        self.plots = LayeredLayout(BoxSpacing())
        self.center.add_layout(self.plots, "center.plots")
        
        self.front_gridlines = LayeredLayout(BoxSpacing())
        self.center.add_layout(self.front_gridlines, "center.front_gridlines")
        
        self.plot_foregrounds = LayeredLayout(BoxSpacing())
        self.center.add_layout(self.plot_foregrounds, "center.plot_foregrounds")

        self.inner_square.set_center(self.center)


    def add_plot(self, plot, name=None, **padding):
        """
        Adds a plot to the plotting area.

        *plot* is the plot to add.

        If supplied, *name* is a string that can be used later to
        retrieve *plot* from this object using :meth:`get_plot`.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        Chart.add_plot(self, plot, name, **padding)
        plot_layout = self._box_component(plot, **padding)
        self.plots.add_layout(plot_layout, name)
                             

    def get_plot_element(self, name=None):
        """
        Returns a plot or chart associated with this chart in the
        plotting area. 

        *name* is the name given to the plot or chart when it was
        added using :meth:`add_plot`.  If no plot or chart
        corresponding to *name* is associated with this chart, this
        method returns ``None``.
        """
        plot_layout = self.plots.get_layout(name)
        if plot_layout is None:
            return None
        rc =  self._unbox_component(plot_layout)
        if isinstance(rc, plots.Plot):
            return rc
        else:
            return None

    def get_plot_elements(self):
        """
        Returns a list of all chart and plot objects associated with
        the plotting area.
        """
        return [self._unbox_component(l) for l in self.plots.layouts_iter()]

    def set_plot_background(self, background, **padding):
        """
        Defines a background image for the plotting area. This image
        or color will be drawn behind any other element.

        *background* is either a Rayon background object or a color
        specification. (See :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(background, backgrounds.Background):
            background = backgrounds.ColorBackground(background)
        self.plot_backgrounds.add_layout(
            SimpleLayout(background, BoxSpacing(**padding)))
        

    def set_plot_foreground(self, foreground, **padding):
        """
        Defines a foreground image for the plotting area. This image
        or color will be drawn behind any other element. Alpha
        transparency is observed in both images and colors.

        *foreground* is either a Rayon background object or a color
        specification. (See :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(foreground, backgrounds.Background):
            foreground = backgrounds.ColorBackground(foreground)
        self.plot_foregrounds.add_layout(
            SimpleLayout(foreground, BoxSpacing(**padding)))

    def add_rear_gridlines(self, grid, **padding):
        """
        Adds grid lines to the plotting area. The grid lines will
        appear behind any other elements in the plotting area (except
        grid lines that have already been added with this
        method). Grid lines will be drawn in the order they were added
        to the chart, with the most recently added ones drawn on top.

        *grid* is a Rayon grid lines object.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)

        .. seealso:: :ref:`classref-grids`
        """
        self.rear_gridlines.add_layout(
            SimpleLayout(grid, BoxSpacing(**padding)))

    def add_front_gridlines(self, grid, **padding):
        """
        Adds grid lines to the plotting area. The grid lines will
        appear on top of any other elements in the plotting area
        (except grid lines that are subsequently added with this
        method). Grid lines will be drawn in the order they were added
        to the chart, with the most recently added ones drawn on top.

        *grid* is a Rayon grid lines object.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)

        .. seealso:: :ref:`classref-grids`
        """
        self.front_gridlines.add_layout(
            SimpleLayout(grid, BoxSpacing(**padding)))




class TiledChart(Chart):
    """
    Subdivides the plotting area into a tiled grid according to a set
    of user-supplied weights, and renders plots and charts into tiles
    in the grid. This chart offers the user more flexibility in
    creating the grid than the :ry:factory:`SimpleTiledChart` chart,
    but has a more complicated interface.

    If supplied, *col_weights* is an iterable containing the integer
    weights each column should have. A list ``[1, 1, 1]`` indicates
    three columns of equal weight; ``[2, 1, 1]`` indicates three
    columns, where the first (left-most) is twice as wide as the other
    two. If *col_weights* is not supplied, the grid will have as many
    equally-weighted columns as needed to tile the required number of
    cells, given the number of rows.

    If supplied, *row_weights* is an iterable of integer weights each
    row should have, with the same semantics as *col_weights*. If not
    supplied, the grid will have as many equally-weighted rows as
    needed to tile the required number of cells, given the number of
    columns.

    One of *row_weights* or *col_weights* must be supplied. If both
    are supplied, the number of rows and columns is fixed, and the
    maximum number of cells is the product of the number of columns
    and the number of rows; attempting to add more than that will
    generate an error.

    *orientation* determines the order in which cells in the grid
    will be filled, and must be one of the strings ``horizontal``,
    ``vertical`` or ``auto``. Horizontal tiling fills the top row
    first, from left to right, then the next row left-to-right, and so
    on until all the cells are filled or no more objects remain to be
    added. Vertical tiling fills the leftmost column first, from top
    to bottom, then the next column to the right top-to-bottom, and so
    on. ``auto`` indicates that vertical tiling should be used if
    *col_weights* is specified, and horizontal tiling should be used
    if *row_weights* is specified. If both *col_weights* and
    *row_weights* are supplied and *orientation* is ``auto``,
    horizontal tiling will be used.
    """
    def __init__(self, backend, col_weights=[], row_weights=[],
                 orientation='vertical', **padding):
        # Note that the documentation doesn't mention backend, because
        # it's documenting toolbox.new_chart, not __init__. Also,
        # **padding is undocumented -- use set_padding() instead.
        #
        # TODO: remove **padding for Rayon 2.0
        # TODO: make orientation default to 'auto' for Rayon 2.0
        if col_weights == [] and row_weights == []:
            raise RayonException("Must specify at least column or row weights")
        Chart.__init__(self, backend)
        # Replace spacing information in superclass with this one
        self.spacing = BoxSpacing(**padding)
        
        self.center = LayeredLayout(self.spacing)

        self.center_backgrounds = LayeredLayout(self.spacing)
        self.center.add_layout(self.center_backgrounds, "center.backgrounds")

        self.grid = GridLayout(self.spacing, col_weights,
                               row_weights, orientation)
        self.center.add_layout(self.grid, "center.grid")

        self.center_foregrounds = LayeredLayout(self.spacing)
        self.center.add_layout(self.center_foregrounds, "center.foregrounds")

        self.inner_square.set_center(self.center)

    def set_plot_background(self, background, **padding):
        """
        Defines a background image for the plotting area. This image
        or color will be drawn behind any other element.

        *background* is either a Rayon background object or a color
        specification. (See :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(background, backgrounds.Background):
            background = backgrounds.ColorBackground(background)
        self.center_backgrounds.add_layout(
            SimpleLayout(background, BoxSpacing(**padding)))

    def set_plot_foreground(self, foreground, **padding):
        """
        Defines a foreground image for the plotting area. This image
        or color will be drawn behind any other element. Alpha
        transparency is observed in both images and colors.

        *foreground* is either a Rayon background object or a color
        specification. (See :ref:`specifying-colors`.)

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        if not isinstance(foreground, backgrounds.Background):
            foreground = backgrounds.ColorBackground(foreground)
        self.center_foregrounds.add_layout(
            SimpleLayout(foreground, BoxSpacing(**padding)))


    def add_plot(self, plot, name=None, **padding):
        """
        Adds a plot to the plotting area.

        *plot* is the plot to add.

        If supplied, *name* is a string that can be used later to
        retrieve *plot* from this object using :meth:`get_plot`.

        *padding* represents zero or more padding arguments. (See
        :ref:`ref-charts-padding` for more information on padding.)
        """
        Chart.add_plot(self, plot, name, **padding)
        plot_layout = self._box_component(plot, **padding)
        self.grid.add_layout(plot_layout, name)

    def get_plot_element(self, name=None):
        """
        Returns a plot or chart associated with this chart in the
        plotting area. 

        *name* is the name given to the plot or chart when it was
        added using :meth:`add_plot`.  If no plot or chart
        corresponding to *name* is associated with this chart, this
        method returns ``None``.
        """
        plot_layout = self.grid.get_layout(name)
        if plot_layout is None:
            return None
        return self._unbox_component(plot_layout)
        
    def get_plot_elements(self):
        """
        Returns a list of all chart and plot objects associated with
        the plotting area.
        """
        return [self._unbox_component(c) for c in self.grid.layouts_iter()]


class SimpleTiledChart(TiledChart):
    """
    Subdivides the plotting area into a tiled grid with the specified
    number of columns and rows, and renders plots and charts into
    tiles in the grid. All cells have equal width and height. To
    control the width and height of columns and rows in the grid, use
    the :ry:factory:`TiledChart` chart.

    *num_columns*, if supplied, is the number of columns in the
    grid. If not supplied, the grid will have as many columns as
    needed to tile the required number of cells into *num_rows*.

    *num_rows*, if supplied, is the number of rows in the grid. If not
    supplied, the grid will have as many rows as needed to tile the
    required number of cells into *num_columns*.

    One of *num_columns* or *num_rows* must be supplied. If both are
    supplied, the maximum number of cells is the product of
    *num_columns* and *num_rows*; attempting to add more than that
    will generate an error.

    *orientation* determines the order in which cells in the grid
    will be filled, and must be one of the strings ``horizontal``,
    ``vertical`` or ``auto``. Horizontal tiling fills the top row
    first, from left to right, then the next row left-to-right, and so
    on until all the cells are filled or no more objects remain to be
    added. Vertical tiling fills the leftmost column first, from top
    to bottom, then the next column to the right top-to-bottom, and so
    on. ``auto`` indicates that vertical tiling should be used if
    *col_weights* is specified, and horizontal tiling should be used
    if *row_weights* is specified. If both *col_weights* and
    *row_weights* are supplied and *orientation* is ``auto``,
    horizontal tiling will be used.
    """
    def __init__(self, backend, num_columns=None, num_rows=None,
                 orientation='auto', **padding):
        def empty_rows(): return num_rows is None or num_rows <= 0
        def empty_cols(): return num_columns is None or num_columns <= 0
        if empty_rows() and empty_cols():
            raise RayonException("Must specify positive number for "
                                 "either rows or columns")
        if orientation not in ('auto', 'horizontal', 'vertical'):
            raise RayonException("Invalid orientation '%s' "
                                 "(Must be 'horizontal', 'vertical', "
                                 "or 'auto')" % orientation)
        if empty_rows():
            # Columns are constrained. Use as many rows as
            # necessary. By default, fill the cells top-to-bottom,
            # from the top left.
            col_weights = [1] * num_columns
            row_weights = []
            if orientation == 'auto':
                orientation = 'vertical'
        elif empty_cols():
            # Rows are constrained. Use as many columns as
            # necessary. By default, fill the cells left-to-right,
            # from the top left.
            row_weights = [1] * num_rows
            col_weights = []
            if orientation == 'auto':
                orientation = 'horizontal'
        else:
            # Both rows and columns are constrained. By default, fill
            # cells left-to-right, from the top left.
            row_weights = [1] * num_rows
            col_weights = [1] * num_columns
            if orientation == 'auto':
                orientation = 'horizontal'
        TiledChart.__init__(
            self, backend, col_weights, row_weights, orientation, 
            **padding)


            
    
        

_chart_types = {
    'square'    : SquareChart,
    'tiled_adv' : TiledChart,
    'tiled'     : SimpleTiledChart,
}

def create(name, backend, **kargs):
    return _chart_types[name](backend, **kargs)

def list_chart_types():
    return tuple((k, v) for k, v in _chart_types.items())
            

__all__ = """
SimpleTiledChart
SquareChart
TiledChart
create
list_chart_types
""".split()
