#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
from StringIO import StringIO
import math
import csv
import types

import dataparse
import datautils
import itertools

from rayon import RayonException


# Exceptions

class RayonDataException(RayonException):
    """
    Base class for exceptions raised in the :mod:`rayon.data` module.
    """
    pass

class RayonDataParseException(RayonDataException):
    """
    Indicates that Rayon had trouble parsing input data from a stream,
    string or file.
    """
    pass





class _ImmutableCollection(object):
    # Implements a collection interface that can't be
    # modified. Subclasses that don't want users to be able to add
    # additional attributes to the object should call lock() as well
    # (presumably after they've set themselves up first).
    def __init__(self):
        self._locked = False
    def __setitem__(self, key, value):
        raise TypeError("'%s' object does not support item assignment" %
                        type(self).__name__)
    def __delitem__(self, ky):
        raise TypeError("'%s' object does not support item deletion" %
                        type(self).__name__)
    def lock(self):
        self._locked = True
    def __setattr__(self, k, v):
        if hasattr(self, "_locked"):
            if self._locked:
                raise TypeError("'%s' object does not support item assignment" %
                                type(self).__name__)
            else:
                return object.__setattr__(self, k, v)
        else:
            # Edge case: caller was lock()
            return object.__setattr__(self, k, v)



class Row(_ImmutableCollection):
    """
    """
    def __init__(self, index, columns, colnames=None):
        _ImmutableCollection.__init__(self)
        # I don't normally like this convention, but it helps avoid
        # conflicts with names in __dict__, so....
        if colnames is not None:
            self.__dict__ = dict((name, columns[i][index])
                                 for i,name in enumerate(colnames))
        self.__index = index
        self.__columns = columns
        self.lock()
    def __len__(self): return len(self.__columns)
    def __getitem__(self, key):
        if type(key) == str:
            return self.__dict__[key]
        else:
            return self.__columns[key][self.__index]
    def __contains__(self, item):
        for i in xrange(len(self.__columns)):
            if self.__columns[i][self.__index] == item:
                return True
        return False
    def __iter__(self):
        return (self.__columns[i][self.__index]
                for i in xrange(len(self.__columns)))
        
    def pprint_asarray(self):
        return "[%s]" % ", ".join("%s" % i for i in self)


class Column(_ImmutableCollection):
    # Documented in classref-data.rst
    
    def __init__(self, data):
        _ImmutableCollection.__init__(self)

        # Rationalize data to a tuple. This should cover iterators as
        # well as lists and other tuples.
        data = tuple(i for i in data)
        
        # Test data to ensure it supports the required statistical
        # functions:
        try:
            if len(data) > 0:
                x = data[0]
                math.floor(x)
                x ** 2
            # Functions should cope with empty datasets
            self.__is_numeric = True
        except:
            self.__is_numeric = False

        self.data = data
        self.stats = None
        self.__mean = None
        self.__population_variance = None
        self.__population_stdev  = None
        self.__sample_variance = None
        self.__sample_stdev  = None
        self.__unique_values_unsorted = None
        self.__unique_values_sorted = None
        # Not locking the ImmutableCollection because (1) I mutate
        # attributes internally, and (2) all the column data is
        # accessed by index, not attribute. They can add all the
        # properties they want, but they won't be useful.
        #
        # TODO: maybe ImmutableCollection should only guard
        # whitelisted items (like the contents of colnames) from being
        # set. That would probably simplify things.
    def expand(self):
        # Returns this column as a one-element list. The additional list
        # may seem redundant, but this format is consistent with
        # :class:`MultiColumn`, in which the outer list is necessary.
        return [self]
    def sorted_by(self, sorted_indices):
        # Return a new :ry:factory:`Column` with the same data as this one,
        # sorted according to *sorted_indices*.
        #
        # This method is used internally in Rayon, but it is not part
        # of the public interface, and should not be documented.
        new_data = [self.data[x] for x in sorted_indices]
        return Column(new_data)
    # TODO: statistics() now deprecated. remove references to this
    # method from demo code
    def statistics(self):
        return self._statistics()
    def _statistics(self):
        if self.stats is None:
            self.stats = datautils.Statistics(self.data,
                                              numeric=self.__is_numeric)
        return self.stats
    def min(self):
        """
        Returns the smallest element in the column data.

        If the data in this column is not numeric, this method will
        raise an error.
        """
        return self._statistics().min()
    def max(self):
        """
        Returns the largest element in the column data.
        
        If the data in this column is not numeric, this method will
        raise an error.
        """
        return self._statistics().max()
    def percentile(self, p):
        """
        Returns the value of the *p* th percentile element in the
        column data. *p* should be an integer between 1 and 100.

        If the data in this column is not numeric, this method will
        raise an error.
        """
        return self._statistics().percentile(p)
    def mean(self):
        """
        Returns the mean of the data in the column.

        If the data in this column is not numeric, this method will
        raise an error.
        """
        if not self.__is_numeric:
            raise RayonDataException("Can't call mean "
                                     "with non-numeric data")
        if self.__mean is None:
            self.__mean = sum(self.data) / len(self.data)
        return self.__mean

    def variance(self):
        """
        Returns the population variance of the data in the column
        data. (The data is assumed to represent the total population.)

        If the data in this column is not numeric, this method will
        raise an error.
        """
        if not self.__is_numeric:
            raise RayonDataException("Can't call variance "
                                     "with non-numeric data")
        if self.__population_variance is None:
            self.__population_variance = (sum((x - self.mean())**2
                                              for x in self.data) /
                                          len(self.data))
        return self.__population_variance
    def stdev(self):
        """
        Returns the population standard deviation of the column
        data. (The data is assumed to represent the total population.)

        If the data in this column is not numeric, this method will
        raise an error.
        """
        if not self.__is_numeric:
            raise RayonDataException("Can't call stdev "
                                     "with non-numeric data")
        if self.__population_stdev is None:
            self.__population_stdev = math.sqrt(self.variance())
        return self.__population_stdev

    def sample_variance(self):
        """
        Returns the sample variance of the data. (The data is
        assumed to represent a sample of a larger population.)

        If the data in this column is not numeric, this method will
        raise an error.
        """
        if not self.__is_numeric:
            raise RayonDataException("Can't call sample_variance "
                                     "with non-numeric data")
        if self.__sample_variance is None:
            self.__sample_variance = (sum((x - self.mean())**2
                                          for x in self.data) /
                                      (len(self.data) - 1))
        return self.__sample_variance

    def sample_stdev(self):
        """
        Returns the sample standard deviation of the column
        data. (The data is assumed to represent a sample of a larger
        population.)

        If the data in this column is not numeric, this method will
        raise an error.
        """
        if not self.__is_numeric:
            raise RayonDataException("Can't call sample_stdev "
                                     "with non-numeric data")
        if self.__sample_stdev is None:
            self.__sample_stdev = math.sqrt(self.sample_variance())
        return self.__sample_stdev


    def uniq(self, return_sorted=False):
        """
        Returns a list of all unique values in column. If
        *return_sorted* is ``True``, return the items in sorted
        order. Otherwise, the are returned in the order in which they
        appear in the column.
        """
        def uniq_unsorted():
            if self.__unique_values_unsorted is None:
                sentinel = set()
                unique_values = []
                for i in self.data:
                    if i not in sentinel:
                        unique_values.append(i)
                        sentinel.add(i)
                self.__unique_values_unsorted = unique_values
            return self.__unique_values_unsorted

        def uniq_sorted():
            if self.__unique_values_sorted is None:
                self.__unique_values_sorted = sorted(uniq_unsorted())
            return self.__unique_values_sorted

        
        if return_sorted:
            return uniq_sorted()
        else:
            return uniq_unsorted()

        
    def __len__(self): return len(self.data)
    def __getitem__(self, key): return self.data[key]
    def __contains__(self, item): return item in self.data
    def __iter__(self): return self.data.__iter__()
    def __eq__(self, other):
        try:
            if len(self) != len(other):
                return False
            for a, b in itertools.izip(iter(self), iter(other)):
                if a != b:
                    return False
            return True
        except:
            # If there are any problems in any of this, they're
            # probably due to other not being a Column, which means
            # it's != self
            return False

    def __ne__(self, other):
        return not (self == other)
            
            


class ConstantColumn(Column):
    """
    A synthetic column that behaves exactly like a regular
    :ry:factory:`Column` object. However, every item in the column contains
    the same value.

    *val* is the value of each element the :class:`ConstantColumn`
    object; *length* is the length of the :class:`ConstantColumn`
    object.
    
    """
    def __init__(self, val, length):
        # not calling Column.__init__ on purpose TODO: this probably
        # means I should have a ColumnImplementation class or
        # something that gets the Column.__init__ stuff out of my
        # way....
        if length <= 0:
            raise RayonDataException("Length of ConstantColumn must be > 0")
        self.val = val
        self.length = length
        self.stats = None
    # expand() is inherited from Column
    def sorted(self, sorted_indices):
        return Column(self, self.val, self.length)
    def statistics(self):
        if self.stats is None:
            return datautils.Statistics([self.val], presorted=True)
    def min(self): return self.val
    def max(self): return self.val
    def mean(self): return self.val
    def percentile(self, p): return self.val
    def sample_stdev(self): return 0
    def sample_variance(self): return 0
    def population_stdev(self): return 0
    def population_variance(self): return 0
    def uniq(self, return_sorted=False): return [self.val]
    def __len__(self): return self.length
    def __getitem__(self, key):
        if key > self.length - 1 or abs(key) > self.length:
            raise IndexError
        else:
            return self.val
    def __contains__(self, item): return item == self.val
    def __iter__(self): return (self.val for _ in xrange(self.length))


class MultiColumn(Column):
    # A column composed of multiple sub-columns of data that should be
    # treated as one column (e.g., a composite key). This class is
    # returned as the result of a :meth:`Dataset.meld` operation.

    # :class:`MultiColumn` objects work exactly like :ry:factory:`Column`
    # objects, except that their elements are tuples containing the
    # corresponding values at that position in the melded columns.
    def expand(self):
        # Return a list of :ry:factory:`Column` objects representing the
        # subcolumns.
        return [Column(x) for x in zip(*self.data)]
        

class Dataset(object):
    # Documented in classref-data.rst

    @classmethod
    def from_stream(cls, stream, colnames=None, typemap=None, delimiter="|"):
        """
        Creates a dataset from data read from a stream. The format of
        the streamed data is described in :ref:`On-disk format
        <data-on-disk-format>`.

        *stream* is a file-like object that supports reading. The
        caller is responsible for closing the stream.

        If supplied, *colnames* is a sequence of strings containing
        column names for the data. There must be as many names in the
        sequence as there are columns in the data. If not specified,
        column names from the file will be used; if the file contains
        no column names, the :ry:factory:`Dataset` will be created without
        column names.

        If supplied, *typemap* is a list of functions that take a
        string representation of an object and return the
        object. There must be as many functions in the sequence as
        there are columns in the data. If not specified, strings
        composed entirely of numbers (optionally with a single period)
        will be converted to numbers, and everything else will be left
        as strings.

        If supplied, *delimiter* is the character used in the file to
        delimit fields in a record. The default delimiter is the pipe
        (``|``) character. The delimiter may also be specified in the
        file format; in that case, this argument is ignored.
        """
        try:
            parser  = dataparse.from_stream(
                stream, colnames, typemap, delimiter)
        except Exception, e:
            raise RayonDataParseException("Couldn't parse input", e)
        coldata = [[] for _ in xrange(parser.numcols())]
        try:
            for row in parser.row_iter():
                for idx, el in enumerate(row):
                    coldata[idx].append(el)
        except dataparse.TypeConversionException, e:
            # Reformat and reraise
            try:
                fn_name, input_val, cause = e.args
                if cause is None:
                    msg = "Error converting '%s' using '%s'" % (
                        str(input_val), fn_name)
                else:
                    msg = "Error converting '%s' using '%s': %s" % (
                        str(input_val), fn_name, str(cause))
            except:
                msg = "Type conversion error"

            raise RayonDataParseException(msg)
        return cls([Column(c) for c in coldata],
                   parser.colnames(), parser.header())



    
    @classmethod
    def from_file(cls, fname, colnames=None, typemap=None, delimiter="|"):
        """
        Creates a dataset from a file. The format of the file is
        described in the documentation for :ref:`On-disk format
        <data-on-disk-format>`.

        The *colnames*, *typemap* and *delimiter* arguments are
        described in :meth:`from_stream`.

        This method is equivalent to calling
        ``Dataset.from_stream(open(fname, 'r'), ...)``
        """
        f = open(fname, 'r')
        try:
            return cls.from_stream(f, colnames, typemap, delimiter)
        finally:
            f.close()


    @classmethod
    def from_string(cls, string, colnames=None, typemap=None, delimiter="|"):
        """
        Creates a dataset from a data contained in *string*. The
        format of the string is described in the documentation for
        :ref:`On-disk format <data-on-disk-format>`.

        The *colnames*, *typemap* and *delimiter* arguments are
        described in :meth:`from_stream`.
        
        Note that the string data may contain headers specifying the
        delimiter and column names, which may be overridden by
        *colnames* and *delimiter* arguments , just as they are in
        :meth:`from_file` and :meth:`from_stream`
        """
        strstream = StringIO(string)
        try:
            return cls.from_stream(strstream, colnames, typemap, delimiter)
        finally:
            strstream.close()


            
    @classmethod
    def from_csv_stream(cls, stream, colnames=None, typemap=None, **kwargs):
        """
        Creates a dataset from a stream containing data in a CSV
        format. *stream* is a file-like object that supports
        reading.

        The *colnames* and *typemap* arguments are described in
        :meth:`from_stream`. Any remaining keyword arguments are
        passed to the constructor for :meth:`csv.Dialect`. If none are
        provided, the nature of the csv file is determined via
        inspection.
        """
        if kwargs:
            dialect = csv.Dialect(**kwargs)
        else:
            dialect = None
        reader, cnames = dataparse.from_csv_stream(stream, colnames,
                typemap=typemap, dialect=dialect)
        return cls.from_rows(reader, colnames=cnames)



    @classmethod
    def from_csv_file(cls, fname, colnames=None, typemap=None, **kwargs):
        """
        Creates a dataset from a file containing data in a CSV format.

        The *colnames* and *typemap* arguments are described in
        :meth:`from_stream`. Additional keyword arguments are
        described in :meth:`from_csv_stream`.

        This method is equivalent to calling
        ``Dataset.from_csv_stream(open(fname, 'r'), ...)``
        """
        f = open(fname, 'r')
        try:
            return cls.from_csv_stream(f, colnames, typemap, delimiter)
        finally:
            f.close()

        
    @classmethod
    def from_csv_string(cls, string, colnames=None, typemap=None, **kwargs):
        """
        Creates a dataset from data contained in *string* in CSV format.

        The *colnames* and *typemap* arguments are described in
        :meth:`from_stream`.  Any remaining keyword arguments are
        passed to the constructor for :meth:`csv.Dialect`. If none are
        provided, the nature of the csv file is determined via
        inspection.

        Note that the string data may specify column names, which may
        be overridden by *colnames*, just as it is in
        :meth:`csv_from_file` and :meth:`csv_from_stream`
        """
        strstream = StringIO(sring)
        try:
            cls.from_csv_stream(ststream, colnames, typemap, **kwargs)
        finally:
            strstream.close()
    

    @classmethod
    def from_rows(cls, row_data, colnames=None, header=None, sortkey=None):
        """
        Creates a dataset from an iterable (list, tuple, iterator,
        etc.)  of rows in a dataset.

        *row_data* is an iterable containing rows of data. The rows
        may be lists, tuples, iterators or dictionaries. Every element
        in *row_data* must be of equal length. If *row_data* contains
        dictionaries, they must additionally all have the same keys.

        If specified, *colnames* is a list of strings . Each item
        represents the name of the corresponding column in the
        dataset.

        If *row_data* contains dictionaries and *colnames* is
        supplied, every element in *colnames* must be a key in the
        dictionary contents of *row_data*. If *colnames* is not
        supplied, the order of columns in the dataset will be the
        alphanumeric sort order of the keys of the dictionaries in
        *row_data*.
        
        If specified, *header* is a dictionary whose keys and values
        are both strings containing mostly arbitrary metadata about
        the dataset. Some header names, however, are reserved; see
        :ref:`Special Headers <data-special-headers>`.

        If specified, *sortkey* is a function that takes a
        :ref:`row <ref-row-objects>` object and returns a comparable. (See the
        :meth:`sort` method for a more precise definition.) This
        function will be used to determine the current sort order of
        the dataset. If not supplied, the current sort order will be
        insert order.
        """
        # De-iteratorize
        try:
            row_data = tuple(x for x in row_data)
        except TypeError:
            raise RayonDataException("iterable object expected")

        if len(row_data) == 0:
            # Degenerate case
            if colnames is None:
                # Really degenerate case
                return cls([Column([])], colnames, header, sortkey)
            else:
                return cls([Column([]) for _ in colnames], colnames,
                           header, sortkey)
        else:
            # Handle dictionaries
            if isinstance(row_data[0], dict):
                if colnames is None:
                    col_order = sorted(row_data[0].keys())
                else:
                    col_order = colnames
                new_columns = [[r[name] for r in row_data]
                               for name in col_order]
                return cls.from_columns(new_columns, colnames, header, sortkey)

            # Handle lists, tuples, iterators
            row_data = tuple(tuple(x) for x in row_data)
            return cls.from_columns(zip(*row_data), colnames, header, sortkey)

    
    @classmethod
    def from_row_dictionaries(cls, row_data, colnames=None, header=None,
                              sortkey=None):
        """
        Creates a dataset from *row_data*, a list of dictionaries
        representing rows in a dataset. Each dictionary should have
        the same keys.

        If *colnames* is not supplied, the order of the columns is the
        alphanumeric sort order of the row keys. Otherwise, column
        order is taken from the ordering of names in
        *colnames*. Otherwise, the *colnames*, *header* and *sortkey*
        arguments are as described in :ry:factory:`Dataset`.
        """
        # Deprecated. Use from_rows.
        # TODO: Remove in Rayon 2.0
        if colnames is None:
            col_order = sorted(row_data[0].keys())
        else:
            col_order = colnames
        new_columns = [[r[name] for r in row_data] for name in col_order]
        return cls.from_columns(new_columns, colnames, header, sortkey)
            

    @classmethod
    def from_columns(cls, col_data, colnames=None, header=None, sortkey=None):
        """
        Creates a dataset from an iterable (list, tuple, iterator,
        etc.) of columns in a dataset.

        *col_data* can be a list of lists, iterables or column
        objects. All elements in  *col_data*
        must be of equal length.

        If specified, *colnames* is a list of strings . Each item
        represents the name of the corresponding column in the
        dataset.

        If specified, *header* is a dictionary whose keys and values
        are both strings containing mostly arbitrary metadata about
        the dataset. Some header names, however, are reserved; see
        :ref:`Special Headers <data-special-headers>`.

        If specified, *sortkey* is a function that takes a
        :ref:`row <ref-row-objects>` object and returns a comparable. (See the
        :meth:`sort` method for a more precise definition.) This
        function will be used to determine the current sort order of
        the dataset. If not supplied, the current sort order will be
        insert order.
        """
        # De-iteratorize
        try:
            col_data = tuple(x for x in col_data)
        except TypeError:
            raise RayonDataException("iterable object expected")

        if len(col_data) == 0:
            # Degenerate case
            if colnames is None:
                # Really degenerate case
                return cls([Column([])], colnames, header, sortkey)
            else:
                return cls([Column([]) for _ in colnames], colnames,
                           header, sortkey)
        else:
            if isinstance(col_data[0], Column):
                return cls(col_data, colnames, header, sortkey)
            else:
                return cls(
                    [Column(c) for c in col_data], colnames, header, sortkey)

    @classmethod
    def from_column_objects(cls, cols, colnames=None,
                            header=None, sortkey=None):
        """
        Creates a dataset from a list :ry:factory:`Column` objects. 

        *colnames*, *header* and *sortkey* are described
        in :ry:factory:`Dataset`.

        This method is equivalent to calling the :ry:factory:`Dataset`
        constructor.
        """
        # Deprecated. Use from_columns instead
        # TODO: remove in Rayon 2.0
        return cls(cols, colnames, header, sortkey)


    def _bootstrap_column_names(self, columns, colnames):
        colnames = list(colnames)
        if len(columns) != len(colnames):
            raise RayonDataException(
                "%d column names mismatches %d columns" %
                (len(colnames), len(columns)))
        self.colindex_by_name = dict(
            (name, index) for index, name in enumerate(colnames))
        
    def __init__(self, columns=[], colnames=None, header=None, sortkey=None):
        # See class definition docstring for constructor documentation

        # Supply defaults if missing
        if header is None:
            self.header_lines = {}
        else:
            self.header_lines = header
        if colnames is None:
            self.colindex_by_name = {}
        else:
            self._bootstrap_column_names(columns, colnames)
        # The data
        self.columns = columns

        # Sort if a sort key was supplied (will do nothing if sortkey
        # is None)
        self.sort(sortkey)


    def get_column_names(self):
        """
        Returns a list of names of all columns in the dataset, in order
        from lowest index to highest. If no column names are
        specified, this method returns an empty list.
        """
        return [k for k, v in sorted(self.colindex_by_name.items(),
                                     key=lambda a: a[1])]
                
    

    def set_column_names(self, colnames):
        """
        Replaces the :ry:factory:`Dataset` object's column names (if they
        exist) with *colnames*.

        *colnames* is an iterable of strings. The number of elements
        in *colnames* should correspond to the number of columns in
        the dataset.
        """
        self._bootstrap_column_names(self.columns, colnames)

    def get_column_name_from_index(self, idx):
        """
        Returns the name corresponding to a column referenced by
        *idx*. *idx* can be a name, in which case it is simply
        returned if it refers to a column in this
        dataset. (Consequently, this method can be used to convert a
        value that may be a name or an index to a name.)
        """
        if type(idx) == type(""):
            if not self.colindex_by_name.has_key(idx):
                raise RayonDataException("Invalid column specification '%s'" %
                                         idx)
            return idx
        else:
            try:
                return self.get_column_names()[idx]
            except IndexError:
                raise RayonDataException("Invalid column specification '%s'" %
                                         idx)

    def get_column_index_from_name(self, name):
        """
        Returns the numeric index corresponding to a column referenced
        by *name*. *name* can be numeric index, in which case it is
        simply returned if it refers to an index in this
        dataset. (Consequently, this method can be used to convert a
        value that may be a name or an index to an index.)
        """
        if type(name) == type(""):
            try:
                return self.colindex_by_name[name]
            except KeyError:
                raise RayonDataException("Invalid column specification '%s'" %
                                         name)
        else:
            try:
                self.get_column_names()[name]
            except:
                raise RayonDataException("Invalid column specification '%s'" %
                                         name)
            return name
            
        

    # XXX TODO Deprecate  and remove 
    def column_names(self):
        return self.get_column_names()

    
    # Data "shape"
    # XXX TODO Deprecate these and remove them
    def numrows(self):
        return self.get_num_rows()
    def numcols(self):
        return self.get_num_cols()


    # Data "shape" -- new and improved interface (names are more
    # consistent with other NetSA python code)
    def get_num_rows(self):
        """
        Returns the number of rows in the dataset.
        """
        try:
            return len(self.columns[0])
        except IndexError:
            return 0

    def get_num_columns(self):
        """
        Returns the number of columns in the dataset
        """
        return len(self.columns)
        
    def get_num_cols(self):
        """
        Returns the number of columns in the dataset
        """
        # Deprecated. Use get_num_columns instead.
        # TODO: remove for Rayon 2.0
        return len(self.columns)
    
    # Headers
    # XXX TODO Deprecate these and remove them
    def header(self, k, v=None):
        if v is  None:
            return self.get_header(k)
        else:
            return self.set_header(k, v)
    def headers(self):
        return self.get_headers()
    def header_names(self):
        return self.get_header_names()

    # Headers -- new and improved interface (names are more consistent
    # with other NetSA python code)
    def get_header(self, name):
        """
        Gets the value of the header named *name*
        """
        return self.header_lines.get(name, None)
    def set_header(self, name, value):
        """
        Sets the value of the header named *name* to *value*
        """
        self.header_lines[name] = value
    def del_header(self, name):
        """
        Remove the header named *name* from the list of headers. If
        *name* does not exist, this method does nothing.
        """
        try:
            del self.header_lines[k]
        except KeyError:
            # Ignore key errors for removal
            pass
    def get_headers(self):
        """
        Returns an iterator of 2-tuples over the headers of this
        dataset. Each 2-tuple is a *(name, value)* pair representing
        one header. Order is undefined.
        """
        return ((k, v) for k, v in self.header_lines.items())
    def get_header_names(self):
        """
        Returns an iterator of strings over the names of the headers in
        this dataset. Order is undefined.
        """
        return (k for k in self.header_lines.keys())
        
        

    # Data access
    def column(self, colname_or_index, ignore_sort=False):
        """
        A synonym for :meth:`get_column`.
        """
        return self.get_column(colname_or_index, ignore_sort)
    
        
    def get_column(self, colname_or_index, ignore_sort=False):
        """
        Given a column index or name, returns the corresponding
        :ry:factory:`Column` object.

        *colname_or_index* is either the index (an integer) or name
        (a string) referring to the column.
        
        If *ignore_sort* is ``False``, return the column in the order
        of the current sort. Otherwise, the column will be sorted in
        insert order, even if the dataset is currently sorted by some
        key.
        """
        if not isinstance(colname_or_index, int):
            try:
                colindex = self.colindex_by_name[colname_or_index]
            except KeyError:
                raise RayonDataException("Invalid column specification '%s'" %
                                         colname_or_index)
        else:
            colindex = colname_or_index

        try:
            if self.sort_key is None or ignore_sort:
                return self.columns[colindex]
            else:
                return self.columns[colindex].sorted_by(self.sorted_indices)
        except IndexError:
            raise RayonDataException("Invalid column specification '%s'" %
                                     colname_or_index)
        
        

    def row(self, index, ignore_sort=False):
        """
        A synonym for :meth:`get_row`
        """
        return self.get_row(index, ignore_sort)

    def get_row(self, index, ignore_sort=False):
        """
        Returns the row in the dataset corresponding to
        *index*. Negative numbers index from the end of the dataset,
        as in Python lists.

        If *ignore_sort* is ``False``, return the row corresponding to
        *index* according to the dataset's current sort order. If
        *ignore_sort* is ``True``, return the row corresponding to
        *index* according to the dataset's native sort order.
        """
        if not ignore_sort:
            index = self.sorted_indices[index]
        return Row(index, self.columns, self.column_names() or None)

    def iter_ignore_sorted(self):
        """
        Iterate over the rows in the dataset according to the
        dataset's native sort order, regardless of the current sort
        order. (To get an iterator over the rows of the dataset in the
        current sort order, use Python's :func:`iter` function.)
        """
        for i in xrange(self.numrows()):
            yield self.get_row(i, True)

    def iter_columns(self):
        """
        Iterate over the columns in the dataset, in order from the
        lowest index to the highest.
        """
        return iter(self.columns)

    def __iter__(self):
        for i in xrange(self.numrows()):
            yield self.get_row(i)


    # Adding or removing columns
    def append_column(self, new_col, name=None):
        """
        Append *new_col* to this object. *new_col* must be either a
        :ry:factory:`Column` object or an iterable sequence (list, tuple,
        iterator, etc.).

        If this :ry:factory:`~rayon.data.Dataset` object already contains columns,
        *new_col* must have the same number of elements as the
        existing columns. *new_col* is appended to the \"right\" of
        the other columns, e.g.::

            >>> # t is a Toolbox object
            >>> raw = [[1,2,3], ['a', 'b', 'c'], [3,2,1]]
            >>> d = t.new_dataset_from_columns(raw)
            >>> d.append_column([4,5,6])
            >>> d.get_num_cols()
            4
            >>> [i for i in d.get_column(3)]
            [4, 5, 6]

        It is an error to append a column to a :ry:factory:`Dataset` object
        without a column name if the :ry:factory:`Dataset` object already
        contains column names::

            >>> # t is a Toolbox object
            >>> raw = [[1,2,3], ['a', 'b', 'c'], [3,2,1]]
            >>> colnames = ['foo', 'bar', 'baz']
            >>> d = t.new_dataset_from_columns(raw, colnames=colnames)
            >>> d.append_column([4,5,6])
            RayonDataException: must specify a column name

        The column will be inserted in native sort order, regardless
        of the current sort order::

            >>> # t is a Toolbox object
            >>> raw = [[1,2,3], ['a', 'b', 'c'], [3,2,1]]
            >>> d = t.new_dataset_from_columns(raw)
            >>> d.sort(lambda r: r[2])
            >>> d.append_column([4,5,6])
            >>> d.sort()
            >>> [[i for i in r] for r in d]
            [[3, 'c', 1, 6], [2, 'b', 2, 5], [1, 'a', 3, 4]]

        To append a column to a sorted :ry:factory:`Dataset` object and
        have it appear in sorted order, first clone the
        :ry:factory:`Dataset` object using :meth:`filter`, then append the
        column to the new :ry:factory:`Dataset`::

            >>> from rayon.data import *
            >>> raw = [[1,2,3], ['a', 'b', 'c'], [3,2,1]]
            >>> d = Dataset.from_columns(raw)
            >>> d.sort(lambda r: r[2])
            >>> d2 = d.filter(True)
            >>> d2.append_column(Column[4,5,6])
            >>> [[i for i in r] for r in d2]
            [[3, 'c', 1, 4], [2, 'b', 2, 5], [1, 'a', 3, 6]]
        """
        if not isinstance(new_col, Column):
            try:
                new_col = Column(tuple(new_col))
            except TypeError:
                raise RayonDataException(
                    "can only append iterable object to Dataset")

        if len(self.get_column_names()) != 0 and name is None:
            raise RayonDataException("must specify a column name")

        if (self.get_num_cols() != 0
            and len(new_col) != self.get_num_rows()):
            raise RayonDataException(
                "appended columns must have the same length as "
                "existing columns")
        
        colindex = self.get_num_cols()
        self.columns.append(new_col)
        if len(self.get_column_names()) != 0:
            self.colindex_by_name[name] = colindex

        return
            
    # Mapping
    def copy_header(self):
        # Utility function for map/filter functions
        if len(self.header_lines) == 0:
            return None
        else:
            return dict(self.header_lines)
    def copy_column_names(self):
        # Utility function for map/filter functions
        colnames = self.column_names()
        if len(colnames) == 0:
            return None
        else:
            return colnames

    
    def map(self, fn, copy_colnames=False, colnames=None, iter_sorted=True,
            copy_headers=False):
        """
        Produces a new :ry:factory:`Dataset` object by iterating over the
        rows of this :ry:factory:`Dataset` object, calling *fn* on each
        row. *fn* takes a :ref:`row <ref-row-objects>` object and returns an array of
        values corresponding to a record in the new dataset.

        If specified, *colnames* will be used to specify column names
        in the new dataset. If *copy_colnames* is ``True`` and
        *colnames* is not specified, this dataset's column names will
        be used in the generated dataset. (This will generate an error
        if the number of columns in the resulting set differ from the
        number of columns in this one.

        If *copy_headers* is ``True``, this dataset's headers will be
        used as the headers of the result set. By default, the dataset
        will be created with empty headers.

        By default, this function will iterate over the dataset in its
        current sort order; to iterate in native sort order, set
        *iter_sorted* to ``False``.
        """
        if colnames is None:
            if copy_colnames:
                colnames = self.copy_column_names()

        if copy_headers:
            header = self.copy_header()
        else:
            header = None
            

                
        if iter_sorted:
            iterator = iter(self)
        else:
            iterator = self.iter_ignore_sorted()
        
        new_cols = tuple(Column(c) for c in
                         zip(*[fn(r) for r in iterator]))
        

        return Dataset(columns=new_cols, colnames=colnames, header=header)
        
        
    
    # Filtering
    def copy_header(self):
        # Utility function for filter functions
        if len(self.header_lines) == 0:
            return None
        else:
            return dict(self.header_lines)
    def copy_column_names(self):
        # Utility function for filter functions
        colnames = self.column_names()
        if len(colnames) == 0:
            return None
        else:
            return colnames

    def filter(self, filter_fn, keep_pass=True, insert_sorted=False,
               copy_sort_key=True, copy_header=True, copy_colnames=True,
               limit=None):
        """
        Returns a new :ry:factory:`Dataset` object based on this one,
        including either all the rows passing the filter function or
        all the rows failing the filter function.

        *filter_fn* is a function taking a :ref:`row <ref-row-objects>` object and
        returning ``True`` or ``False``; *filter_fn* may also be one
        of the values ``True`` or ``False``, meaning to pass
        (``True``) or fail (``False``) every row in the dataset.

        If *keep_pass* is ``True``, then if *filter_fn* returns (or
        is) ``True`` for a given row, the row will be included in the
        filtered dataset. If *keep_pass* is ``False``, the logic is
        inverted, and only rows for which *filter_fn* returns (or is)
        ``False`` will be included in the filtered dataset.

        If *insert_sorted* is ``True``, the native sort order of the
        filtered dataset will be the parent dataset's current sort
        order. If *insert_sorted* is ``False``, the native sort order
        of the filtered dataset will be the parent dataset's native
        sort order, regardless of the current sort order.

        If *copy_sort_key* is ``True``, the current sort order from
        the parent dataset will be copied into the filtered
        dataset. This does not change the native sort order.

        If *copy_header* is ``True``, the parent dataset's headers
        will be copied into the output dataset. Otherwise, the output
        dataset will have no headers.

        If *copy_colnames* is ``True``, the parent dataset's column
        names will be copied into the output dataset. Otherwise, the
        output dataset will have no column names.

        If supplied, *limit* is the maximum number of rows to return
        from this filter. By default, all rows matching *filter_fn*
        will be returned.
        """
        if filter_fn is True:
            filter_fn = lambda row: True
        elif filter_fn is False:
            filter_fn = lambda row: False
        if not keep_pass:
            real_filter_fn = lambda row: not filter_fn(row)
        else:
            real_filter_fn = filter_fn

        if copy_header:     header = self.copy_header()
        else:               header = None
        if copy_colnames:   colnames = self.copy_column_names()
        else:               colnames = None
        if copy_sort_key:   sortkey = self.sort_key
        else:               sortkey = None

        if insert_sorted:
            row_iter = iter(self)
        else:
            row_iter = self.iter_ignore_sorted()

        if limit is None:
            new_columns = zip(*tuple(row for row in row_iter
                                     if real_filter_fn(row)))
        else:
            new_columns = zip(*tuple(row for row in row_iter
                                     if real_filter_fn(row))[:limit])
        
        return type(self).from_columns(new_columns, colnames=colnames,
                                       header=header, sortkey=sortkey)


    def filter_pass(self, filter_fn, insert_sorted=False,
                    copy_sort_key=True, copy_header=True,
                    copy_colnames=True, limit=None):
        """
        Produces a new :ry:factory:`Dataset` based on this one, containing
        only rows passing *filter_fn*.

        *filter_fn* is a function taking a :ref:`row <ref-row-objects>` object and
        returning ``True`` or ``False``; *filter_fn* may also be one
        of the values ``True`` or ``False``, meaning to pass
        (``True``) or fail (``False``) every row in the dataset.

        If *insert_sorted* is ``True``, the native sort order of the
        filtered dataset will be the parent dataset's current sort
        order. If *insert_sorted* is ``False``, the native sort order
        of the filtered dataset will be the parent dataset's native
        sort order, regardless of the current sort order.

        If *copy_sort_key* is ``True``, the current sort order from
        the parent dataset will be copied into the filtered
        dataset. This does not change the native sort order.

        If *copy_header* is ``True``, the parent dataset's headers
        will be copied into the output dataset. Otherwise, the output
        dataset will have no headers.

        If *copy_colnames* is ``True``, the parent dataset's column
        names will be copied into the output dataset. Otherwise, the
        output dataset will have no column names.

        If supplied, *limit* is the maximum number of rows to return
        from this filter. By default, all rows matching *filter_fn*
        will be returned.

        .. seealso::  :meth:`filter_both`, :meth:`filter_fail`
        """
        return self.filter(filter_fn, True, insert_sorted,
                           copy_sort_key, copy_header, copy_colnames, limit)


    def filter_fail(self, filter_fn, insert_sorted=False,
                    copy_sort_key=True, copy_header=True,
                    copy_colnames=True, limit=None):
        """
        Produces a new :ry:factory:`Dataset` based on this one, containing
        only rows failing *filter_fn*.

        *filter_fn* is a function taking a :ref:`row <ref-row-objects>` object and
        returning ``True`` or ``False``; *filter_fn* may also be one
        of the values ``True`` or ``False``, meaning to pass
        (``True``) or fail (``False``) every row in the dataset.

        If *insert_sorted* is ``True``, the native sort order of the
        filtered dataset will be the parent dataset's current sort
        order. If *insert_sorted* is ``False``, the native sort order
        of the filtered dataset will be the parent dataset's native
        sort order, regardless of the current sort order.

        If *copy_sort_key* is ``True``, the current sort order from
        the parent dataset will be copied into the filtered
        dataset. This does not change the native sort order.

        If *copy_header* is ``True``, the parent dataset's headers
        will be copied into the output dataset. Otherwise, the output
        dataset will have no headers.

        If *copy_colnames* is ``True``, the parent dataset's column
        names will be copied into the output dataset. Otherwise, the
        output dataset will have no column names.

        If supplied, *limit* is the maximum number of rows to return
        from this filter. By default, all rows matching *filter_fn*
        will be returned.

        .. seealso::  :meth:`filter_both`, :meth:`filter_pass`
        """
        return self.filter(filter_fn, False, insert_sorted,
                           copy_sort_key, copy_header, copy_colnames, limit)


    def filter_both(self, filter_fn, insert_sorted=False,
                    copy_sort_key=True, copy_header=True,
                    copy_colnames=True, limit=None, fail_limit=None):
        """
        Produces a tuple of two new :ry:factory:`Dataset` objects based on
        this one, with rows passing *filter_fn* in the first, and
        rows failing *filter_fn* in the second. This method is
        equivalent to calling ``a, b = (data.filter_pass(fn, ...),
        data.filter_fail(fn, ...))``, but is more efficient.

        *filter_fn* is a function taking a :ref:`row <ref-row-objects>` object and
        returning ``True`` or ``False``; *filter_fn* may also be one
        of the values ``True`` or ``False``, meaning to pass
        (``True``) or fail (``False``) every row in the dataset.

        If *insert_sorted* is ``True``, the native sort order of the
        filtered datasets will be the parent dataset's current sort
        order. If *insert_sorted* is ``False``, the native sort order
        of the filtered datasets will be the parent dataset's native
        sort order, regardless of the current sort order.

        If *copy_sort_key* is ``True``, the current sort order from
        the parent dataset will be copied into the filtered
        datasets. This does not change the native sort order.

        If *copy_header* is ``True``, the parent dataset's headers
        will be copied into the output datasets. Otherwise, the output
        datasets will have no headers.

        If *copy_colnames* is ``True``, the parent dataset's column
        names will be copied into the output datasets. Otherwise, the
        output datasets will have no column names.

        If supplied, *limit* is the maximum number of passing rows to
        return from this filter. By default, all rows matching
        *filter_fn* will be returned in the first :ry:factory:`Dataset`
        object.

        If supplied, *fail_limit* is the maximum number of failing
        rows to return from this filter. By default, all rows failing
        *filter_fn* will be returned in the second :ry:factory:`Dataset`
        object.

        .. seealso:: :meth:`filter_fail`, :meth:`filter_pass`
        """
        if copy_header:     header = self.copy_header()
        else:               header = None
        if copy_colnames:   colnames = self.copy_column_names()
        else:               colnames = None
        if copy_sort_key:   sortkey = self.sort_key
        else:               sort_key = None

        if insert_sorted:
            row_iter = iter(self)
        else:
            row_iter = self.iter_ignore_sorted()

        pass_records = []
        fail_records = []
        for row in row_iter:
            if filter_fn(row):
                if limit is None or len(pass_records) < limit:
                    pass_records.append(tuple(row))
            else:
                if fail_limit is None or len(fail_records) < fail_limit:
                    fail_records.append(tuple(row))
            # Break out if we've satisfied both limits
            if (limit is not None and len(pass_records) == limit and
                fail_limit is not None and len(fail_records) == fail_limit):
                break

        # Convert pass_records and fail_records to datasets and return
        return tuple(
            type(self).from_rows(rec, colnames=colnames,
                                 header=header, sortkey=sortkey)
            for rec in (pass_records, fail_records))


    # Partitioning
    def partition(self, key_colname_or_index,
                  insert_sorted=False, copy_sort_key=True,
                  copy_header=True, copy_colnames=True):
        """
        Creates several new :ry:factory:`Dataset` objects, partitioned
        along the unique values of one of the parent dataset's
        columns. The new datasets are returned as a tuple of pairs;
        the first pair is the common value in the partitioning column;
        the second is a new dataset containing the matching rows.

        The following examples illustrate the use of :meth:`partition`
        on a dataset *d* with the following rows:

           red   | 1 | 0
           red   | 1 | 1
           green | 1 | 2
           green | 1 | 3
           blue  | 1 | 4
           blue  | 1 | 5

        The following example partitions *d* on the first column,
        producing three datasets (one for each color), each containing
        two rows::
           
           >>> from pprint import pprint
           >>> p0 = d.partition(0)
           >>> pprint(p0)
           (('red', <rayon.data.Dataset object at 0x...>),
            ('green', <rayon.data.Dataset object at 0x...>),
            ('blue', <rayon.data.Dataset object at 0x...>))
           >>> print(p0[0][1].to_string())
           red|1|0
           red|1|1

        Partitioning *d* on the second column produces a single
        dataset containing all the rows::

           >>> p1 = d.partition(1)
           >>> pprint(p1)
           ((1, <rayon.data.Dataset object at 0x10058eb90>),)
           >>> print p1[0][1].to_string()
           red|1|0
           red|1|1
           green|1|2
           green|1|3
           blue|1|4
           blue|1|5

        As each value in the third column is unique, artitioning *d*
        on this column produces produces six datasets of one row
        each::

           >>> p2 = d.partition(2)
           >>> pprint(p2)
           pprint(p2)
           ((0, <rayon.data.Dataset object at 0x...>),
            (1, <rayon.data.Dataset object at 0x...>),
            (2, <rayon.data.Dataset object at 0x...>),
            (3, <rayon.data.Dataset object at 0x...>),
            (4, <rayon.data.Dataset object at 0x...>),
            (5, <rayon.data.Dataset object at 0x...>))
           >>> print p2[0][1].to_string()
           red|1|0
           >>> print p2[1][1].to_string()
           red|1|1
           >>> print p2[2][1].to_string()
           green|1|2
        
        *key_colname_or_index* is the name or index of the column on
        which to partition.
        
        If *copy_sort_key* is ``True``, the current sort order from
        the parent dataset will be copied into the partitioned
        datasets. This does not change the native sort order.

        If *insert_sorted* is ``True``, the rows will be inserted into
        the partitioned datasets in the order indicated by the
        parent dataset's current sort order. If *insert_sorted* is
        ``False``, the rows will be inserted into the partitioned
        datasets according to the parent dataset's native sort
        order, regardless of the current sort order.

        If *copy_header* is ``True``, the parent dataset's headers
        will be copied into the output dataset. Otherwise, the output
        dataset will have no headers.

        If *copy_colnames* is ``True``, the parent dataset's column
        names will be copied into the output dataset. Otherwise, the
        output dataset will have no column names.

        This function returns a list of 2-tuples *(keyval, dataset)*,
        where *keyval* is one of the unique values in the column
        referred to by *key_colname_or_index*, and *dataset* is a
        :ry:factory:`Dataset` object containing the rows in the parent
        dataset that contain that value in *key_colname_or_index*.
        """
        key_col = self.get_column(key_colname_or_index)
        by_keys = key_col.uniq()
        
        if copy_header:     header = self.copy_header()
        else:               header = None
        if copy_colnames:   colnames = self.copy_column_names()
        else:               colnames = None
        if copy_sort_key:   sortkey = self.sort_key
        else:               sortkey = None

        if insert_sorted:
            row_iter = iter(self)
        else:
            row_iter = self.iter_ignore_sorted()

        partition_rows = dict()

        for row in row_iter:
            keyval = row[key_colname_or_index]
            partition_rows.setdefault(keyval, []).append(row)

        return tuple((k, type(self).from_rows(partition_rows[k],
                                              colnames=colnames,
                                              header=header,
                                              sortkey=sortkey))
                     for k in by_keys)

    # Sorting
    def sort(self, key=None):
        """
        Changes the current sort order according to the *key*. If
        *key* is ``None``, revert to native sort order.

        *key* is a function (or callable object) taking a :ref:`row <ref-row-objects>`
        object and returning a comparable value. (That is, a value for
        which the ``<``, ``>``, ``<=``, ``>=`` and ``==`` operators
        have meaning, that can be compared with others of its type to
        determine ordering.) *key* can also be a class whose
        constructor takes a :ref:`row <ref-row-objects>` object and returns an
        instance suitable for comparison. (i.e., definining
        ``__lt__``, ``__gt__``, etc.)

        Some examples::

           # Sorts on the first 2 columns of d
           d.sort(key=lambda r: (r[0], r[1]))
           
           # This is equivalent to the above...
           class Comparator(object):
               def __init__(self, r):
                   self.to_cmp = (r[0], r[1])
               def __lt__(self, other):
                   return self < other
               def __gt__(self, other):
                   return self > other
               def __le__(self, other):
                   return self <= other
               def __ge__(self, other):
                   return self >= other
               def __eq__(self, other):
                   return self == other
               def __ne__(self, other):
                   return self != other

           d.sort(key=Comparator)

           # ...as is this
           def CallableComparator(object):
               def __call__(self, row):
                   return r[0], r[1]

           d.sort(key=CallableComparator())
        """
        # This method is called from __init__, so we need to make sure
        # the bits we need are intialized before we go on.
        def _bootstrap_sort():
            self.sorted_indices = list(xrange(self.numrows()))
            self.sort_key = None
    
        try:
            self.sorted_indices
        except:
            _bootstrap_sort()
        
        if key is None:
            # As it happens, bootstrapping is identical to creating
            # the naive search order, which is what we do when no key
            # is supplied.
            _bootstrap_sort()
        else:
            to_sort = sorted((key(row), i) for i, row in enumerate(self))
            self.sorted_indices = [i for _, i in to_sort]
            self.sort_key = key


    # Writing to file



    def to_file(self, fname, delimiter="|"):
        """
        Writes the contents of this dataset to a file. The file will
        be created if it does not exist, and will be overwritten if it
        does exist. The format of the output is described in
        :ref:`data-on-disk-format`.

        Generally, this method supports writing any opaque datatype
        that can be represented as a string, provided that the
        string representation does not contain the character used as
        the field delimiter.

        At this time, composite columns (such as those returned by the
        :meth:`meld` method) are *not* supported; to write datasets
        with melded columns to disk, first flatten them using the
        :meth:`flatten` method.

        This method is equivalent to calling::

            # d is the dataset
            return d.to_stream(open(fname, 'w'))

        *fname* is the name of the file to write.

        *delimiter* should be a single character that will be used in
        the stream to delimit fields in a record.
        """
        f = open(fname, 'w')
        try:
            return self.to_stream(f, delimiter)
        finally:
            f.close()


    def to_stream(self, stream, delimiter="|"):
        """
        Writes the contents of a dataset to a stream. The format of
        the output is described in :ref:`data-on-disk-format`.

        Generally, this method supports writing any opaque datatype
        that can be represented as a string, provided that the
        string representation does not contain the character used as
        the field delimiter.

        At this time, composite columns (such as those returned by the
        :meth:`meld` method) are *not* supported; to write datasets
        with melded columns to disk, first flatten them using the
        :meth:`flatten` method.

        *stream* is a file-like object that supports writing.

        *delimiter* should be a single character that will be used in
        the stream to delimit fields in a record.
        """
        dataparse.to_stream(stream, self.column_names(), self.header_lines,
                            iter(self), self.numrows(),  delimiter)


    def to_string(self, delimiter="|"):
        """
        Writes the contents of this dataset to a string, then returns
        the string.

        Generally, this method supports writing any opaque datatype
        that can be represented as a string, provided that the
        string representation does not contain the character used as
        the field delimiter.

        At this time, composite columns (such as those returned by the
        :meth:`meld` method) are *not* supported; to write datasets
        with melded columns to disk, first flatten them using the
        :meth:`flatten` method.

        This method is equivalent to calling::

            # d is the dataset
            sio = StringIO()            
            d.to_stream(sio)
            return sio


        *delimiter* should be a single character that will be used in
        the stream to delimit fields in a record.
        """
        out = StringIO()
        try:
            self.to_stream(out, delimiter)
            return out.getvalue()
        finally:
            out.close()

    # Other operations

    def meld(self, *names_or_indices):
        """
        Creates a single composite :ry:factory:`Column` object from the
        multiple columns referenced in *names_or_indices*. This is
        useful, for example, in data where several columns together
        form a unique key.
        
        If *names_or_indices* is not supplied, a :ry:factory:`Column`
        object will be created consisting of all columns in the
        dataset.
        """
        if not len(names_or_indices):
            cnames = self.column_names()
            if len(cnames):
                names_or_indices = self.column_names()
            else:
                names_or_indices = list(xrange(self.numcols()))
        to_meld = tuple(tuple(self.column(n)) for n in names_or_indices)
        return MultiColumn(zip(*to_meld))


    def flatten(self, new_colnames=None):
        """
        Creates a copy of this dataset, converting any composite
        :ry:factory:`Column` objects into one :ry:factory:`Column` object for
        each element in the composite. Note that this may change the
        column indices of subsequent columns.

        If *new_colnames* is supplied, it should be an iterable of
        strings, which will be used as column names for the new
        dataset, with the first new column getting the first name from
        *new_colnames*, and so on. If the number of column names in
        *new_colnames* does not match the number of columns in the
        flattened dataset, a :exc:`RayonDataException` will be
        raised. If *new_colnames* is ``None`` (the default), this
        method will try to use the names of the parent dataset as the
        names of the output dataset. (This will only be possible if
        the :ry:factory:`Dataset` object contains no melded columns.) If
        that isn't possible, the output dataset will not contain
        column names.

        .. seealso:: :meth:`meld`
        """
        # Call expand() on each column to get a list of expanded
        # columns for each real column. Then flatten that list for
        # input into the new dataset
        flattened_columns = \
        [c for col in list(c.expand() for c in self.columns) for c in col]

        # TODO: we should be actually _using_ new_colnames....
        if new_colnames is None:
            if len(flattened_columns) == len(self.get_column_names()):
                new_colnames = self.get_column_names()
            else:
                new_colnames = None
                
        return type(self)(flattened_columns, colnames=new_colnames,
                          header=self.copy_header(), sortkey=self.sort_key)


    # TODO: ecrasons l'infame
    def to_string_as_array_by_rows(self):
        for row in self:
            print row.pprint_asarray()


    # Comparison/Equality

    def __eq__(self, other):
        try:
            if self.get_column_names() != other.get_column_names():
                return False
            for c1, c2 in itertools.izip(self.iter_columns(),
                                         other.iter_columns()):
                if c1 != c2:
                    return False
            return True
        except:
            # As with Column, almost any problem int the try block is
            # going to be due to other not being a Dataset, and hence
            # not equal to self
            return False

    def __ne__(self, other):
        return not (self == other)
