#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Code for serializing and deserializing L{Dataset} objects.
"""
import miscutils, csv
from copy import copy

from rayon import RayonException

class DataParseException(RayonException):
    # Internal base class of exceptions in this module -- should
    # always be caught and handled before reaching a user.
    pass

class TypeConversionException(DataParseException):
    # Problem converting a value to the requested type in the typemap
    pass
    

class PushbackReader(object):
    """Wrapper around a read-only file object. Contains facilities for
    pushing data back onto the stream after it's been read. Silently
    ignores attempts to write to it.

    The PushbackReader keeps one line's worth of data around from the
    previous read. This line may be pushed back into the stream with
    the push_back() method."""
    def __init__(self, stream):
        self.stream = stream
        self.line = None
        self.buffer_is_live = False
    def __iter__(self):
        return self
    def readline(self):
        if self.buffer_is_live:
            self.buffer_is_live = False
            line = self.line
            return line
        else:
            self.line = self.stream.readline()
            return self.line
    def pushback(self):
        self.buffer_is_live = True
    def next(self):
        line = self.readline()
        if line == "":
            raise StopIteration()
        return line



def type_from_string(spec):
    # Generate appropriate typemaps for data with typemaps in
    # headers. Try to do the right thing for converters in as-yet
    # unimported modules.
    if "." in spec:
        pivot = spec.rfind(".")
        mod_name = spec[:pivot]
        fn_name = spec[pivot+1:]
        # fn_name is irrelevant in this call -- it can be any string
        # and still work (but it won't work without the arg -- see
        # __import__ docs)
        mod = __import__(mod_name, globals(), locals(), fn_name)
        return getattr(mod, fn_name)
    else:
        # Assume it's a builtin
        return __builtins__[spec]



class StreamParser(object):
    def __init__(self, colnames, header, typemap, rows, delimiter):
        self.__colnames = colnames
        self.__header = header
        self.__typemap = typemap
        self.__rows = rows
        self.__delimiter = delimiter
        self.__numcols = None
    def colnames(self):
        if self.__colnames is None:
            return None
        # Only return column names we have typeable data for
        return self.__colnames[:self.numcols()]
    def header(self): return self.__header
    def numcols(self):
        if self.__numcols is None:
            line = self.__rows.readline()
            try:
                # Throw away blank lines
                while line.isspace():
                    line = self.__rows.readline()
                items = line.split(self.__delimiter)
                numcols = len(items)
            finally:
                self.__rows.pushback()
            # Take the smallest of the number of typemaps and the
            # number of columns, and ignore the rest. If you have more
            # columns than typemaps, that means data will be ignored.
            self.__numcols = min(numcols, len(self.__typemap))
        return self.__numcols
    def row_iter(self):
        # If we got more typemaps than data columns, ignore the extras
        typemap = self.__typemap[:self.numcols()]
                                             
        for line in self.__rows:
            # Continue to ignore blank lines
            line = line.strip()
            if not line:
                continue
            # Ignore commented-out lines
            if line.startswith("#"):
                continue
            items = line.split(self.__delimiter)[:self.numcols()]
            tmp = []
            def convert(fn, inval):
                try:
                    return fn(inval)
                except ValueError, e:
                    raise TypeConversionException(fn.__name__, inval, e)
            yield [convert(typemap[idx], i.strip())
                   for idx, i in enumerate(items)]

class RowReader(object):
    def __init__(self, colnames, header, typemap, rows, delimiter):
        self.__colnames = colnames
        self.__header = header
        self.__typemap = typemap
        self.__rows = rows
        self.__delimiter = delimiter
        self.__numcols = None
    def colnames(self):
        if self.__colnames is None:
            return None
        # Only return column names we have typeable data for
        return self.__colnames[:self.numcols()]
    def header(self): return self.__header
    def numcols(self):
        if self.__numcols is None:
            line = self.__rows.readline()
            try:
                # Throw away blank lines
                while line.isspace():
                    line = self.__rows.readline()
                items = line.split(self.__delimiter)
                numcols = len(items)
            finally:
                self.__rows.pushback()
            # Take the smallest of the number of typemaps and the
            # number of columns, and ignore the rest. If you have more
            # columns than typemaps, that means data will be ignored.
            self.__numcols = min(numcols, len(self.__typemap))
        return self.__numcols
    def row_iter(self):
        # If we got more typemaps than data columns, ignore the extras
        typemap = self.__typemap[:self.numcols()]
                                             
        for line in self.__rows:
            # Continue to ignore blank lines
            line = line.strip()
            if not line:
                continue
            items = line.split(self.__delimiter)[:self.numcols()]
            tmp = []
            yield [typemap[idx](i.strip()) for idx, i in enumerate(items)]
            
        



def from_stream(stream, orig_colnames, orig_typemap, orig_delimiter):
    """Used by L{Dataset.from_stream} to marshall data into a L{Dataset}."""
    # Parse the prologue
    prologue_lines = []
    stream = PushbackReader(stream)
    has_data = False
    for line in stream:
        if line.isspace():
            continue
        # Header lines start with exactly 2 #. Comment lines start
        # with # and anything else.
        if line.startswith("##"):
            cleanline = line[2:].strip()
            if cleanline.startswith("#"):
                # It's a comment - remove one additional # and store
                cleanline = line[1:].strip()
                if not cleanline.isspace():
                    prologue_lines.append(cleanline)
            else:
                # It's a header line
                prologue_lines.append(cleanline)
        elif line.startswith("#"):
            cleanline = line[1:].strip()
            if not cleanline.isspace():
                prologue_lines.append(cleanline)
        else:
            has_data = True
            stream.pushback()
            break

    # Provisionally grab informal colnames line, if available
    # (It will be overwritten if we find a Column-Names header)
    if len(prologue_lines) and prologue_lines[-1].find(':') == -1:
        col_line = prologue_lines.pop()
    else:
        col_line = None
        


    # The rest of the code needs col_line to be whatever column names
    # there are. Figure it out.


    # Parse header, ignoring comments that may be in there
    def parse_header_line(h):
        k, v = h.split(':', 1)
        v = v.strip()
        return k, v
    header = dict(parse_header_line(h) for h in prologue_lines
                  if ':' in h)

    # Is there a formal column spec? If so, that takes precedence.
    # Then remove it -- user might change column names.
    if header.has_key('Column-names'):
        col_line = header['Column-names']
        del header['Column-names']
        

    # Get delimiter for rows and colnames line. Then remove it so it
    # doesn't get in whe way if we serialize later.
    delim_char = header.get('Delimiter', orig_delimiter)
    try:
        del header['Delimiter']
    except KeyError:
        pass
    

    # Use colnames line unless user has specified column names
    if orig_colnames is not None:
        colnames = orig_colnames
    else:
        # Parse colnames line, if available
        if col_line is None:
            colnames = None
        else:
            colnames = [n.strip() for n in col_line.split(delim_char)]


    # Typemap, if it's in the header and an explicit typemap wasn't
    # passed in (i.e., typemap is None)
    if orig_typemap is None:
        if header.has_key('Typemap'):
            typemap = [type_from_string(s)
                       for s in header['Typemap'].split(delim_char)]
        else:
            # Grab the first row of data (throwing away blank lines as
            # we go), get the typemap from that, and push it back.
            while True:
                line = stream.readline().strip()
                if not line:
                    continue
                else:
                    break
            typemap = [miscutils.infer_type_from_string(item.strip())
                       for item in line.split(delim_char)]
            stream.pushback()
    else:
        typemap = orig_typemap
        if header.has_key('Typemap'):
            del header['Typemap']


    p = StreamParser(colnames, header, typemap, stream, delim_char)
    return p
        




def to_stream(stream, colnames, header, row_iter, numrows, delimiter):
    """Used by L{Dataset.to_stream} to serialize a Dataset to a stream."""
    for k, v in header.items():
        stream.write("## %s: %s\n" % (k, v))
    if len(colnames) > 0:
        stream.write("## Column-names: %s\n" % delimiter.join(colnames))
    for idx, row in enumerate(row_iter):
        stream.write(delimiter.join(str(e) for e in row))
        # Omit the final newline
        if idx != numrows - 1:
            stream.write("\n")



def from_csv_stream(stream, colnames=None, typemap=[], dialect=None):
    """Used by L{Dataset.from_csv_file} to marshall data into a L{Dataset}."""

    pos = stream.tell()
    sample = stream.read(1024)
    stream.seek(pos)

    if not sample:
        return stream, colnames

    wilbur = csv.Sniffer()
    wilbur.preferred.append('|')

    if dialect is None:
        dialect = wilbur.sniff(sample)

    has_colnames = wilbur.has_header(sample)

    stream = PushbackReader(stream)
    reader = csv.reader(stream, dialect)
            
    if has_colnames:
        try:
            cnames = []
            for i,x in enumerate(reader.next()):
                x = x.strip()
                cnames.append(x or i)
        except StopIteration:
            pass
    else:
        cnames = None

    if colnames:
        if len(colnames) != len(cnames):
            raise ValueError, "given colnames mismatch found %s" % str(cnames)
    else:
        colnames = cnames
        
    for row in reader:
        if row:
            row = [row[x] for x in xrange(len(colnames))]
            break
    stream.pushback()

    typemap_orig = typemap
    typemap = []
    for i,c in enumerate(colnames):
        c = c.strip()
        try:
            typemap.append(typemap_orig[i])
        except (IndexError, KeyError, TypeError):
            try:
                typemap.append(typemap_orig[c])
            except (IndexError, KeyError, TypeError):
                typemap.append(miscutils.infer_type_from_string(row[i].strip()))

    reader = ([typemap[i](r[i].strip())
                for i in xrange(len(colnames))] for r in reader)

    return reader, colnames


__all__ = ['from_stream', 'to_stream', 'from_csv_stream']
