#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import math
import copy
import datetime
import calendar
import datetime
import itertools

from rayon import RayonException
import scales
import data

import netsa.data.times

# TODO: Remove anything in there that doesn't require sorting: I'm
# doing mean, stdev, etc in the Column class already, so it's
# redundant. Should probably rename, too.
class Statistics(object):
    """Median, percentile and interquartile-range statistics on a
    univariate data set."""
    def __init__(self, data, presorted=False, numeric=False):
        if presorted:
            self.data = data
        else:
            try:
                self.data = sorted(data)
            except:
                raise data.RayonDataException("Column data cannot be sorted")
        self.__mean = None
        self.__population_variance = None
        self.__population_stdev = None
        self.__sample_variance = None
        self.__sample_stdev = None

    def percentile(self, p):
        "Value of pth percentile element in the dataset."
        # TODO: range checking (e.g., there is no 100th percentile)

        # Special case len(data) == 0 and len(data) == 0
        if len(self.data) == 0: return None
        if len(self.data) == 1: return self.data[0]

        idx = len(self.data) * p / 100
        if idx == math.floor(idx):
            return self.data[int(idx)]
        else:
            idx = math.floor(idx)
            return (self.data[int(idx)] + self.data[int(idx) + 1]) / 2.0

    def iqr(self):
        "Inter-quartile range"
        return self.percentile(75) - self.percentile(25)

    def _iqr_lo_idx(self, multiplier):
        """Highest value that is more than multiplier * iqr less than the
        25-percentile value."""
        threshold = self.percentile(25) - (multiplier * self.iqr())
        for idx in xrange(len(self.data)):
            if self.data[idx] > threshold:
                if idx == 0:
                    return 0
                else:
                    return idx - 1
    def iqr_lo(self):
        return self.data[self._iqr_lo_idx(1.5)]

    def _iqr_hi_idx(self, multiplier):
        """Lowest value that is more than multiplier * iqr greater than the
        75-percentile value."""
        threshold = self.percentile(75) + (multiplier * self.iqr())
        for idx in xrange(len(self.data) - 1, 0, -1):
            if self.data[idx] < threshold:
                if idx == (len(self.data) - 1):
                    return len(self.data) - 1
                else:
                    return idx + 1
    def iqr_hi(self):
        return self.data[self._iqr_hi_idx(1.5)]

    def mild_outliers(self):
        """Values of all mild (< 3 * iqr) outliers"""
        vals = []
        lo_15_idx = self._iqr_lo_idx(1.5)
        lo_30_idx = self._iqr_lo_idx(3)
        hi_15_idx = self._iqr_hi_idx(1.5) + 1
        hi_30_idx = self._iqr_hi_idx(3)
        if hi_30_idx != len(self.data) - 1:
            hi_30_idx += 1
        vals.extend(self.data[lo_30_idx : lo_15_idx])
        vals.extend(self.data[hi_15_idx : hi_30_idx])
        return vals

    def extreme_outliers(self):
        """Values of all extreme (>= 3 * iqr) outliers"""
        vals = []
        lo_30_idx = self._iqr_lo_idx(3)
        hi_30_idx = self._iqr_hi_idx(3)
        vals.extend(self.data[:lo_30_idx])
        if hi_30_idx != len(self.data) - 1:
            vals.extend(self.data[hi_30_idx + 1:])
        return vals
    
    def max(self):
        try:
            return self.data[-1]
        except IndexError:
            return None
    def min(self):
        try:
            return self.data[0]
        except IndexError:
            return None

    def mean(self):
        """Mean of column data. It's more efficient to get this from
        the L{Column} than to create a C{Statistics} object and call
        this method, since the latter method incurs the penalty of
        sorting the data."""
        if not self.__is_numeric:
            raise data.RayonDataException("Can't call mean "
                                          "with non-numeric data")
        if self.__mean is None:
            self.__mean = sum(self.data) / len(self.data)
        return self.__mean
        
    def variance(self):
        """Return the population variance of the data in the column
        data. (The data is assumed to represent the total population.)"""
        if not self.__is_numeric:
            raise data.RayonDataException("Can't call variance "
                                          "with non-numeric data")
        if self.__population_variance is None:
            self.__population_variance = (sum((x - self.mean())**2
                                              for x in self.data) /
                                          len(self.data))
        return self.__population_variance
    def stdev(self):
        """Return the population standard deviation of the column
        data. (The data is assumed to represent the total population.)"""
        if not self.__is_numeric:
            raise data.RayonDataException("Can't call stdev "
                                          "with non-numeric data")
        if self.__population_stdev is None:
            self.__population_stdev = math.sqrt(self.variance())
        return self.__population_stdev

    def sample_variance(self):
        """Return the sample variance of the data. (The data is
        assumed to represent a sample of a larger population."""
        if not self.__is_numeric:
            raise data.RayonDataException("Can't call sample_variance "
                                          "with non-numeric data")
        if self.__sample_variance is None:
            self.__sample_variance = (sum((x - self.mean())**2
                                          for x in self.data) /
                                      (len(self.data) - 1))
        return self.__sample_variance

    def sample_stdev(self):
        """Return the sample standard deviation of the column
        data. (The data is assumed to represent a sample of a larger
        population."""
        if not self.__is_numeric:
            raise data.RayonDataException("Can't call sample_stdev "
                                          "with non-numeric data")
        if self.__sample_stdev is None:
            self.__sample_stdev = math.sqrt(self.sample_variance())
        return self.__sample_stdev

# def dev2usr_x(cr, x):
#     return cr.device_to_user_distance(x, 0)[0]

# def dev2usr_y(cr, y):
#     return cr.device_to_user_distance(0, y)[1]


# def rebin(bin_size, data):
#     # Note that rebin works on lists, not Datasets. TODO: change
#     new_bins = []
#     # Sort data once
#     data = sorted(data, cmp=lambda a,b: cmp(a[0], b[0]))
#     bin_stats = Statistics([i[0] for i in data], presorted=True)
#     for i in xrange(bin_stats.min(), bin_stats.max(), bin_size):
#         new_bins.append([i, 0])

#     # TODO: eliminate short bins by distributing remainder
#     # across all other bins evenly. Shouldn't matter for the
#     # current set of data, however.
#     curr_idx = 0
#     last_idx = 0
#     for label, size in data:
#         found = False
#         for idx in xrange(last_idx - 1, len(new_bins) - 1):
#             if new_bins[idx + 1][0] > label:
#                 found = True
#                 curr_idx = idx
#                 break
#         if found:
#             new_bins[curr_idx][1] += size
#         else:
#             new_bins[-1][1] += size
#         last_idx = curr_idx
#     return new_bins



class Parameters(object):
    """An object for holding parameters to, for instance, a
    visualization.

    This object should provide a subset of the optparse.OptionValues
    class's capabilities, so that an OptionValues can be used in its
    stead. This will let visualization functions take their parameters
    from either the command line or other functions.
    """
    def __init__(self, **defaults):
        self.__dict__ = defaults
    def dict(self, include={}, exclude=[]):
        newdict = copy.copy(self.__dict__)
        newdict.update(include)
        for key in exclude:
            newdict.pop(key, None)
        return newdict
    def has_key(self, k):
        return self.__dict__.has_key(k)


# Binning data

class _BinsWithZeroes(object):
    def __init__(self, bin_name, colnames, min_bin, max_bin, bin_size):
        self.bin_name = bin_name
        self.colnames = colnames

        self.min_bin = min_bin
        self.max_bin = max_bin
        self.bin_size = bin_size

        def bin_walker():
            return xrange(int(min_bin),
                          (int(max_bin + bin_size)),
                          int(bin_size))

        # self.bins = [[bin, val1, val2, ...],
        #              ...] for each bin
        self.bins = [[b] + [0 for c in colnames] for b in bin_walker()]
        self.bins_by_idx = dict((b, idx) for idx, b in enumerate(bin_walker()))
    def add_to_bin(self, bin_label, row):
        idx = self.bins_by_idx[bin_label]
        bin = self.bins[idx]
        for colidx, colname in enumerate(self.colnames):
            bin[colidx + 1] += row[colname]
    def get_bin_iter(self):
        return iter(self.bins)

class _BinsWithoutZeroes(object):
    def __init__(self, bin_name, colnames, min_bin, max_bin, bin_size):
        self.bin_name = bin_name
        self.colnames = colnames
        self.min_bin = min_bin
        self.max_bin = max_bin
        self.bin_size = bin_size
        self.bins = {}
    def add_to_bin(self, bin_val, row):
        def make_bin():
            return [0 for c in self.colnames]
        if not self.bins.has_key(bin_val):
            self.bins[bin_val] = [0 for c in self.colnames]
        bin = self.bins[bin_val]
        for colidx, colname in enumerate(self.colnames):
            bin[colidx] += row[colname]
    def get_bin_iter(self):
        # TODO: verify that int(bin_size) == bin_size always.
        for bin_val in xrange(int(self.min_bin),
                              int(self.max_bin + self.bin_size),
                              int(self.bin_size)):
            if self.bins.has_key(bin_val):
                yield [bin_val] + self.bins[bin_val]
    
        

def bin_data(d, bin_column=0, bin_size=1,
             output_zeroes=True, output_negatives=True,
             max_bin=None, min_bin=None, ignore_columns=[]):
    """
    Aggregates rows in *d* based on *bin_col*. Returns a new
    :class:`~rayon.data.Dataset` object containing the binned
    data. The data types of the data in the columns in the dataset
    (except those in ignore columns) must support addition.

    This function may sort the input dataset according to
    *bin_column*.

    *d* is a :class:`~rayon.data.Dataset` object that will be binned.

    *bin_column* is the index or name of a column in *d* that contains
    the values to sort into bins.

    *bin_size* is the size of each bin.

    If *output_zeroes* is ``True`` (the default), the output dataset
    will contain rows for every bin between the lowest bin and the
    highest, including bins which have zero values. If *output_zeroes*
    is ``False``, the output dataset will contain only bins with
    non-zero values.

    *output_negative* behaves like *output_zeroes*, for negative
    numbers.

    If supplied, *max_bin* and *min_bin* are, respectively, the
    largest and smallest bin values. Data in *bin_column* falling
    below *max_bin* and above the sum of *max_bin* and *bin_size* will
    be ignored. If *max_bin* or *min_bin* are not multiples of
    *bin_size*, they will be adjusted up (for *max_bin*) or down (for
    *min_bin*) to the nearest multiple of *bin_size*.

    If supplied *ignore_columns* is a sequence of column names in
    *d*. These columns will be ignored during binning, and the output
    dataset will not contain them.
    """
    if max_bin is None or min_bin is None:
        d.sort(lambda r: r[bin_column])
    if max_bin is None:
        dmax = d.get_row(-1)[bin_column]
        max_bin = (dmax // bin_size) * bin_size
    else:
        max_bin = ((max_bin // bin_size) * bin_size) + bin_size
    if min_bin is None:
        dmin = d.get_row(0)[bin_column]
        min_bin = (dmin // bin_size) * bin_size
    else:
        min_bin = ((min_bin // bin_size) * bin_size)

    # Make sure we're dealing with column names
    bin_column = d.get_column_name_from_index(bin_column)
    ignore_columns = set(d.get_column_name_from_index(i)
                         for i in ignore_columns)
        
    # Get output ordering for new dataset (omit bin column)
    keep_columns = [c for c in d.get_column_names()
                    if c != bin_column and c not in ignore_columns]

    if output_zeroes:
        binner = _BinsWithZeroes(bin_column, keep_columns,
                                 min_bin, max_bin, bin_size)
    else:
        binner = _BinsWithoutZeroes(bin_column, keep_columns,
                                 min_bin, max_bin, bin_size)
        
    for row in d:
        bin_val = (row[bin_column] // bin_size) * bin_size
        binner.add_to_bin(bin_val, row)

    columns = list(itertools.izip(*list(binner.get_bin_iter())))
    return data.Dataset(columns, colnames=[bin_column] + keep_columns)
    

# Tools for data trending

def moving_average(input, window=300):
    """
    Calculates a moving average over a set of time-series data. This
    function returns a new :ry:factory:`Dataset` object with two columns:
    ``time`` is a :class:`datetime.datetime` representing the center
    time of that window, and ``value`` is the average at the center of
    that window. The number of rows in this :ry:factory:`Dataset` will not
    correspond to those in *input*. (New observations are added to
    the return data when a window changes -- that is, when a new
    observation enters the window or an old observation leaves it.)

    *input* is a a 2-column :ry:factory:`Dataset`. The type of first column is
    :class:`datetime.datetime` objects; the type of the second is
    numeric.

    *window* is the size of the window, in seconds.
    """

    times = []
    vals  = []
    output = []

    def emit():
        output.append([netsa.data.times.make_datetime(sum(times) / len(times)),
                       sum(vals) / len(vals)])

    for row in input:
        newest = calendar.timegm(row[0].timetuple())
        val  = row[1]
        while len(times) > 0 and newest - times[0] > window:
            emit()
            times.pop(0)
            vals.pop(0)
        times.append(newest)
        vals.append(val)
    emit()
    return data.Dataset.from_rows(output, colnames=['time', 'value'])

def kernel_epanechnikov(x):
    """
    The Epanechnikov kernel smoother. The default smoother for
    :func:`regression_nw`.
    """
    if abs(x) >= math.sqrt(5):
        return None
    else:
        return .75 * (1 - ((x**2)/5)) / math.sqrt(5)


def kernel_gauss(x, throw_oor=False):
    """
    The Gaussian kernel smoother. Can be used as a parameter to 
    :func:`regression_nw`.
    """
    return ((2 * math.pi)**(-1/2)) * (math.e**(((-1 * x)**2)/2))


def binary_search(indata, sortcolidx, x):
    # Searches through a sorted Dataset for X in the column referenced
    # by sortcolidx. Should probably be moved into Dataset.
    start = 0
    end = indata.numrows() - 1
    while True:
        length = end - start
        if length == 1:
            return start
        candidate = start + (length // 2)
        cand_val = indata.column(sortcolidx)[candidate]
        if cand_val == x:
            return candidate
        elif cand_val > x:
            end = candidate
        elif cand_val < x:
            start = candidate


def regression_single_nw(x, indata, bandwidth, x_colidx, y_colidx, kernel):
    mid_idx = binary_search(indata, x_colidx, x)
    # Unlike regression_single_nw_naive, this function takes advantage
    # of the locality of the kernel smoother we're using to avoid
    # processing every data point. It throws away some real
    # information, though, so it will be less precise.

    # Get normalized weights for x. (Collect these over the whole
    # dataset for now. May or may not be the right thing to do.)
    normalized_weights = 0

    # Go down first. Stop when we hit the end of the set, or if
    # weighted_value returns None (suggesting that the kernel function
    # is out of range).
    curr_idx = mid_idx
    while curr_idx >= 0:
        row = indata.row(curr_idx)
        nw = kernel((x - row[x_colidx]) / bandwidth)
        if nw is None:
            break
        normalized_weights += nw
        curr_idx -= 1

    # Now go up, with the same drill
    curr_idx = mid_idx + 1
    while curr_idx < indata.numrows():
        row = indata.row(curr_idx)
        nw = kernel((x - row[x_colidx]) / bandwidth)
        if nw is None:
            break
        normalized_weights += nw
        curr_idx += 1

    # For the actual regression estimate (r_hat), do the same thing as
    # with normalized weights -- go out from mid_idx and stop when/if
    # the kernel tells you.
    def weighted_value(row):
        kernel_val = kernel((x - row[x_colidx])/bandwidth)
        if kernel_val is None:
            return None
        weights = kernel_val/normalized_weights
        return row[y_colidx] * weights
        
    r_hat = 0
    curr_idx = mid_idx
    while curr_idx >=0:
        row = indata.row(curr_idx)
        val = weighted_value(row)
        if val is None:
            break
        r_hat += val
        curr_idx -= 1

    curr_idx = mid_idx + 1
    while curr_idx < indata.numrows():
        row = indata.row(curr_idx)
        val = weighted_value(row)
        if val is None:
            break
        r_hat += val
        curr_idx += 1
        
    return r_hat

def regression_single_nw_naive(x, indata, bandwidth,
                               x_colidx, y_colidx, kernel):
    # The naive estimator runs over the whole dataset. The output,
    # seems to be smoother and more sensible, but takes longer than
    # regression_single_nw.
    
    # Normalized weights
    normalized_weights = 0
    for row in indata:
        nw = kernel((x - row[x_colidx]) / bandwidth)
        if nw is not None:
            normalized_weights += nw

    r_hat = 0
    for row in indata:
        kernel_val = kernel((x - row[x_colidx])/bandwidth)
        if kernel is None:
            continue
        weights = kernel_val / normalized_weights
        r_hat += row[y_colidx] * weights

    return r_hat
    

def regression_nw(indata, output_size, bandwidth=None,
                  x_colidx=0, y_colidx=1,
                  kernel=kernel_epanechnikov):
    """
    Calculates a specified number of points on the regression function
    of input, using the Nadaraya-Watson kernel estimator. The input
    and output are assumed to be two-dimensional points represented in
    the Cartesian coordinate system. (The input dataset can have more
    than two columns, however; use *x_colidx* and *y_colidx* to
    identify the columns representing *x* and *y* position.)

    *indata* is a :class:`rayon.data.Dataset` object containing the
    data to regress.

    *output_size* is the desired number of points along the function
    to return.

    *bandwidth* is the expected "spread" of the data distribution; if
    it is ``None``, the bandwidth is computed from *indata*.

    *x_colidx* is the index or name of the column in *indata* that
    represents the *x* coordinate.

    *y_colidx* is the index or name of the column in *indata* that
    represents the *y* coordinate.

    *kernel* is a function that takes a float and returns a float. It
    is the kernel function to be used for smoothing.

    This function returns a :ry:factory:`Dataset` object with *output_size*
    rows and two columns named (in order) ``x`` and ``y``,
    representing the *x* and *y* coordinates of points along the
    regression function.
    """
    indata.sort(key=lambda row: row[x_colidx])
    try:
        x_col = indata.column(x_colidx, False) # avoid sorting the column

        # Pick a semi-sensible value for bandwidth (5% of x range), if
        # unspecified
        if bandwidth is None:
            bandwidth = (x_col.max() - x_col.min()) * .05

        if output_size > indata.numrows():
            output_size = indata.numrows()

        interval = (x_col.max() - x_col.min()) / output_size
        output = []
        x_curr = x_col.min()
        while x_curr <= x_col.max():
            y_curr = regression_single_nw(x_curr, indata,
                                          bandwidth=bandwidth,
                                          x_colidx=x_colidx,
                                          y_colidx=y_colidx,
                                          kernel=kernel)
            output.append((x_curr, y_curr))
            x_curr += interval

        return data.Dataset.from_rows(output, colnames=['x', 'y'])
    finally:
        indata.sort()

def regression_ols(indata, x_colidx=0, y_colidx=1):
    """
    Derive the ordinary least-squares regression line of y on x in
    *indata*.

    *indata* is a :class:`rayon.data.Dataset` object containing the
    data to regress.

    *x_colidx* is the index or name of the column in *indata* that
    represents the *x* coordinate.

    *y_colidx* is the index or name of the column in *indata* that
    represents the *y* coordinate.

    This function returns a :ry:factory:`Dataset` object with two rows and
    two columns named (in order) ``x`` and ``y``. The first row
    represents a point on the regression line at the lowest value of
    *x_colidx*; the second represents a point on the regression line
    at the highest value of *x_colidx*.
    """
    # Some references for the math in this section:

    # http://www.stat.wmich.edu/s216/book/node126.html
    # http://en.wikipedia.org/wiki/Pearson_product-moment_correlation_coefficient

    x = indata.column(x_colidx)
    y = indata.column(y_colidx)
    # Get pearson correlation
    r = sum((_x - x.mean()) / x.sample_stdev() *
            (_y - y.mean()) / y.sample_stdev()
            for _x, _y in zip(x, y)) / (len(x) - 1)

    # Get slope (b) and intercept (a) of regression line 
    b = (y.sample_stdev() / x.sample_stdev()) * r
    a = y.mean() - (b * (x.mean()))

    # Compute predicted y for first and last points of x
    actual_x = (x[0], x[-1])
    predicted_y = [a + (b * _x) for _x in actual_x]

    return data.Dataset.from_columns([actual_x, predicted_y],
                                     colnames=['x', 'y'])
        
    




trend_algorithms = {
    "moving_avg": moving_average,
    "kernel"    : regression_nw,
    "ols"       : regression_ols
}

def trend_data(trend_name, indata, **trend_args):
    try:
        trend_fn = trend_algorithms[trend_name]
        return trend_fn(indata, **trend_args)
    except KeyError:
        raise RuntimeError("No such trend function: '%s'" % trend_name)


smooth_data = trend_data

__all__ = """
Statistics
smooth_data
moving_average
regression_nw
regression_ols
kernel_gauss
kernel_epanechnikov
""".split()
