#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Event handling.

We have three classes of event:

   1. Backend events, emitted by the rendering backend. Examples:
      C{mouse_left_down}, C{mouse_motion}, C{idle}. Only implementors
      of new backends will need to worry about these.
  
   2. Low-level Rayon events. These events are about as primitive as
      the backend events, but are generic, so Rayon components can
      express their behavior in terms of these events without tying
      themselves to a specific backend. Users may act on these events,
      or they may listen for high-level events. Implementors of new
      plot types and the like also have to worry about
      these. Examples: C{mouse_left_down}, C{mouse_motion},
      C{tooltip}. (Not C{idle}, as some backends may not emit that
      event or make it possible to infer it. This layer exists, in
      part, to smooth over such inconsistencies in low-level APIs)

   3. High-level Rayon events. Rayon components may synthesize these
      events in response to combinations of low-level events.
      Example: C{bar_selected}, C{points_selected}.

To translate between these event types, we have a series of
I{Adapters} and I{Listeners}:

   - Adapters receive events (typically from the rendering backend),
     translate them into different events, and dispatch them to
     Listeners.

   - Listeners receive events and act on them.

An Adapter is a kind of Listener, but for this discussion it will be
treated separately.

An L{EventSource} is implemented in the backend and associated with
an event emitter suitable to the backend rendering system. It emits
low-level Rayon events.

The L{EventListener} is a base class for anything that listens to
low-level Rayon events (including other adapters -- see below). As you
might guess, it receives low-level Rayon events and acts on them.

A L{EventListener} receives low-level Rayon events and emits events
specific to a component type. It is associated with a specific
instance of that component type, and with a C{EventSource}
associated with the backend object on which the component is
rendering.

Generally, C{EventListener}s are bound to a chart applied to a
rendering surface. C{EventListener}s are registered with the
C{EventListener} associated with the (chart, surface) tuple on which
we wish to capture events using the C{EventListener}'s
C{register_adapter} method.


Handling Events
===============

Event handlers take the following arguments:

   - C{target} - in a C{EventListener}, the target is the chart
     being acted upon. In a C{EventListener}, the target is the
     chart component being acted on.

   - C{evt} - an instance of C{RayonEvent} describing the event being
     handled.

When an event handler is done, it may return one of the following
three codes, both of which are defined in this module:

   - C{CONTINUE} - Continue processing other handlers registered for
     this event.

   - C{STOP} - Do not process any more handlers for this event.

Returning nothing or explicitly returning C{None} are equivalent to
returning C{CONTINUE}.

Rayon Event Interface
---------------------

The low-level events that may be emitted by C{EventSource} are:

  - C{mouse_left_down} - User has pushed the left mouse button down
  
  - C{mouse_left_up} - User has released the left mouse button
  
  - C{mouse_left_dclick} - User has double-clicked with the left mouse button
  
  - C{mouse_right_down} - User has pushed the right mouse button down

  - C{mouse_right_up} - User has released the right mouse button
  
  - C{mouse_right_dclick} - User has double-clicked with the right mouse button
  
  - C{mouse_motion} - User is moving the mouse
  
  - C{tooltip} - The GUI wishes to emit a tooltip

The following events are only emitted by C{EventListener}s:

  - C{component_enter} - Mouse is entering the component

  - C{component_exit} - Mouse is leaving the component

Additional subclasses of C{EventListener} may implement events
specific to the type of component being listened to. (For example,
when the user selects one or more points.)

Implementing Event Handlers
---------------------------

To implement an event handler, define a method named C{on_<event
name>}. For example, to handle the C{tooltip} event, define an
C{on_tooltip_} method.

Example
-------

Given the following listener (note that C{bars_at} may return multiple
bars, if they overlap):

>>> class MyBarListener(events.EventListener):
...     def on_tooltip__(self, target, evt):
...         bars = target.bars_at(evt.x(), evt.y()
...         if len(bars) > 0:
...             barlist = \"\\n\".join(\"  * %s\" % str(b) for b in bars)
...             evt.set_tooltip(\"You're hovering over:\\n%s\" % barlist)
...

The following describes how you would associate it with a Rayon
component you wish to listen on, and the adapter for associated with
the low-level UI element you will receive events from::.

>>> chart = make_barchart() # details left to the reader
>>> # panel is a UI element that emits events. Get a low-level adapter
>>> # for the event from the chart
>>> adapter = chart.get_event_source(panel)
>>> # Get the component in the chart whose events you want to
>>> # listen to (e.g., a barplot)
>>> component = chart.get_plot().get_subplot(\"bars\")
>>> # Create the listener associate with that component and
>>> # register it with the adapter
>>> listener = MyBarListener(component)
>>> listener.register_with(adapter)

The result would be that, if you hovered the mouse over a bar in
C{bars}, a tooltip would appear saying::

   You're hovering over:
     * <rayon.plots.BarPlot.Bar instance at 0x857b0>

"""

import sys



# Event return codes
CONTINUE = None
STOP     = "stop"

def check_rayon_event_implemented(obj, methodname):
    cname = obj.__class__.__name__
    if cname  == 'RayonEvent':
        raise RuntimeError("%s not implemented on %s." % (methodname, cname))

class EventSource(object):
    """
    A wrapper over an object in the rendering backend that emits
    events. (E.g., a GUI component.)
    """
    def activate(self):
        """
        Sets the :class:`EventSource` object to begin emitting
        events. If the :class:`EventSource` object is already
        activated, calling this method does nothing.
        """
        pass
    def deactivate(self):
        """
        Sets the :class:`EventSource` object to stop emitting
        events. If the :class:`EventSource` object is already
        deactivated, calling this method does nothing.
        """
        pass
    def subscribe(self, evt_name, handler):
        # subscribe() is called by EventListeners to register interest
        # in evt_name
        pass
    
    
class RayonEvent(object):
    """
    Contains information about UI events.
    """
    def __init__(self):
        if self.__class__.__name__ == 'RayonEvent':
            raise RuntimeError("Cannot instantiate RayonEvent.")
    def is_mouse_right_button_pressed(self):
        """
        Returns ``True`` if the right mouse button is pressed, and
        ``False`` if the right mouse button is not pressed. Returns
        ``None`` if the event backend can't determine the mouse button
        state.
        """
        check_rayon_event_implemented(self, 'is_mouse_right_button_pressed')
        
    def is_mouse_left_button_pressed(self):
        """
        Returns ``True`` if the left mouse button is pressed, and
        ``False`` if leftright mouse button is not pressed. Returns
        ``None`` if the event backend can't determine the mouse button
        state.
        """
        check_rayon_event_implemented(self, 'is_left_right_button_pressed')
        
    def is_mouse_moving(self):
        """
        Returns ``True`` if the mouse is in motion, and ``False`` if
        it is not. Returns ``None`` if the event backend can't
        determine the mouse motion state.
        """
        check_rayon_event_implemented(self, 'is_mouse_moving')
        
    def get_x(self):
        """
        Returns the X coordinate of the mouse in Cartesian
        coordinates, relative to the left edge of the responsible
        component's bounding box.

        Returns ``None`` if the event backend can't determine mouse
        positions.
        """
        check_rayon_event_implemented(self, 'get_x')

    def x(self):
        """
        A shortcut method for calling :meth:`RayonEvent.get_x`.
        """
        return self.get_x()
        
    def get_y(self):
        """
        Returns the Y coordinate of the mouse in Cartesian
        coordinates, relative to the top edge of the responsible
        component's bounding box.

        Returns ``None`` if the event backend can't determine mouse
        positions.
        """
        check_rayon_event_implemented(self, 'get_y')

    def y(self):
        """
        A shortcut method for calling :meth:`RayonEvent.get_y`.
        """
        return self.get_y()
        
    def get_position(self):
        """
        Returns a tuple ``(x, y)``, giving the mouse's position

        Returns ``None`` if the event backend can't determine mouse
        positions.
        """
        x = self.get_x()
        y = self.get_y()
        if x is None or y is None:
            return None
        else:
            return (x, y)
        


# The below class structure was a bit of a mess -- EventListener was a
# parent class with just one child, and the child had some
# half-implemented additional handlers on it. The revised version is
# below; this is left for informational purposes, and should be
# removed at some point.
        
# class EventListener(object):
#     def __init__(self):
#         self.handlers = [
#             ('mouse_left_down', self.on_mouse_left_down_),
#             ('mouse_left_up', self.on_mouse_left_up_),
#             ('mouse_left_dclick', self.on_mouse_left_dclick_),
#             ('mouse_right_down', self.on_mouse_right_down_),
#             ('mouse_right_up', self.on_mouse_right_up_),
#             ('mouse_right_dclick', self.on_mouse_right_dclick_),
#             ('mouse_motion', self.on_mouse_motion_),
#             ('tooltip', self.on_tooltip_)]

        
#     def on_mouse_left_down_(self, target, evt):
#         """
#         Called when the left mouse button is pressed while the mouse
#         is positioned over the listening 
#         """
#         pass
    
#     def on_mouse_left_up_(self, target, evt):
#         pass

#     def on_mouse_left_dclick_(self, target, evt):
#         pass

#     def on_mouse_right_down_(self, target, evt):
#         pass

#     def on_mouse_right_up_(self, target, evt):
#         pass

#     def on_mouse_right_dclick_(self, target, evt):
#         pass

#     def on_mouse_motion_(self, target, evt):
#         pass

#     def on_tooltip_(self, target, evt):
#         pass

#     def register_with(self, event_source):
#         """Register the handlers on this listener with the event source."""
#         for evt_name, handler in handlers:
#             event_source.subscribe(evt_name, handler)


        
# # TODO: on_component_enter/leave looks like a good idea, but it never
# # got fully implemented
# class ComponentListener(EventListener):
#     """Listens to low-level Rayon events and emits component-specific
#     events."""
#     def __init__(self, component):
#         EventListener.__init__(self)
# #        self.handlers.append(('component_enter', self.on_component_enter_))
# #        self.handlers.append(('component_exit', self.on_component_exit_))
#         self.component = component
#     def get_component(self):
#         return self.component
#     def register_with(self, event_source):
#         # Nope, no call to the superclass method...

#         def create_custom_handler(evt_name, handler):
#             def custom_handler(target, evt):
#                 if target == self.component:
#                     handler(target, evt)
#             custom_handler.__name__ = "%s_%s" % (evt_name, self.component)
#             return custom_handler

#         for evt_name, h in self.handlers:
#             event_source.subscribe(evt_name, create_custom_handler(evt_name, h))
            
# #    def on_component_enter_(self, target, evt): pass
# #    def on_component_exit_(self, target, evt): pass
    



class EventListener(object):
    """
    Monitors the events emitted by an :class:`EventSource` object,
    calling event handling functions at the appropriate times.  Rayon
    users can use subclasses of :class:`EventListener` to respond to
    events in customized ways by overriding the appropriate event
    handling methods.

    Event handling methods begin with ``on_`` and end with an
    underscore (``_``). They take two arguments. *target* is the
    target of the event, which is the same object passed as
    *component* to the :class:`EventListener` when it is
    created. *evt* is a :class:`RayonEvent` object describing the
    event being handled.

    *component* is a Rayon component -- a
    :class:`~rayon.borders.Border` :class:`~rayon.charts.Chart`,
    :class:`~rayon.layouts.Layout` or :class:`~rayon.plots.Plot`
    object. This :class:`EventListener` object will listen for and
    respond to events directed at this component.
    """
    def __init__(self, component):
        self.handlers = [
            ('mouse_left_down', self.on_mouse_left_down_),
            ('mouse_left_up', self.on_mouse_left_up_),
            ('mouse_left_dclick', self.on_mouse_left_dclick_),
            ('mouse_right_down', self.on_mouse_right_down_),
            ('mouse_right_up', self.on_mouse_right_up_),
            ('mouse_right_dclick', self.on_mouse_right_dclick_),
            ('mouse_motion', self.on_mouse_motion_),
            ('tooltip', self.on_tooltip_)]
        self.component = component


    def get_component(self):
        """
        Returns the component associated with this
        :class:`EventListener` object.
        """
        return self.component

    def register_with(self, event_source):
        """
        Associates this :class:`EventListener` object with
        *event_source*, an :class:`EventSource` object.
        """
        def create_custom_handler(evt_name, handler):
            def custom_handler(target, evt):
                if target == self.component:
                    handler(target, evt)
            custom_handler.__name__ = "%s_%s" % (evt_name, self.component)
            return custom_handler

        for evt_name, h in self.handlers:
            event_source.subscribe(evt_name, create_custom_handler(evt_name, h))
        
        
    def on_mouse_left_down_(self, target, evt):
        """
        Called when the left mouse button is pressed while the mouse
        is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass
    
    def on_mouse_left_up_(self, target, evt):
        """
        Called when the left mouse button is released while the mouse
        is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_mouse_left_dclick_(self, target, evt):
        """
        Called when the left mouse button is \"double clicked\"
        (presed and released twice in quick succession) while the
        mouse is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_mouse_right_down_(self, target, evt):
        """
        Called when the right mouse button is pressed while the mouse
        is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_mouse_right_up_(self, target, evt):
        """
        Called when the right mouse button is released while the mouse
        is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_mouse_right_dclick_(self, target, evt):
        """
        Called when the right mouse button is \"double clicked\"
        (presed and released twice in quick succession) while the
        mouse is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_mouse_motion_(self, target, evt):
        """
        Called when the left mouse is moved while the mouse
        is positioned over the component associated with the
        :class:`EventListener` object.
        """
        pass

    def on_tooltip_(self, target, evt):
        """
        Called when a tooltip should be displayed about the associated
        component.
        """
        pass

