#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
import quickselect
import sys

class KDNode:
    def __init__(self, val, dimension, left, right):
        self.val = val
        self.dimension=dimension
        self.left=left
        self.right=right

    def _print(self, depth=0, recurse=True):
        leader = " " * depth
        acc = [""]
        def doprint(msg):
            acc[0] += "%s%s\n" % (leader, msg)

        #doprint("Depth: %s" % depth)

        if self.dimension is None:
            doprint("TN: %s" % self.val)
        else:
            doprint("IN: Cut at %s" % self.val)
            if recurse:
                if self.left is None:
                    doprint("L: NONE")
                else:
                    doprint("L:")
                    acc[0] += self.left._print(depth+1)
                
                if self.right is None:
                    doprint("R: NONE")
                else:
                    doprint("R:")
                    acc[0] += self.right._print(depth+1)
        return acc[0]
            
    def __str__(self):
        return self._print()
    def ident(self):
        if self.dimension is None:
            return '"%s"' % str(self.val)
        else:
            return '"%s (%d) (%x)"' % (self.val, self.dimension, id(self))
    def dot(self, frame=True):
        str = ""
        if frame:
            str += "digraph {\n"
            str += 'size="11,8";\n'
            str += 'graph [rankdir=LR, size="11,8"];\n'

        if self.left is not None:
            str += self.left.dot(frame=False)
            str += '%s -> %s;\n' % (self.ident(), self.left.ident())

        if self.right is not None:
            str += '%s -> %s;\n' % (self.ident(), self.right.ident())
            str += self.right.dot(frame=False)
        if frame:
            str += "}"
        return str


def print_indent(lvl, msg):
    print >>sys.stderr, "%s%s" % (" " * lvl, msg)


class Indivisible(Exception):
    # Used internally by build() to determine if a dataset can no
    # longer be subdivided.
    pass

def build(data, width=None, depth=0, ttl=None):
    """Build a kd-tree from data. build modifies data in-place as it
    goes; make a copy and pass it into build() if ordering matters.

    @type data: list
    @param data: points to tree-ify

    @type width: int | None
    @param width: consider only first $width dimensions in building the
                 tree. (Default: use all dimensions)

    @type depth: int
    @param depth: used internally by build

    @type ttl: int | None
    @param ttl: used internally by build
    """

    def debug(msg):
        #print_indent(depth, msg)
        pass

    # For debugging:
    #if depth > 100:
    #    raise RuntimeError("Tree depth may not exceed 100")

    if len(data) == 0:
        raise RuntimeError("Empty data")

    if len(data) <= 5:
        debug( "DATA: %s" % data)


    if width is None:
        width = len(data[0])

    dimension = depth % width

    if len(data) == 1:
        # Leaf node
        debug("___Storing___: %s" % data[0])
        return KDNode([data[0]], None, None, None)
    else:
        # Get median value to use as pivot
        median_idx = len(data)//2 - 1
        debug( ("Data length: %d. Selecting median (%dth value) "
               "in dimension %d" % (len(data), median_idx, dimension)))
        sortkey = lambda p: p[dimension]
        pivot_val = quickselect.select(data, median_idx,
                                       left=0, right=len(data)-1,
                                       key=sortkey)

        # Partition data for children
        left = []
        right = []
        for d in data:
            if d[dimension] <= pivot_val[dimension]:
                left.append(d)
            else:
                right.append(d)

        debug( "Cut: %s[%d]: %s | %s" % (pivot_val[dimension],
                                         dimension, left, right))


        # Do we have empty sides? If so, handle underlying pathologies
        # in data...
        if len(left) == 0:
            # All data is > the cut point on this dimension. Shouldn't happen
            raise RuntimeError("empty left side")
        if len(right) == 0:
            # We can't subdivide the data along this dimension.
            # Recurse into the other dimensions and see if we can
            # further subdivide along any of them. If we can, do that
            # instead. If not, make this a leaf bucket with all the
            # remaining data in it.
            debug("Hmm, empty right side....")
            if ttl is None:
                # Start the counter. We will search all other
                # dimensions to see if we can subdivide further.
                ttl = width - 1 + depth
                # Try to subdivide the dimensions below this one. If
                # we can't, just put everything in a bucket.
                try:
                    debug("<<<<<<<<<<<Catching Indivisible")
                    node = build(left, width, depth+1, ttl)
                    debug(">>>>>>>>>>Not catching Indivisible")
                    return node
                except Indivisible:
                    debug("put all this shit in a bucket")
                    debug("___Storing___: %s" % left)
                    return KDNode(left, None, None, None)
            else:
                debug("TTL: %d" % ttl)
                debug("Depth: %d" % depth)
                if depth > ttl:
                    # We've tried all other dimensions. Give up
                    raise Indivisible()
                else:
                    # Go down another level. (If we can find a cut
                    # point in the other dimensions, the "normally
                    # distributed" case will continue with a default
                    # ttl, implicitly resetting ours.)
                    return build(left, width, depth+1, ttl)
        else:
            # Data is distributed into both
            debug("Building left tree...")
            left = build(left, width, depth+1)
            debug("Building right tree...")
            right = build(right, width, depth+1)
            return KDNode(val=pivot_val[dimension], dimension=dimension,
                          left=left, right=right)
        



num_compares = 0

def range_search(node, lo, hi, dims=None, lvl=0):
    """Given a hyperplane bounded by I{lo} and I{hi}, each a
    hyperpoint with I{dims} dimensions, give all the points in the
    tree this hyperplane contains."""
    global num_compares
    if node is None:
        return []

    def debug(msg):
        pass
        #print_indent(lvl, msg)

    candidates = []

    if node.dimension is None:
        # leaf node
        debug("leaf node: %s" % node.val)
        num_dims = len(node.val[0])
        for point in node.val:
            num_compares += 1
            match = True
            for idx in xrange(num_dims):
                if ((lo[idx] != '*' and point[idx] < lo[idx]) or
                    (hi[idx] != '*' and point[idx] > hi[idx])):
                    debug("%s not in %s-%s" % (point[idx], lo[idx], hi[idx]))
                    match = False
                    break
            if match:
                candidates.append(point)
    else:
        if lo[node.dimension] != '*' and node.val < lo[node.dimension]:
            # node is below our range for this axis
            debug("cut point: %s [%d] (below)" % (node.val, node.dimension))
            debug("going right...")
            candidates.extend(range_search(node.right, lo, hi, dims, lvl+1))
            debug("done")
        elif hi[node.dimension] != '*' and node.val > hi[node.dimension]:
            # node is above our range for this axis
            debug("cut point: %s [%d] (above)" % (node.val, node.dimension))
            debug("going left...")
            candidates.extend(range_search(node.left, lo, hi, dims, lvl+1))
            debug("done")
        else:
            # node is within our range, or equals one of the bounds (since
            # values equal to this node's value could live in either
            # child, we have to check both)
            debug("cut point: %s [%d] (within)" % (node.val, node.dimension))
            debug("going left...")
            candidates.extend(range_search(node.left, lo, hi, dims, lvl+1))
            debug("going right...")
            candidates.extend(range_search(node.right, lo, hi, dims, lvl+1))
            debug("done")
    return candidates

import random
import time

def log(msg):
    print >>sys.stderr, "%s: %s" % (time.time(), msg)


def tree_test(lo, hi, nelems=200):
    global num_compares
    num_compares = 0
    x = [[random.randint(0,99), random.randint(0, 99)] for _ in xrange(nelems)]
    #x = [[0, 1], [3, 10], [8, 9], [9, 17], [10, 4],
    #     [10, 11], [11, 6], [11, 19], [12, 12], [13, 15]]
    #log("x = %s (%d)" % (x, len(x)))
    log("%d items in x" % len(x))
    log("--------building tree")
    ptree = build(x)
    #log("tree is: %s" % ptree)
    log("--------searching for results")
    results = range_search(ptree, lo, hi)
    log("Number of comparisons: %d" % num_compares)
    log("--------searching for verified results")
    v_results = [point for point in x
                 if (point[0] >= lo[0] and point[0] <= hi[0] and
                     point[1] >= lo[1] and point[1] <= hi[1])]
    print sorted(results)
    print "--------------"
    print sorted(v_results)


def tree_perftest():
    for nelems in xrange(1000, 101000, 1000):
        for num_dims in xrange(2, 11):
            build_times = []
            search_times = []
            for _ in xrange(5):
                x = [[random.randint(0, 99) for a in xrange(num_dims)]
                     for b in xrange(nelems)]
                stime = time.time()
                tree = build(x)
                build_time = time.time() - stime
                lo = [20 for a in xrange(num_dims)]
                hi = [80 for a in xrange(num_dims)]
                stime = time.time()
                results = range_search(tree, lo, hi)
                search_time = time.time() - stime
                build_times.append(build_time)
                search_times.append(search_time)
            sys.stdout.write("%s,%s,%s,%s\n" %
                             (nelems, num_dims,
                              sum(build_times) / len(build_times),
                              sum(search_times) / len(search_times)))
            sys.stdout.flush()


#tree_perftest()

def dot_test():
    x = [[_, 50, 100] for _ in xrange(10)]
    t = build(x)
    sys.stdout.write(t.dot())
    sys.stdout.flush()


#dot_test()
