#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division

from rayon import RayonException, DEBUG
import rayon.base as base
from rayon.backend import units_from_spec, LayoutUnits


# TODO: make a NullSpacing singleton and use it as the default
# spacing.


def point_add(p1, p2):
    # I do this a lot, but it's more convenient to treat points as
    # tuples most of the time, so this is a function instead of a
    # class.
    return (p1[0] + p2[0], p1[1] + p2[1])



#
# Spacing/padding information
#

class BoxSpacing(object):
    """Spacing for a box with dynamic width and height determined by
    available space.

    A C{BoxSpacing} stores spacing information about its associated
    L{Layout}, and should only ever be used from within that
    manager's C{layout} method.
    """
    def __init__(self, tpad=0, bpad=0, lpad=0, rpad=0, width=None, height=None):
        self.tpad = units_from_spec(tpad)
        self.bpad = units_from_spec(bpad)
        self.lpad = units_from_spec(lpad)
        self.rpad = units_from_spec(rpad)
        if width is None:
            self.width = width
        else:
            self.width = units_from_spec(width)
        if height is None:
            self.height = height
        else:
            self.height = units_from_spec(height)
    def left_padding(self, ctx, width):
        return self.lpad.to_pixels(ctx, width)
    def right_padding(self, ctx, width):
        return self.rpad.to_pixels(ctx, width)
    def top_padding(self, ctx, height):
        return self.tpad.to_pixels(ctx, height)
    def bottom_padding(self, ctx, height):
        return self.bpad.to_pixels(ctx, height)
    def vertical_padding(self, ctx, height):
        return self.top_padding(ctx, height) + self.bottom_padding(ctx, height)
    def horizontal_padding(self, ctx, width):
        return self.left_padding(ctx, width) + self.right_padding(ctx, width)
    def inner_height(self, ctx, height):
        rc = height - (self.tpad.to_pixels(ctx, height) +
                       self.bpad.to_pixels(ctx, height))
        if rc < 0: return 0
        return rc
    def inner_width(self, ctx, width):
        rc = width - (self.lpad.to_pixels(ctx, width) +
                      self.rpad.to_pixels(ctx, width))
        if rc < 0: return 0
        return rc
    def total_width(self, ctx, width):
        if self.width is None:
            return width
        else:
            return self.width.to_pixels(ctx, width)
    def total_height(self, ctx, height):
        if self.height is None:
            return height
        else:
            return self.height.to_pixels(ctx, height)
    def padded_topleft(self, ctx, real_topleft, width, height):
        # Given the top left corner of a box being padding with this
        # BoxSpacing, return the top left corner of drawable space.
        x, y = real_topleft
        return (x + self.lpad.to_pixels(ctx, width),
                y + self.tpad.to_pixels(ctx, height))
    def padded_bottomright(self, ctx, real_bottomright, width, height):
        # Given the bottom right corner of a box being padding with this
        # BoxSpacing, return the bottom right corner of drawable space.
        x, y = real_bottomright
        rc = (x - self.rpad.to_pixels(ctx, width),
              y - self.bpad.to_pixels(ctx, height))
        return rc
    def __repr__(self):
        return ("BoxSpacing(tpad=%s, bpad=%s, "
                "lpad=%s, rpad=%s, width=%s, height=%s" % (
                repr(self.tpad), repr(self.bpad),
                repr(self.lpad), repr(self.rpad),
                repr(self.width), repr(self.height)))

# Bounding Box information

class LayoutNode(object):
    """A region of canvas that has been subdivided by a L{Layout}. A
    C{LayoutNode} contains its own bounding information and, if the
    area was further subdivided, a set of C{LayoutNode}s representing
    those child regions."""
    def __init__(self, children, topleft, bottomright, zorder, component,
                 leaf=False):
        """Note that I{bottomright} is the point directly I{after} the
        bottom right corner. (IOW, the bounding box runs
        C{[I{topleft}, I{bottomright})}.)
        """
        self.children = children
        self.topleft = topleft
        self.bottomright = bottomright
        self.zorder = zorder
        self.component = component
        self.leaf = leaf
    def is_leaf(self): return self.leaf
    def get_component(self): return self.component
    def __contains__(self, elem):
        x, y = elem
        return (self.topleft[0] <= x and self.bottomright[0] > x and
                self.topleft[1] <= y and self.bottomright[1] > y)
    def get_eligible_children(self, x, y):
        # Return list of children with claims on (x, y), sorted by z-order
        return sorted([child for child in self.children if (x, y) in child],
                      key=lambda c: c.zorder)
    def search(self, x, y, callback):
        nodes = [self]
        leaves = []
        levels = 0
        def p(msg):
            print "%s%s" % (" " * levels, msg)
        while len(nodes) > 0:
            childnodes = []
            for node in nodes:
                if node.is_leaf():
                    #try:
                    #    p("LEAF  : %s" % node.get_component().get_component())
                    #except:
                    #    p("LEAF  : %s" % node.get_component())
                    callback(node, x, y, node.topleft)
                else:
                    #p("BRANCH: %s" % node.get_component())
                    kids = node.get_eligible_children(x, y)
                    childnodes.extend(kids)
            nodes = childnodes
            levels += 1
        return
    def __str__(self):
        if self.is_leaf():
            start, end = "<", ">"
        else:
            start, end = "[", "]"
        try:
            clabel = "%s(%s)" % (
                type(self.component).__name__,
                type(self.component.get_component()).__name__)
        except AttributeError:
            clabel = type(self.component).__name__
        s = "(%s <-> %s; z=%s) %s" % (
            self.topleft, self.bottomright, self.zorder, clabel)
        return start + s + end
    def pprint(self, indent=0):
        print (" " * indent) + str(self)
        for child in self.children:
            child.pprint(indent+1)




# The main Layout class

class Layout(base.Component):
    """Base class for objects managing layout of one or more contained
    components."""
    def __init__(self, spacing):
        assert self.__class__.__name__ != "Layout", \
               "Layout class should not be instantiated"
        self.spacing = spacing
    def get_spacing(self):
        """
        Returns the :class:`BoxSpacing` object associated with this
        layout.
        """
        return self.spacing
    def set_spacing(self, new_spacing):
        """
        Replace the existing spacing information with *new_spacing*, a
        :class:`BoxSpacing` object.
        """
        self.spacing = new_spacing
    def is_layoutable(self): return True
    def layout(self, ctx, topleft, width=None, height=None, call_depth=0):
        """Allocate space to children based on I{width} and I{height}
        and call them to further subdivide or draw. If you are
        subclassing C{Layout}, note that C{layout} should
        handle its own padding. (I.e., translate the origin past the
        left and top padding and draw within the portion of the width
        and height that is not used for padding.)

        @type  ctx: L{rayon.backend_base.Context}
        @param ctx: Context object with which to draw components.

        @type  topleft: tuple(float, float)
        @param topleft: (x, y) position of top left corner of this
            layout on the canvas

        @type  width: number
        @param width: width of area in which to draw, in canvas
            units. If C{None}, use the width of the canvas.

        @type  height: number
        @param height: height of area in which to draw, in canvas
            units. If C{None}, use the height of the canvas.

        @type call_depth: int
        @param call_depth: nesting level of layout call. Useful for
            debugging.

        @rtype: tuple(list, tuple, bool)
        @return: double of a list of this node's children (as
            L{LayoutNodes}) and the actual part of the box in which
            the layout drew. (This will probably be smaller than the
            allowable space due to padding.) The bounding box is a
            tuple C{(topleft, bottomright)}, where the top left of the
            box begins at C{topleft} the bottom right ends I{just
            before} C{bottomright}. (In other words, C{bottomright}
            isn't included in the bounding box.) The final element is
            C{True} if this node is a leaf node (i.e., it contains no
            other C{Layout} objects), otherwise C{False}.
        """
        try:
            if width is not None:
                width = width + 0
            if height is not None:
                height = height + 0
        except:
            raise RayonException("width and height should be numbers or None "
                                 "(got '%s' and '%s')" %
                                 (type(width), type(height)))
        # xXx debug
        #print "%s%s(ctx, %s, %s, %s, %d)" % (
        #        " " * call_depth, type(self).__name__, topleft,
        #        width, height, call_depth)
        
        # Draw a rectangle around the layout in debug mode
        if DEBUG:
            if width is not None and height is not None:
                ctx.draw_rectangle(0, 0, width, -height,
                                   1, (255, 0, 0), (0, 0, 0, 10), fixup=False)

