#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Classes for drawing graphic primitives on a canvas relative to an
(x,y) coordinate on the canvas.

Classes in this module may be invoked in two ways. They can be
referenced and instantiated directly, or they may be instantiated via
the L{create} function, which consults a mapping of string labels to
classes.
"""
from __future__ import division
import math
from base import Stamp
import miscutils

class Marker(Stamp):
    """
    Base class for objects which draw marks at a particular Cartesian
    coordinate position on the canvas.
    """
    def draw(self, ctx, x, y):
        """
        Draw the marker. The Marker should take care to leave the
        state of the GraphicsContext and DC object as they were when
        passed in.
        """
        pass
    def dimensions(self):
        """
        Describe width and height of marked area.

        @return: C{(w, h)}, where C{w} is the width of the bounding box
            and C{h} is the height."""
        pass
    def bbox(self, x, y):
        """ Describe bounding box of marker for given C{(x, y)}
        coordinate.

        @type x: number
        @param x: position on X axis
        @type y: number
        @param y: position on Y axis

        @return: A tuple C{(x1, y1, x2, y2)} containint the x and y
            coordinates of the upper left and lower right points
            respectively of the box bounding the marker."""
        pass


class Dot(Marker):
    """
    Draws a filled circle. The origin is the center of the circle.


    *radius* is the radius of the circle in device units
    (points/pixels).

    *color* is a color specification (see :ref:`layout-color-specs`)
    of the dot.
    """
    @classmethod
    def generic(cls, color, size):
        return cls(size, color)
    
    def __init__(self, radius=.5, color=(0,0,0)):
        self.color  = miscutils.parse_color(color)
        self.radius = radius
    def draw(self, ctx, x, y):
        ctx.draw_ellipse(x, y, self.radius, self.radius,
                         stroke_color=self.color, fill_color=self.color)
    def dimensions(self):
        return (self.radius * 2, self.radius * 2)
    def bbox(self, x, y):
        return (x - self.radius, y - self.radius,
                x + self.radius, y + self.radius)

class Circle(Marker):
    """
    Draws a circle. The origin of the mark is the center of the
    circle.

    *radius* is the radius of the circle in device units
    (points/pixels).

    *stroke_color* is the color of the edge of the
    circle.

    *fill_color* is the color of the circle's interior.

    *stroke_width* is the width (in points or pixels) of the edge of
    the circle.
    """
    @classmethod
    def generic(cls, color, size):
        return cls(size, color)
    
    def __init__(self, radius=.5, stroke_color=(0,0,0), fill_color=(0,0,0,0),
                 stroke_width=1):
        self.radius = radius
        self.stroke_color  = miscutils.parse_color(stroke_color)
        self.fill_color = miscutils.parse_color(fill_color)
        self.stroke_width = stroke_width
    def draw(self, ctx, x, y):
        ctx.draw_ellipse(x, y, self.radius, self.radius,
                         stroke_color=self.stroke_color,
                         fill_color=self.fill_color,
                         stroke_width=self.stroke_width)
    def dimensions(self):
        return (self.radius * 2, self.radius * 2)
    def bbox(self, x, y):
        return (x - self.radius, y - self.radius,
                x + self.radius, y + self.radius)




class VerticalLine(Marker):
    """
    Draws a vertical line segment. The origin of the mark is a point
    along the line. (The exact point depends on the values of the
    *below* and *above* parameters.

    *below* is the length of the line segment below the origin, in
    points/pixels.

    *above* is the length of the line segment above the origin, in
    points/pixels.

    *color* is a color specification (see :ref:`layout-color-specs`)
    of both line segments.

    *width* is the width of both line segments in points/pixels.
    """
    @classmethod
    def generic(cls, color, size):
        return cls(size, size, color, 1)
    
    def __init__(self, below=5, above=5, color=(0,0,0), width=1):
        self.below = below
        self.above = above
        self.color = miscutils.parse_color(color)
        self.width = width
    def draw(self, ctx, x, y):
        ctx.draw_line(x, y - self.above, x, y + self.below, self.color,
                      self.width, True)
    def dimensions(self):
        # TODO: something reasonable about width
        return 1, self.below + self.above + 1
    def bbox(self, x, y):
        # TODO: something reasonable about width
        return x, y - self.below, x, y + self.above

class HorizontalLine(Marker):
    """
    Draws a horizontal line segment. The origin of the mark is a point
    along the line. (The exact point depends on the values of the
    *before* and *after* parameters.)

    *before* is the length of the line segment to the left of the
    origin, in points/pixels.

    *after* is the length of the line segment to the right of the
    origin, in points/pixels.

    *color* is a color specification (see :ref:`layout-color-specs`)
    of both line segments.

    *width* is the width of both line segments in points/pixels.
    """
    @classmethod
    def generic(cls, color, size):
        return cls(size, size, color, 1)

    def __init__(self, before=5, after=5, color=(0,0,0), width=1):
        self.before = before
        self.after  = after
        self.color  = miscutils.parse_color(color)
        self.width  = width
    def draw(self, ctx, x, y):
        ctx.draw_line(x - self.before, y, x + self.after, y, self.color,
                      self.width, fixup=True)
    def dimensions(self):
        # TODO: something reasonable about width
        return 1, self.below + self.above + 1
    def bbox(self, x, y):
        # TODO: something reasonable about width
        return x - self.before, y, x + self.after, y



class Cross(Marker):
    """
    Draws a cross. (i.e., An equally-sized horizontal line and a
    vertical line crossing at their respective midpoints.) The origin
    of the mark is the point at which the vertical and horizontal line
    segments intersect.

    *below* is the length of the vertical line segment below the
    origin, in points/pixels.

    *above* is the length of the vertical line segment above the
    origin, in points/pixels.

    *before* is the length of the horizontal line segment to the left
    of the origin, in points/pixels.

    *after* is the length of the horizontal line segment to the right
    of the origin, in points/pixels.

    *color* is a color specification (see :ref:`layout-color-specs`)
    of both the horizontal and vertical line segments.

    *width* is the width of both the horizontal and vertical line
    segments in points/pixels.
    """
    @classmethod
    def generic(cls, color, size):
        return cls(size, size, size, size, color)

    def __init__(
        self, below=5, above=5, before=5, after=5, color=(0,0,0)):
        self.below  = below
        self.above  = above
        self.before = before
        self.after  = after
        self.color  = miscutils.parse_color(color)
    def draw(self, ctx, x, y):
        ctx.draw_lines([[x, y - self.below, x, y + self.above],
                        [x - self.before, y, x + self.after, y]],
                       color=self.color, width=1)
    def dimensions(self):
        return self.before + self.after + 1, self.above + self.below + 1
    def bbox(self, x, y):
        return x - self.before, y - self.above, x + self.after, y + self.below

class NullMarker(Marker):
    """
    Makes no mark. This can be used as the *marker* argument a
    :class:`LabeledMarker`, which can in turn be passed to something
    which expects a :class:`LabeledMarker`, if the intent is to only
    draw a label.
    """
    @classmethod
    def generic(cls, color, size):
        return cls()
    def dimensions(self):
        return 0, 0
    def bbox(self, x, y):
        return x, y, x, y

# Labeled markers

class LabeledMarker(object):
    # Documented in classref-text.rst
    angles = {
        "e"  : 0,
        "ne" : 45,
        "n"  : 90,
        "nw" : 135,
        "w"  : 180,
        "sw" : 225,
        "s"  : 270,
        "se" : 315,
    }

    def __init__(self, marker, labeler, label_position="n", label_spacing=5):
        self.marker = marker
        self.labeler = labeler
        self.label_position = label_position
        self.label_spacing = label_spacing

    def get_label_bounding_box(self, ctx, x, y, txt):
        """Get the bounding box for just the marker label.

        @rtype: tuple
        @return: the top left, bottom left, top right and bottom right
            corners of the marker label's bounding box, respectively."""
        # lbl_x and lbl_y are relative to x and y. Also,
        # xy_from_rtheta assumes that the Y axis goes up, when, in
        # cairo, it goes down.
        rel_lbl_x, rel_lbl_y = miscutils.xy_from_rtheta(
            self.label_spacing, self.angles[self.label_position])
        abs_x = x + rel_lbl_x
        abs_y = y - rel_lbl_y
        bbox = self.labeler.get_bounding_box(ctx, txt)
        return tuple((bbox_x + abs_x, bbox_y + abs_y)
                     for bbox_x, bbox_y in bbox[:-1])

    def get_label_dimensions(self, ctx, x, y, txt):
        """Get the bounding box for just the marker label, in terms of
        width, height and the top left corner.

        @rtype: tuple @return the top left corner, width and height of
            the marker label's bounding box."""
        tl, bl, tr, br = self.get_label_bounding_box(ctx, x, y, txt)
        w = tr[0] - tl[0]
        h = br[1] - tr[1]
        return tl, w, h
        
        
    def draw(self, ctx, x, y, width, height, txt):
        """ Draw the marker and label.

        @type ctx: backend_*.Context
        @param ctx: context into which to draw
        
        @type x: int
        @param x: x-position of label.
        
        @type y: int
        @param y: y-position of label

        @type  width: float
        @param width: width of enclosing component (e.g., a
            L{borders.Border}).

        @type  height: float
        @param height: height of enclosing component (e.g., a
            L{borders.Border}).

        @type txt: string
        @type txt: Contents of the label
        """
        # lbl_x and lbl_y are relative to x and y. Also,
        # xy_from_rtheta assumes that the Y axis goes up, when, in
        # cairo, it goes down.
        lbl_x, lbl_y = miscutils.xy_from_rtheta(
            self.label_spacing, self.angles[self.label_position])
        if self.labeler is not None:
            self.labeler.draw(ctx, x + lbl_x, y - lbl_y, width, height, txt)
        if self.marker is not None:
            self.marker.draw(ctx, x, y)

    
# Functional interface


_marker_classes = {
    'dot'   : Dot,
    'circle': Circle, 
    'vline' : VerticalLine,
    'hline' : HorizontalLine,
    'cross' : Cross,
    'none'  : NullMarker
}

def create(marker_style, **marker_params):
    try:
        return _marker_classes[marker_style](**marker_params)
    except KeyError, e:
        raise SyntaxError("No such marker style %s" % str(e))


def create_generic(marker_style, color, size):
    try:
        return _marker_classes[marker_style].generic(color, size)
    except KeyError, e:
        raise SyntaxError("No such marker style %s" % str(e))
    

def list_styles():
    return sorted(_marker_classes.keys())

def list_marker_types():
    return tuple((k, v) for k, v in _marker_classes.items())
