#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division

import math
import bisect
import re

import datautils
import netsa.data.times

class State(object):
    """A generic class for passing properties around."""
    def __init__(self, **params): self.__dict__ = params

def merge_with_defaults(args, defaults):
    """ Create a unified list of parameters from user-specified and
        default options """
    merged = dict(defaults)
    merged.update(args)
    return datautils.Parameters(**merged)



# class PlotPadding(object):
#     """ Padding around the drawable area of the plot. """
#     def __init__(self, left, right, top, bottom):
#         self._padding = (left, right, top, bottom)
#     def left(self): return self._padding[0]
#     def right(self): return self._padding[1]
#     def top(self): return self._padding[2]
#     def bottom(self): return self._padding[3]
#     def vertical(self): return self._padding[0] + self._padding[1]
#     def horizontal(self): return self._padding[2] + self._padding[3]




## Nice Numbers ##########################################################

# Making these available from rayon because netsa-python isn't an
# "officially" released API, but these functions are very useful when
# doing data visualization.
from netsa.data.nice import *
    
## Gradients #############################################################


def gradient(start, end, numsteps):
    """Provide a smooth gradient between n-dimensional point start and
    n-dimensional point end, in numsteps steps."""

    # rgb is an n-element array (n == number of dimensions). Each
    # element contains the start and step values for that
    # dimension. (Usually this is an rgb value, hence the name, though
    # it might be an rgba value, or anything, really.)
    rgb = [(start[x], (end[x] - start[x]) / (numsteps - 1))
           for x in xrange(len(start))]
    return [[dstart + (dstep * i) for dstart, dstep in rgb]
            for i in xrange(numsteps)]



def gradient_colors(start, end, numsteps):
    """Provide a gradient between the start and end colors in numsteps steps.

    @type start: tuple of ints between 0-255
    @param start: rgb (or rgba) values for start color
    @type end: tuple of ints between 0-255
    @param end: rgb (or rgba) values for end color
    @type numsteps: int
    @param numsteps: how many steps to transition between start and end
                    color (including start and end)"""
    
    # Scale the color to a value between 0 and 1, then rescale back to
    # values between 0 and 255 in hopes that it won't be too blocky.
    start = [x/255 for x in start]
    end   = [x/255 for x in end]

    def int_round(n):
        if (n % 255) < 128:
            return int(math.floor(n))
        else:
            return int(math.ceil(n))

    return [[int_round(x * 255) for x in color]
            for color in gradient(start, end, numsteps)]


    


## Data-covering ranges ##################################################

def get_cover_range(stats, include_zero=False):
    """Provide a data range that covers the minimum and maximum values
    in I{stats}, and is anchored at zero if the data is either all
    positive (or 0) or all negative (or 0).

    @type  stats: C{Statistics}-like object
    @param stats: object representing a data range. Must support
        C{min()} and C{max()} methods. C{Statistics} object meets this
        need. So does a C{data.Column} object.

    @rtype: tuple
    @return: a tuple C{(lo, hi)} containing the low and high values of
        the range.
    """
    if not include_zero:
        return stats.min(), stats.max()
    
    if stats.min < 0: lo = stats.min()
    else:             lo = 0
    if stats.max < 0: hi = 0
    else:             hi = stats.max()

    return lo, hi

## Munging tick data #####################################################

def tick_data_iter(data):
    def issequence(x):
        try:
            len(x)
            return True
        except:
            return False
    for tick, label in data:
        if label is None:
            if type(tick) == type(""):
                label = tick
            elif issequence(tick):
                label = tuple(str(_) for _ in tick)
            else:
                label = str(tick)
        yield tick, label
    


## Parsing ###############################################################

re_number = re.compile(r"""
    ^ -?
    [0-9]+
 (\.[0-9]*)? $
""", re.VERBOSE)

def infer_type_from_string(s):
    # Guess what type to use for a given input. Number-like items get
    # float, date-like items get datetime, everything else gets str
    def str_is_datetime(s):
        # It would be more efficient to add a function to
        # netsa.data.times that checks if a string is in the right
        # format. But the performance hit is probably minimal.
        try:
            netsa.data.times.make_datetime(s)
            return True
        except ValueError:
            return False
    rc = None
    if re_number.match(s):
        if s.find(".") != -1:
            rc = float
        else:
            rc = int
    elif str_is_datetime(s):
        rc = netsa.data.times.make_datetime
    else:
        rc = str
    return rc


# Cribbed from http://www.w3.org/TR/css3-color/#svg-color
color_lookup = {
    'aliceblue': (240,248,255,255),
    'antiquewhite': (250,235,215,255),
    'aqua': (0,255,255,255),
    'aquamarine': (127,255,212,255),
    'azure': (240,255,255,255),
    'beige': (245,245,220,255),
    'bisque': (255,228,196,255),
    'black': (0,0,0,255),
    'blanchedalmond': (255,235,205,255),
    'blue': (0,0,255,255),
    'blueviolet': (138,43,226,255),
    'brown': (165,42,42,255),
    'burlywood': (222,184,135,255),
    'cadetblue': (95,158,160,255),
    'chartreuse': (127,255,0,255),
    'chocolate': (210,105,30,255),
    'coral': (255,127,80,255),
    'cornflowerblue': (100,149,237,255),
    'cornsilk': (255,248,220,255),
    'crimson': (220,20,60,255),
    'cyan': (0,255,255,255),
    'darkblue': (0,0,139,255),
    'darkcyan': (0,139,139,255),
    'darkgoldenrod': (184,134,11,255),
    'darkgray': (169,169,169,255),
    'darkgreen': (0,100,0,255),
    'darkgrey': (169,169,169,255),
    'darkkhaki': (189,183,107,255),
    'darkmagenta': (139,0,139,255),
    'darkolivegreen': (85,107,47,255),
    'darkorange': (255,140,0,255),
    'darkorchid': (153,50,204,255),
    'darkred': (139,0,0,255),
    'darksalmon': (233,150,122,255),
    'darkseagreen': (143,188,143,255),
    'darkslateblue': (72,61,139,255),
    'darkslategray': (47,79,79,255),
    'darkslategrey': (47,79,79,255),
    'darkturquoise': (0,206,209,255),
    'darkviolet': (148,0,211,255),
    'deeppink': (255,20,147,255),
    'deepskyblue': (0,191,255,255),
    'dimgray': (105,105,105,255),
    'dimgrey': (105,105,105,255),
    'dodgerblue': (30,144,255,255),
    'firebrick': (178,34,34,255),
    'floralwhite': (255,250,240,255),
    'forestgreen': (34,139,34,255),
    'fuchsia': (255,0,255,255),
    'gainsboro': (220,220,220,255),
    'ghostwhite': (248,248,255,255),
    'gold': (255,215,0,255),
    'goldenrod': (218,165,32,255),
    'gray': (128,128,128,255),
    'green': (0,128,0,255),
    'greenyellow': (173,255,47,255),
    'grey': (128,128,128,255),
    'honeydew': (240,255,240,255),
    'hotpink': (255,105,180,255),
    'indianred': (205,92,92,255),
    'indigo': (75,0,130,255),
    'ivory': (255,255,240,255),
    'khaki': (240,230,140,255),
    'lavender': (230,230,250,255),
    'lavenderblush': (255,240,245,255),
    'lawngreen': (124,252,0,255),
    'lemonchiffon': (255,250,205,255),
    'lightblue': (173,216,230,255),
    'lightcoral': (240,128,128,255),
    'lightcyan': (224,255,255,255),
    'lightgoldenrodyellow': (250,250,210,255),
    'lightgray': (211,211,211,255),
    'lightgreen': (144,238,144,255),
    'lightgrey': (211,211,211,255),
    'lightpink': (255,182,193,255),
    'lightsalmon': (255,160,122,255),
    'lightseagreen': (32,178,170,255),
    'lightskyblue': (135,206,250,255),
    'lightslategray': (119,136,153,255),
    'lightslategrey': (119,136,153,255),
    'lightsteelblue': (176,196,222,255),
    'lightyellow': (255,255,224,255),
    'lime': (0,255,0,255),
    'limegreen': (50,205,50,255),
    'linen': (250,240,230,255),
    'magenta': (255,0,255,255),
    'maroon': (128,0,0,255),
    'mediumaquamarine': (102,205,170,255),
    'mediumblue': (0,0,205,255),
    'mediumorchid': (186,85,211,255),
    'mediumpurple': (147,112,219,255),
    'mediumseagreen': (60,179,113,255),
    'mediumslateblue': (123,104,238,255),
    'mediumspringgreen': (0,250,154,255),
    'mediumturquoise': (72,209,204,255),
    'mediumvioletred': (199,21,133,255),
    'midnightblue': (25,25,112,255),
    'mintcream': (245,255,250,255),
    'mistyrose': (255,228,225,255),
    'moccasin': (255,228,181,255),
    'navajowhite': (255,222,173,255),
    'navy': (0,0,128,255),
    'oldlace': (253,245,230,255),
    'olive': (128,128,0,255),
    'olivedrab': (107,142,35,255),
    'orange': (255,165,0,255),
    'orangered': (255,69,0,255),
    'orchid': (218,112,214,255),
    'palegoldenrod': (238,232,170,255),
    'palegreen': (152,251,152,255),
    'paleturquoise': (175,238,238,255),
    'palevioletred': (219,112,147,255),
    'papayawhip': (255,239,213,255),
    'peachpuff': (255,218,185,255),
    'peru': (205,133,63,255),
    'pink': (255,192,203,255),
    'plum': (221,160,221,255),
    'powderblue': (176,224,230,255),
    'purple': (128,0,128,255),
    'red': (255,0,0,255),
    'rosybrown': (188,143,143,255),
    'royalblue': (65,105,225,255),
    'saddlebrown': (139,69,19,255),
    'salmon': (250,128,114,255),
    'sandybrown': (244,164,96,255),
    'seagreen': (46,139,87,255),
    'seashell': (255,245,238,255),
    'sienna': (160,82,45,255),
    'silver': (192,192,192,255),
    'skyblue': (135,206,235,255),
    'slateblue': (106,90,205,255),
    'slategray': (112,128,144,255),
    'slategrey': (112,128,144,255),
    'snow': (255,250,250,255),
    'springgreen': (0,255,127,255),
    'steelblue': (70,130,180,255),
    'tan': (210,180,140,255),
    'teal': (0,128,128,255),
    'thistle': (216,191,216,255),
    'tomato': (255,99,71,255),
    'turquoise': (64,224,208,255),
    'violet': (238,130,238,255),
    'wheat': (245,222,179,255),
    'white': (255,255,255,255),
    'whitesmoke': (245,245,245,255),
    'yellow': (255,255,0,255),
    'yellowgreen': (154,205,50,255),
}

class InvalidColorSpec(Exception):
    """
    Exception raised when an attempt is made to parse an improper
    color specification.
    """
    pass

def parse_color_string(color):
    """Convert a string color specification into an RGBA tuple.

    The syntax of the color spec is identical to that specified by
    L{http://www.w3.org/TR/CSS2/syndata.html#color-units} and
    L{http://www.w3.org/TR/css3-color}, with the exception
    that alpha channels are supported in hexadecimal notation as well
    as functional notation.

    As an example, here are all the valid ways to specify full red
    (100% red, 0% green, 0% blue, 100% alpha):

       * red
       * #f00
       * #f00f
       * #ff0000
       * #ff0000ff
       * #F00
       * #F00F
       * #FF0000
       * #FF0000FF
       * rgb(255, 0, 0)
       * rgba(255, 0, 0, 1)
       * rgb(100%, 0%, 0%)
       * rgba(100%, 0%, 0%, 1)
    """
    # All quotes are from the CSS2 spec, unless otherwise
    # noted. Editorial comments are in square brackets:
    def parse_rgb_color_spec(spec_items):
        num_pct = 0
        parsed_items = []
        for si in spec_items:
            if si.endswith("%"):
                num_pct += 1
                si = float(si[:-1])
                parsed_items.append(int(255 * (si / 100)))
            else:
                parsed_items.append(int(si))
        if num_pct not in (0, 3):
            raise InvalidColorSpec(
                "All or none of spec should be expressed in percentages")
        # "Values outside the device gamut should be clipped or mapped
        # into the gamut when the gamut is known: the red, green, and
        # blue values must be changed to fall within the range
        # supported by the device."
        clipped_items = []
        for pi in parsed_items:
            if pi < 0: pi = 0
            if pi > 255: pi = 255
            clipped_items.append(pi)
        return tuple(clipped_items)
    orig_color = color
    try:
        if color_lookup.has_key(color.lower()):
            return color_lookup[color.lower()]
        if color.startswith("#"):
            # "The format of an RGB value in hexadecimal notation is a
            # '#' immediately followed by either three or six
            # hexadecimal characters." [We allow 4- and 8-character
            # sequences, too, for alpha, even though it's explicitly
            # ignored in the spec.]
            color = color[1:]
            if len(color) == 3 or len(color) == 4:
                return tuple(int(i + i, 16) for i in list(color))
            elif len(color) == 6 or len(color) == 8:
                return tuple(int(color[i:j], 16) for i, j in
                        zip(xrange(0, len(color), 2),
                            xrange(2, len(color) + 2, 2)))
            else:
                raise InvalidColorSpec(
                    "Color spec should have 3, 4, 6, or 8 characters")
        else:
            if color.startswith('rgb(') and color.endswith(')'):
                # "The format of an RGB value in the functional
                # notation is 'rgb(' followed by a comma-separated
                # list of three numerical values (either three integer
                # values or three percentage values) followed by ')'."
                color = color[4:-1]
                rgb = [c.strip() for c in color.split(",")]
                return parse_rgb_color_spec(rgb)
            elif color.startswith('rgba(') and color.endswith(')'):
                # The format of an RGBA value in the functional
                # notation is 'rgba(' followed by a comma-separated
                # list of three numerical values (either three integer
                # values or three percentage values), followed by an
                # <alphavalue>, followed by ')'.
                #
                # "[An <alphavalue> is] the uniform opacity setting to
                # be applied across an entire object. Any values
                # outside the range 0.0 (fully transparent) to 1.0
                # (fully opaque) will be clamped to this range."
                # [http://www.w3.org/TR/css3-color/#alphavaluedt]
                color = color[5:-1]
                rgba = [c.strip() for c in color.split(",")]
                rgb = parse_rgb_color_spec(rgba[:-1])
                # [Alpha must be clamped, too]
                alpha = float(rgba[-1])
                if alpha > 1: alpha = 1
                if alpha < 0: alpha = 0
                # [Convert to 0-255 range]
                alpha = int(alpha * 255)
                
                return rgb + (alpha,)
            else:
                raise InvalidColorSpec("Can't parse")
            return tuple(out_values)
    except InvalidColorSpec, e:
        raise ValueError("Invalid color specification '%s': %s" % (
                orig_color, str(e)))


def parse_color(color):
    if type(color) == str:
        color = parse_color_string(color)
    if len(color) == 3:
        return color + (255,)
    else:
        return tuple(color)

## Polar/Cartesian Conversion ############################################

def x_from_rtheta(r, theta):
    """Derive the cartesian C{x} coordinate from the polar coordinate
    represented by I{r} and I{theta}.

    @type  r: number
    @param r: radial coordinate

    @type  theta: number
    @param theta: angular coordinate, in degrees

    @rtype: float
    @return: x coordinate
    """
    theta = math.radians(theta)
    x = r * math.cos(theta)


def y_from_rtheta(r, theta):
    """Derive the Cartesian C{x} coordinate from the polar coordinate
    represented by I{r} and I{theta}

    @type  r: number
    @param r: radial coordinate

    @type  theta: number
    @param theta: angular coordinate, in degrees

    @rtype: float
    @return: y coordinate
    """
    theta = math.radians(theta)
    return r * math.sin(theta)


def xy_from_rtheta(r, theta):
    """Derive the Cartesian C{x} coordinate from the polar coordinate
    represented by I{r} and I{theta}

    @type  r: number
    @param r: radial coordinate

    @type  theta: number
    @param theta: angular coordinate, in degrees

    @rtype: C{(float, float)} tuple
    @return: x and y coordinates
    """
    theta = math.radians(theta)
    return (r * math.cos(theta)), (r * math.sin(theta))


def r_from_xy(x, y):
    """Derive the radial polar coordinate from the Cartesian
    coordinate represented by I{x} and I{y}.

    Note that this function assumes the C{y} axis goes up, which is
    the opposite of the native canvas coordinate systems of, say,
    Cairo and WxWidgets.

    @type  x: number
    @param x: x coordinate

    @type  x: number
    @param x: x coordinate

    @rtype: number
    @return: radial coordinate
    """
    return math.sqrt(x**2 + y**2)


def theta_from_xy(x, y):
    """Derive the radial and angular polar coordinates from the
    Cartesian coordinate represented by I{x} and I{y}.

    Note that this function assumes the C{y} axis goes up, which is
    the opposite of the native canvas coordinate systems of, say,
    Cairo and WxWidgets.

    @type  x: number
    @param x: x coordinate

    @type  x: number
    @param x: x coordinate

    @rtype: number
    @return: angular coordinate in degrees
    """
    # No more efficiency to be gained from implementing it separately,
    # so:
    return rtheta_from_xy(x, y)[1]

def rtheta_from_xy(x, y):
    """Derive the radial and angular polar coordinates from the
    Cartesian coordinate represented by I{x} and I{y}.

    Note that this function assumes the C{y} axis goes up, which is
    the opposite of the native canvas coordinate systems of, say,
    Cairo and WxWidgets.

    @type  x: number
    @param x: x coordinate

    @type  x: number
    @param x: x coordinate

    @rtype: C{(number, number)} tuple
    @return: radial coordinate and angular coordinate in degrees
    """
    if x == 0 and y == 0:
        return 0, 0

    r = r_from_xy(x, y)
    prototheta = math.asin(y/r)
    if x < 0:
        theta = math.degrees((prototheta * -1) + math.pi)
    else:
        theta = math.degrees(prototheta)
    # Normalize so theta is always positive
    if theta < 0:
        theta = 180 + (theta * -1)
    return r, theta

