#  Copyright (C) 2011 by Carnegie Mellon University.
#
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#
#  NO WARRANTY
#
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE,
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT,
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES.
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE
#  DELIVERABLES UNDER THIS LICENSE.
#
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie
#  Mellon University, its trustees, officers, employees, and agents from
#  all claims or demands made against them (and any related losses,
#  expenses, or attorney's fees) arising out of, or relating to Licensee's
#  and/or its sub licensees' negligent use or willful misuse of or
#  negligent conduct or willful misconduct regarding the Software,
#  facilities, or other rights or assistance granted by Carnegie Mellon
#  University under this License, including, but not limited to, any
#  claims of product liability, personal injury, death, damage to
#  property, or violation of any laws or regulations.
#
#  Carnegie Mellon University Software Engineering Institute authored
#  documents are sponsored by the U.S. Department of Defense under
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains
#  copyrights in all material produced under this contract. The U.S.
#  Government retains a non-exclusive, royalty-free license to publish or
#  reproduce these documents, or allow others to do so, for U.S.
#  Government purposes only pursuant to the copyright license under the
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import base
import borders
import markers
import text
import ticks
import lines
import events
import data
import scales
import miscutils

from rayon import RayonException

import datetime
import sys
import itertools

# Plots

NO_VALUE = object()

def IDENTITY_SCALE():
    return lambda x: x

def string_scale():
    # String conversion for a Rayon scale. None == "", anything that
    # can't be converted to str == "".
    def to_string(x):
        if x is None: return ""
        try:
            return str(x)
        except:
            return ""
    return to_string

class Plot(base.Panel):
    """
    Abstract base class for classes that provide primary data
    visualization in Rayon.

    *axis_args* are parameters whose names take the form
    ``axis_spec``, where ``axis`` is one of the axis labels
    specified by a subclass of :class:`Plot`, and ``spec`` is one of
    the literals ``scale`` or ``data``. The value of the argument will
    be the value of that axis's scale or data, as appropriate. As an
    example, the value of ``foo_data`` will be the
    :class:`~rayon.data.Column` object used as data for visualizing
    the ``foo`` axis; ``foo_scale`` will be the
    :class:`~rayon.scales.Scale` object used to transform the data
    values into the coordinate space of the visualization.

    It is an error to specify data or a scale for an axis that is not
    one of the :class:`Plot` subclass's named axes.
    """
    axes = None
    """
    A tuple of strings labels specifying the axes this plot
    visualizes. All subclasses must populate this value.
    """
    default_scales = None
    """
    A dictionary whose keys are axis labels named in :data:`axes`
    and whose values are either no-arg constructors or factory
    functions for scales (either :class:`~rayon.scales.Scale` objects
    or functions) that can be used as scales for the given axis if no
    scale is provided by the user.
    """
    def __init__(self, **kargs):
        base.Panel.__init__(self)
        if self.__class__.__name__ == "Plot":
            raise RayonException(
               "Plot class should not be instantiated")
        if self.axes is  None:
               raise RayonException(
                   "'%s' does not define an 'axes' attribute" %
                   self.__class__.__name__)
        self.options = {}
        self.scales  = {}
        self.data    = {}
        self.metadata = {}

        # Process the arguments
        axis_data = {}
        axis_scales = {}
        for k, v in kargs.items():
            underbar = k.rfind("_")
            axis, scale_or_data = k[:underbar], k[underbar+1:]
            if axis not in self.axes:
                raise RayonException("Invalid axis '%s'" % axis)
            if scale_or_data == 'scale':
                axis_scales[axis] = v
            elif scale_or_data == 'data':
                # Don't add if value is none
                if v is not None:
                    # Take iterables of all sorts by converting them
                    # to lists. (We take lists as column data now,
                    # lists are cool.)
                    if isinstance(v, data.Column):
                        axis_data[axis] = v
                    else:
                        axis_data[axis] = list(v)
                        
            else:
                raise TypeError("Plot.__init__() got an "
                                "unexpected keyword argument '%s'" % k)
        # Add default scales for axes that don't have them
        def mk_no_arg_constructor(non_callable):
            def no_arg_constructor():
                return non_callable
            return no_arg_constructor
        for axis, default_scale in self.default_scales.items():
            if not axis_scales.has_key(axis):
                if not callable(default_scale):
                    default_scale = mk_no_arg_constructor(default_scale)
                axis_scales[axis] = default_scale()
        # Set the data and scales
        self.set_data(**axis_data)
        self.set_scales(**axis_scales)

    def set_options(self, **kargs):
        self.options.update(kargs)

    def get_default_scale(self, axis):
        try:
            candidate = self.default_scales[axis]
        except KeyError:
            return None
        if callable(candidate):
            x = candidate()
            return x
        x = lambda _: candidate
        return x

    def get_scale(self, axis):
        """
        Returns the scale associated with *axis*, or ``None`` if no
        scale is (yet) associated with this axis.
        """
        if axis not in self.axes:
            raise RayonException("'%s' not an axis in '%s'" %
                                 (axis, self.__class__.__name__))
        return self.scales.get(axis, None)

    def get_data(self, axis):
        """
        Returns the :class:`~rayon.data.Column` object associated with
        *axis*, or ``None`` if no data is (yet) associated with this
        axis.
        """
        if axis not in self.axes:
            raise RayonException("'%s' not an axis in '%s'" %
                                 (axis, self.__class__.__name__))
        return self.data.get(axis, None)

    def _set_scale(self, axis, scale):
        if axis not in self.axes:
            raise RayonException("'%s' not an axis in '%s'" %
                                 (axis, self.__class__.__name__))
        new_scales = dict(self.scales)
        # If scale isn't callable, make a callable that returns scale
        # as a value.
        if not callable(scale):
            # Can't just do:
            #  scale = lambda x: scale
            # Because that would just return whatever scale's bound to
            # (in this case, a lambda that returns scale)
            value = scale
            scale = lambda x: value
        # Detach the data from from the old scale
        d = self.get_data(axis)
        if new_scales.has_key(axis):
            if d is not None:
                if hasattr(new_scales[axis], 'detach'):
                    new_scales[axis].detach(self, axis)
        # Attach the data to the new scale
        if d is not None:
            if hasattr(scale, 'attach'):
                scale.attach(self.data[axis], self, axis)
        # Add the new scale
        new_scales[axis] = scale
        self.scales = new_scales

    def set_scales(self, **kwargs):
        """
        Associates :class:`~rayon.scales.Scale` objects or functions
        with axes in the :class:`Plot` object. *axis_scales* are named
        parameters whose names are axis labels specified in
        :data:`axes`, and whose values are scales (or functions) mapping
        values in the data coordinate space to values in the plot
        coordinate space.
        """
        for axis, scale in kwargs.items():
            self._set_scale(axis, scale)


    def _set_datum(self, axis, datum):
        new_data = dict(self.data)
        if axis not in self.axes:
            raise RayonException("'%s' not an axis in '%s'" %
                                 (axis, self.__class__.__name__))
        # Detach the old data from the scale
        s = self.get_scale(axis)
        if s is not None:
            if hasattr(s, 'detach'):
                s.detach(self, axis)
        # Attach the new data to the scale
        if s is not None:
            if hasattr(s, 'attach'):
                s.attach(datum, self, axis)
        new_data[axis] = datum
        self.data = new_data

    def set_data(self, **kwargs):
        """
        Associates :class:`~rayon.data.Column` objects with axes in
        the :class:`Plot` object. *axis*data* are named parameters
        whose names are axis labels specified in :data:`axes`, and
        whose values are :class:`~rayon.data.Column` objects
        containing the data to be plotted on this axis.
        """
        for axis, datum in kwargs.items():
            if not isinstance(datum, data.Column):
                datum = data.Column(datum)
            self._set_datum(axis, datum)

    def set_metadata(self, **kwargs):
        # Do meta-data differently from data for now. It just seems
        # like people may want to use this incrementally rather than
        # all at once. I may revisit this decision.
        self.metadata.update(kwargs)

    def get_metadata(self):
        # Accessor to the metadata
        return dict(self.metadata)

    def draw(self, ctx, width, height):
        self.draw_(ctx, width, height)

    def get_scaled_value(self, axis, val):
        """
        Returns the value of *val* scaled by the scale for *axis*.
        """
        scale = self.get_scale(axis)
        if scale is None:
            raise RayonException("No scale set for '%s' axis" % axis)
        return scale(val)

    def scaled_value(self, axis, val):
        # Deprecated -- don't use
        return self.get_scaled_value(axis, val)

    def get_scaled_minimum(self, axis):
        """
        Returns the minimum plottable value on the scale associated
        with *axis*. If the scale for *axis* is not a
        :class:`~rayon.scales.Scale` object, returns the scaled value
        of the minimum value in the data associated with *axis*.
        """
        s = self.get_scale(axis)
        if s is None:
            raise RayonException("No scale set for '%s' axis" % axis)
        try:
            return s.get_output_min()
        except AttributeError:
            # It's probably a function. Use the minimum of the data as
            # a proxy.
            d = self.get_data(axis)
            if d is None:
                raise RayonException("Can't get scaled minimum for axis '%s' "
                                     "if scale is a function and data doesn't "
                                     "exit" % axis)
            return self.get_scaled_value(axis, d.min())

    def get_scaled_maximum(self, axis):
        """
        Returns the maximum plottable value on the scale associated
        with *axis*. If the scale for *axis* is not a
        :class:`~rayon.scales.Scale` object, returns the scaled value
        of the maximum value in the data associated with *axis*.
        """
        s = self.get_scale(axis)
        if s is None:
            raise RayonException("Scale doesn't exist for '%s'" % axis)
        try:
            return s.get_output_max()
        except AttributeError:
            # It's probably a function. Use the maximum of the data as
            # a proxy.
            d = self.get_data(axis)
            if d is None:
                raise RayonException("Can't get scaled minimum for axis '%s' "
                                     "if scale is a function and data doesn't "
                                     "exit" % axis)
            return self.get_scaled_value(axis, d.max())



    def _orient_iteration(self, reverse):
        datalen = len(self.data.values()[0])
        if reverse:
            return reversed(xrange(datalen))
        else:
            return xrange(datalen)

    def _validate_axes(self):
        # Ensure there's a scale for every axis
        axes = set(self.axes)
        axes_with_scales = set(self.scales.keys())
        axes_without_scales = axes - axes_with_scales
        if len(axes_without_scales) != 0:
            raise RayonException("Unspecified scales for '%s': '%s'" %
                                 (self.__class__.__name__,
                                  ", ".join(list(axes_without_scales))))

    def get_scaled_points(self, reverse=False):
        """
        Returns an iterator over the data in the plot. Each item in
        the iterator is a tuple containing one element of data from
        each axis, scaled by the scale for that axis.

        If *reverse* is ``True``, start iterating from the end of the
        set of data and go over it in reverse. Otherwise, iteration
        will start from the beginning and proceed until the end.
        """
        self._validate_axes()
        def scaled_value_by_idx(axis, idx):
            # You can have a scale with no data; it must be a constant
            # scale, and it must be able to cope with being passed a
            # value of NO_VALUE.
            if not self.data.has_key(axis):
                return self.scales[axis](NO_VALUE)
            return self.get_scaled_value(axis, self.data[axis][idx])
        iterator = self._orient_iteration(reverse)
        for idx in iterator:
            yield [scaled_value_by_idx(a, idx) for a in self.axes]

    def get_unscaled_points(self, reverse=False):
        """
        Returns an iterator over the data in the plot. Each item in
        the iterator is a tuple containing one element of data from
        each axis. Unlike :meth:`get_scaled_points`, this method
        returns the unscaled, original data.

        If an axis has no data, the value :data:`NO_VALUE` will be
        passed back rather than ``None``, as ``None`` could be a valid
        data value.

        If *reverse* is ``True``, start iterating from the end of the
        set of data and go over it in reverse. Otherwise, iteration
        will start from the beginning and proceed until the end.
        """
        def unscaled_value_by_idx(axis, idx):
            # You can have a scale with no data; as such, absent data
            # is marked as such, rather than causing errors.
            if not self.data.has_key(axis):
                return NO_VALUE
            try:
                return self.data[axis][idx]
            except IndexError:
                return NO_VALUE
        iterator = self._orient_iteration(reverse)
        for idx in iterator:
            yield [unscaled_value_by_idx(a, idx) for a in self.axes]

    def get_metadata_points(self, reverse=False):
        # iterator over metadata. Currently undocumented -- not sure
        # I'm going to keep it, or how.
        iterator = self._orient_iteration(reverse)
        # Note that we're yielding a dictionary, not a tuple, unlike
        # the other methods in this genre.
        for idx in iterator:
            # We don't store metadata on a per-point basis, so
            # iteration over points is -- erm, pointless. For now, at
            # least. However, I'm leaving it in because I think I will
            # make it per-point sometime soon.
            yield dict((k, self.metadata[k])
                       for k in self.metadata.keys())


    def get_all_points(self, reverse=False):
        """
        Returns an iterator over the data in the plot. Each item
        yielded by the iterator is a double of tuples. The first tuple
        contains the scaled values of a single element of data in each
        axis. The second tuple contains the unscaled values. This
        method is equivalent to calling
        ``itertools.izip(plot.get_scaled_points(reverse),
        plot.get_unscaled_points(reverse))``

        If *reverse* is ``True``, start iterating from the end of the
        set of data and go over it in reverse. Otherwise, iteration
        will start from the beginning and proceed until the end.
        """
        return itertools.izip(self.get_scaled_points(reverse),
                              self.get_unscaled_points(reverse))


    def get_all_points_and_metadata(self, reverse=False):
        # get_all_points, for metadata. Currently undocumented, for
        # the same reason get_metadata_points is undocumented.
        return itertools.izip(self.get_scaled_points(reverse),
                              self.get_unscaled_points(reverse),
                              self.get_metadata_points(reverse))




    def components_from_xy(self, x, y):
        return [self]


    # To be implemented by subclasses

    def draw_(self, ctx, width, height):
        """
        Draws the plot onto the rendering surface. Subclasses must
        implement this.

        *ctx* is a :class:``~rayon.backend.Context``
        object. (See :ref:`rendering`.)

        *width* is the width of the drawing region in device units
        (points or pixels).

        *height* is the width of the drawing region in device units
        (points or pixels).
        """

        # Debugging - draw box around plot area
        #ctx.draw_rectangle(0, 0, width, -height, 1, (0,0,0), (0,0,0,0))
        raise NotImplementedError("draw_ must be implemented by subclass")



#
# Documented plot
#

# This is a hack. The resolve_class method in the toolbox gets
# confused with respect to new_plot and new_gridlines, because
# Gridlines objects are Plots. The answer is to declare that
# new_gridlines return Gridlines objects, and new_plot
# returns...something that Gridlines doesn't inherit. That something
# is the DocumentedPlot.
#
# TODO: To fix this, the functionality in plot should probably be
# refactored into one or more functions/internal classes, which are
# shared by both the plotting and grids code. We will no longer be
# inheriting for implementation, so we can change the type inheritance
# (so Gridlines are no longer Plots).
    
class DocumentedPlot(Plot):
    pass


#
# Scatter plots
#


class ScatterPlot(DocumentedPlot):
    """
    .. figure:: images/plot-scatterplot-example-1.*
        :align: center

        Example of a scatterplot.

    Draws a plot in which data are mapped onto the X and Y axes of a
    Cartesian coordinate system. At these coordinates, marks are made
    whose shape, color and size can also be determined based on data.

    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|

    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y``","A real number between 0 and 1","Y (vertical)) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``marker_color``","A color specification (see :ref:`specifying-colors`)","Color of the mark representing the point","``#000000FF``"
       "``marker_shape``","A marker shape ","Shape of the mark representing the point","``dot``"
       "``marker_size``","Integer size (in points/pixels)","Size of the mark representing the point","``1``"
    """
    axes = ('x', 'y', 'marker_color', 'marker_shape', 'marker_size')
    default_scales=dict(x=scales.LinearScale,
                        y=scales.LinearScale,
                        marker_color=(0, 0, 0),
                        marker_shape="dot",
                        marker_size=1)
    def __init__(self, **kargs):
        Plot.__init__(self, **kargs)
        self.marker_cache = {}
        self.selection_start = None
        self.selection_end = None

    def draw_(self, ctx, width, height):
        for x, y, color, shape, size in self.get_scaled_points():
            color = miscutils.parse_color(color)
            if not self.marker_cache.has_key((color, shape, size)):
                self.marker_cache[color, shape, size] = (
                    markers.create_generic(shape, color=color, size=size))
            marker = self.marker_cache[color, shape, size]
            x = x * width
            y = height - (y * height)
            marker.draw(ctx, x, y)

    def set_selection(self, start_tuple, end_tuple):
        """
        Make note of the starting and ending C{(x,y)} coordinates in a
        rectangular selection on the scatter plot.

        @type  start: tuple
        @param start: (x,y) coordinates of one corner of selection
            rectangle, in device units.

        @type  end: tuple
        @param end: (x,y) coordinates of corner of selection rectangle
            opposite to C{start}
        """
        self.selection_start, self.selection_end = start_tuple, end_tuple
    def get_selection(self):
        return self.selection_start, self.selection_end


class ScatterPlotListener(events.EventListener):
    def __init__(self, scatterplot):
        events.EventListener.__init__(self, scatterplot)
        self.drag_start = None
    def on_mouse_left_down_(self, target, evt):
        self.drag_start = evt.get_position()
    def on_mouse_left_up_(self, target, evt):
        if self.drag_start != evt.get_position():
            target.set_selection(self.drag_start, evt.position())
            self.on_points_selected_(target, evt)
        self.drag_start = None


class LabeledMarkerCache(object):
    def __init__(self):
        self.cache = {}
    def get(self, label_key, marker_key, label_position, label_spacing):
        key = (label_key, marker_key, label_position, label_spacing)
        if not self.cache.has_key(key):
            color, shape, size = marker_key
            labeler = text.Labeler(**dict(label_key))
            marker = markers.create_generic(shape, color=color, size=size)
            self.cache[key] = markers.LabeledMarker(marker, labeler,
                                                    label_position,
                                                    label_spacing)
        return self.cache[key]


class LabeledScatterPlot(DocumentedPlot):
    """
    .. figure:: images/plot-labeledscatterplot-example-1.*
        :align: center

        Example of a labeled scatterplot.

    Draws a plot similar to that drawn by a
    :ry:factory:`~rayon.plots.ScatterPlot` plot, but the markers may
    also be labeled.

    By default, label data is converted to a string using ``str``. If
    the label value is ``None`` or the input value cannot be converted
    to a string, the output value is an empty string (meaning no label
    is drawn).

    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|


    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y``","A real number between 0 and 1","Y (vertical) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``marker_color``","A color specification","Color of the mark representing the point","``#000000FF``"
       "``marker_shape``","A marker shape","Shape of the mark representing the point","``dot``"
       "``marker_size``","Integer size (in points/pixels)","Size of the mark representing the point","``1``"
       "``text``","String","Text of the label accompanying the marker","See description"
       "``label_position``","``n``, ``s``, ``e``, or ``w`` (See :ref:`ref-labeledmarkers`)","Position of the label accompanying the marker, relative to the marker","``ne``"
       "``label_spacing``","Integer size (in points/pixels)","Space between label and marker, in points/pixels","``1``"
       "``font_properties``","dictionary","Dictionary of properties of the label text","``{}`` (Empty dictionary)"
       "``border_properties``","dictionary","Dictionary of properties of the label border","``{}`` (Empty dictionary)"
       "``bgcolor``","A color specification (see :ref:`specifying-colors`)","Background color of the label box","``#00000000``"
    """
    axes = ('x', 'y', 'marker_color', 'marker_shape', 'marker_size',
            'text', 'label_position', 'label_spacing',
            'font_properties', 'border_properties', 'bgcolor')
    default_scales = dict(x=scales.LinearScale, y=scales.LinearScale,
                          marker_color=(0, 0, 0),
                          marker_shape="dot",
                          marker_size=1,
                          label_position="ne", label_spacing=1,
                          font_properties={}, border_properties={},
                          bgcolor=(0, 0, 0, 0),
                          text=string_scale)
    def __init__(self, **kargs):
        Plot.__init__(self, **kargs)
        self.label_cache = LabeledMarkerCache()
        self.selection_start = None
        self.selection_end = None
    def set_data(self, **kwargs):
        Plot.set_data(self, **kwargs)
    def draw_(self, ctx, width, height):
        for (x, y, color, shape, size, txt, label_position, label_spacing,
             font_properties, border_properties,
             bgcolor) in self.get_scaled_points():
            scaled_x = x * width
            scaled_y = height - (y * height)

            color = miscutils.parse_color(color)
            label_properties = dict()
            label_properties.update(font_properties)
            label_properties.update(border_properties)
            label_properties['bgcolor'] = bgcolor
            label_key = tuple(sorted(
                (k, v) for k, v in label_properties.items()))
            marker_key = (color, shape, size)

            labeled_marker = self.label_cache.get(label_key, marker_key,
                                                  label_position,
                                                  label_spacing)

            if (self.options.get('avoid_vertical_overflow', False) or
                self.options.get('avoid_horizontal_overflow', False)):
                # Check if the marker, when drawn, will overflow one
                # of the edges of the box. If so, change the alignment
                # accordingly. (The change may not fix the problem,
                # but we leave it in place anyway.)

                # TODO: use whole bounding box, not just the bounding
                # box of the label.

                # Normalize label position so:
                #  1. it's always two characters
                #  2. position one is always "n", "s" or " "
                #  3. position two is always "e", "w" or " "
                lpos = list(label_position)
                if lpos[0] in "ew":
                    lpos.insert(0, " ")
                if lpos[-1] in "ns":
                    lpos.append(" ")

                lbl_topleft, lbl_w, lbl_h = labeled_marker.get_label_dimensions(
                    ctx, scaled_x, scaled_y, txt)

                if self.options.get('avoid_horizontal_overflow', False):
                    if lbl_topleft[0] < 0:
                        # left clipping
                        lpos[1] = "e"
                        font_properties['halign'] = 'left'
                    elif lbl_topleft[0] + lbl_w > width:
                        # right clipping
                        lpos[1] = "w"
                        font_properties['halign'] = 'right'
                if self.options.get('avoid_vertical_overflow', False):
                    if lbl_topleft[1] < 0:
                        # top clipping
                        lpos[0] = "s"
                        font_properties['valign'] = 'bottom'
                    elif lbl_topleft[1] + lbl_h > height:
                        # bottom clipping
                        lpos[0] = "n"
                        font_properties['valign'] = 'top'

                new_label_position = (lpos[0] + lpos[1]).strip()

                new_label_properties = dict(label_properties)
                new_label_properties.update(font_properties)
                new_label_key = tuple(sorted(
                    (k, v) for k, v in new_label_properties.items()))

                if new_label_position != label_position:
                    labeled_marker = self.label_cache.get(new_label_key,
                                                          marker_key,
                                                          new_label_position,
                                                          label_spacing)

            labeled_marker.draw(ctx, scaled_x, scaled_y, width, height, txt)


    def set_selection(self, start_tuple, end_tuple):
        """
        Make note of the starting and ending C{(x,y)} coordinates in a
        rectangular selection on the scatter plot.

        @type  start: tuple
        @param start: (x,y) coordinates of one corner of selection
            rectangle, in device units.

        @type  end: tuple
        @param end: (x,y) coordinates of corner of selection rectangle
            opposite to C{start}
        """
        self.selection_start, self.selection_end = start_tuple, end_tuple
    def get_selection(self):
        return self.selection_start, self.selection_end




#
# Line and field plots
#


class LinePlot(DocumentedPlot):
    """

    .. figure:: images/plot-lineplot-example-1.*
        :align: center

        Example of a line plot.

    Draws a series of Cartesian points as a "line." (More accurately,
    as a connected chain of line segments. The points represent the
    endpoints of segments in the chain.)

    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|

    
    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale."
       "``y``","A real number between 0 and 1","Y (vertical)) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale."
       "``line_color``","A color specification (see :ref:`specifying-colors`)","Color of the line segment drawn between the last point and this one","``#000000FF``"
       "``line_width``","Integer size (in points/pixels)","Width of line segment ending in this point","``2``"
       "``line_style``","A line style (see :ref:`line-styles`)","Style of line segment ending in this point","``solid``"
    """
    axes = ('x', 'y', 'line_color', 'line_width', 'line_style')
    default_scales = dict(x=scales.LinearScale, y=scales.LinearScale,
                          line_color=(0, 0, 0),
                          line_width=2,
                          line_style="solid")
    def __init__(self, **kargs):
        Plot.__init__(self, **kargs)
        self.line_cache = {}
    def draw_(self, ctx, width, height):
        row_iter = self.get_scaled_points()
        from_x, from_y, _, _, _ = row_iter.next()
        from_x = from_x * width
        from_y = height - (from_y * height)
        for to_x, to_y, color, lwidth, style in row_iter:
            color = miscutils.parse_color(color)
            if not self.line_cache.has_key((color, style, lwidth)):
                self.line_cache[color, style, lwidth] = (
                    lines.create_generic(style, color=color, width=lwidth))
            line = self.line_cache[color, style, lwidth]
            to_x = to_x * width
            to_y = height - (to_y * height)
            line.draw(ctx, from_x, from_y, to_x, to_y)
            from_x = to_x
            from_y = to_y


class FieldPlot(DocumentedPlot):
    """

    .. figure:: images/plot-fieldplot-example-1.*
        :align: center

        Example of a field plot.

    Draws a visualization similar to that drawn by a
    :ry:factory:`~rayon.plots.LinePlot` plot. However, a
    :ry:factory:`~rayon.plots.FieldPlot` plot draws two lines. Each
    line shares values for the X axis, but has different values for
    Y. The area between the two lines (the "field") is shaded.

    *fill_color* is a color specification (see
    :ref:`specifying-colors`) of the color of the shading.


    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|


    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y1``","A real number between 0 and 1","The Y (vertical)) axis for the first line","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y2``","A real number between 0 and 1","The Y (vertical)) axis for the second line","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``line_color``","A color specification (see :ref:`specifying-colors`)","Color of the line segment drawn between the last point and this one. The same line color is used for the line segment drawn for the ``y1`` data and the one drawn for the ``y2`` data.","``#0000007F``"
       "``line_width``","Integer size (in points/pixels)","Width of line segment ending in this point. The same line width is used for the line segment drawn for the ``y1`` data and the one drawn for the ``y2`` data.","``2``"
       "``line_style``","A line style (see :ref:`line-styles`)","Style of line segment ending in this point","``dashed``"
    """
    axes = ('x', 'y1', 'y2', 'line_color', 'line_width', 'line_style')
    default_scales = dict(x=scales.LinearScale,
                          y1=scales.LinearScale,
                          y2=scales.LinearScale,
                          line_color=(0, 0, 0, 127),
                          line_width=2,
                          line_style="dashed")
    def __init__(self, fill_color, **kargs):
        Plot.__init__(self, **kargs)
        self.fill_color = miscutils.parse_color(fill_color)
        self.line_cache = {}
    def draw_(self, ctx, width, height):
        # No, we're not currently supporting line color/width/style on
        # a per-point basis, even though I'm demanding it in the
        # interface. But we'll probably want it later, and it would be
        # nice to avoid making a backward-incompatible interface
        # change later, so they're axes, even though they do nothing..
        line_color = self.get_scaled_value('line_color', 0)
        line_width = self.get_scaled_value('line_width', 0)
        def poly_vert_generator():
            for to_x, to_y1, _, _, _, _ in self.get_scaled_points():
                yield (to_x * width), height - (to_y1 * height)
            for to_x, _, to_y2, _, _, _ in self.get_scaled_points(True):
                yield (to_x * width), height - (to_y2 * height)
        def line_segment_generator(reverse):
            row_iter =  self.get_scaled_points(reverse)
            fromrow = row_iter.next()
            for torow in row_iter:
                (from_x, from_y1, from_y2,
                 from_lcolor, from_lwidth, from_lstyle) = fromrow
                (to_x, to_y1, to_y2,
                 to_lcolor, to_lwidth, to_lstyle) = torow
                # TODO: test this
                if reverse:
                    lcolor = miscutils.parse_color(to_lcolor)
                    lwidth = to_lwidth
                    lstyle = to_lstyle
                else:
                    lcolor = miscutils.parse_color(from_lcolor)
                    lwidth = from_lwidth
                    lstyle = from_lstyle
                if not self.line_cache.has_key(
                    (lcolor, lstyle, lwidth)):
                    self.line_cache[lcolor, lstyle, lwidth] = (
                        lines.create_generic(to_lstyle,
                                             color=lcolor,
                                             width=lwidth))
                line = self.line_cache[lcolor, lstyle, lwidth]
                # Use the "top" line if it's the forward pass, and the
                # "bottom" line if it's the reverse pass
                if not reverse:
                    yield (from_x * width, height - (from_y1 * height),
                           to_x * width  , height - (to_y1 * height), line)
                else:
                    yield (from_x * width, height - (from_y2 * height),
                           to_x * width  , height - (to_y2 * height), line)
                fromrow = torow


        # Draw the polygon using a transparent stroke
        ctx.draw_polygon(poly_vert_generator(), line_width,
                         (0, 0, 0, 0),
                         miscutils.parse_color(self.fill_color))

        # Draw the top line
        for from_x, from_y, to_x, to_y, line in line_segment_generator(False):
            line.draw(ctx, from_x, from_y, to_x, to_y)

        # Draw the bottom line
        for from_x, from_y, to_x, to_y, line in line_segment_generator(True):
            line.draw(ctx, from_x, from_y, to_x, to_y)



class FilledLinePlot(DocumentedPlot):
    """

    .. figure:: images/plot-filledlineplot-example-1.*
        :align: center

        Example of a filled line plot.

    Draws a line based on data specified for the X and Y axes in the
    style of a :ry:factory:`LinePlot` plot. The area
    "beneath" the line (that is, in the direction of smaller values on
    the scale), is shaded. See also the :ry:factory:`FieldPlot` plot.

    *fill_color* is a color specification (see
    :ref:`specifying-colors`) of the shading beneath the
    line.


    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|

    
    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y``","A real number between 0 and 1","Y (vertical)) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``line_color``","A color specification (see :ref:`specifying-colors`)","Color of the line segment drawn between the last point and this one","``#0000007F``"
       "``line_width``","Integer size (in points/pixels)","Width of line segment ending in this point","``2``"
       "``line_style``","A line style (see :ref:`line-styles`)","Style of line segment ending in this point","``dashed``"
    """
    axes = ('x', 'y', 'line_color', 'line_width', 'line_style')
    default_scales = dict(x=scales.LinearScale, y=scales.LinearScale,
                          line_color=(0, 0, 0, 127),
                          line_width=2,
                          line_style="dashed")
    def __init__(self, fill_color, fill_top=False, **kargs):
        Plot.__init__(self, **kargs)
        self.fill_color = miscutils.parse_color(fill_color)
        self.fill_top = fill_top
        self.line_cache = {}
    def draw_(self, ctx, width, height):
        # No, we're not currently supporting line color/width/style on
        # a per-point basis, even though I'm demanding it in the
        # interface. I'll probably want it later, and I'd like to
        # avoid making a backward-incompatible interface change later.
        line_color = self.get_scaled_value('line_color', 0)
        line_width = self.get_scaled_value('line_width', 0)
        row_iter = self.get_scaled_points()
        def point_generator():
            pgen = self.get_scaled_points()
            first_x, first_y, _, _, _ = pgen.next()
            max_y = self.get_scaled_maximum('y')
            min_y = self.get_scaled_minimum('y')
            if self.fill_top:
                # Put first point at top
                yield first_x, max_y
            else:
                # Put first point at bottom
                yield first_x, min_y
            yield first_x, first_y
            last_x, last_y = (None, None)
            for row in self.get_scaled_points():
                last_x, last_y, _, _, _ = row
                yield last_x, last_y
            if self.fill_top:
                # Put last point at top
                yield last_x, max_y
            else:
                # Put last point at bottom
                yield last_x, min_y


        def vert_generator():
            first_x = None
            for to_x, to_y in point_generator():
                if first_x is None:
                    first_x = to_x
                yield (to_x * width), height - (to_y * height)
        # Use a transparent stroke...
        ctx.draw_polygon(vert_generator(), line_width, (0, 0, 0, 0),
                         self.fill_color)
        # Then draw the line independently. That limits the lines to
        # the data part of the plot (not the bottom part of the
        # polygon) and lets us treat line color, width, etc. on a
        # per-segment basis.
        row_iter = self.get_scaled_points()
        from_x, from_y, _, _, _ = row_iter.next()
        from_x = from_x * width
        from_y = height - (from_y * height)
        for to_x, to_y, color, lwidth, style in row_iter:
            color = miscutils.parse_color(color)
            if not self.line_cache.has_key((color, style, lwidth)):
                self.line_cache[color, style, lwidth] = (
                    lines.create_generic(style, color=color, width=lwidth))
            line = self.line_cache[color, style, lwidth]
            to_x = to_x * width
            to_y = height - (to_y * height)
            line.draw(ctx, from_x, from_y, to_x, to_y)
            from_x = to_x
            from_y = to_y


#
# Barplots
#


class Bar(object):
    def __init__(self, vertices, xdata, ydata, extra_data=None):
        xs = [_[0] for _ in vertices]
        ys = [_[1] for _ in vertices]

        self.xmin = min(xs)
        self.ymin = min(ys)
        self.xmax = max(xs)
        self.ymax = max(ys)
        self.xdata = xdata
        self.ydata = ydata
        self.extra_data = extra_data
    def overlaps(self, other):
        xmin, xmax, ymin, ymax = other
        return ((xmin <= self.xmax and xmax >= self.xmin) and
                (ymin <= self.ymax and ymax >= self.ymin))

    def __str__(self):
        return "<Bar: X=%s (%s-%s), Y=%s (%s-%s), Extra=%s>" % (
            self.xdata, self.xmin, self.xmax,
            self.ydata, self.ymin, self.ymax,
            self.extra_data)

class VerticalBarPlot(DocumentedPlot):
    """

    .. figure:: images/plot-barplot-example-1.*
        :align: center

        Example of a bar plot.

    Draws a number of vertical bars, with the number and height of
    each bar determined by the data. Data can also be mapped onto the
    bars' color, height, width and border color/width.

    .. seealso:: :ry:factory:`HorizontalBarPlot` 

    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|
    
    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.CategoricalRangeScale` scale"
       "``y``","A real number between 0 and 1","Y (vertical)) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``bar_alignment``","A real number between 0 and 1. 0 == full left, 1 == full right.","Alignment of bar in allotted region.","``.5``"
       "``bar_color``","A color specification (see :ref:`specifying-colors`)","Interior color of the bar.","``#0000FFC0``"
       "``border_color``","A color specification (see :ref:`specifying-colors`)","The border color of the bar.","``#000000FF``"
       "``bar_width``","A real number between 0 and 1","The proportion of the range specified by ``x`` that should be filled with the bar. 0 == no width; 1 == full range.","``1``"
       "``border_width``","Integer size (in points/pixels)","Size of the border.","``0``"
       "``bar_fixup``", "``True`` or ``False``","If ``True``, align bar edges on pixels","``True``"
    """
    axes = ('x', 'y', 'bar_alignment', 'bar_color', 'border_color', 'bar_width',
            'border_width', 'bar_fixup')
    default_scales = dict(x=scales.CategoricalRangeScale, y=scales.LinearScale,
                          bar_alignment=.5, bar_color=(0, 0, 255, 192),
                          border_color=(0,0,0), bar_width=1, border_width=0,
                          bar_fixup=True)
    def __init__(self, **kargs):
        Plot.__init__(self, **kargs)
        self.barlist = []
        self.last_width = None
        self.last_height = None
    def bars_at(self, x, y):
        return filter(lambda b: b.overlaps((x, x, y, y)), self.barlist)
    def draw_(self, ctx, width, height):
        if width != self.last_width or height != self.last_height:
            self.width = width
            self.height = height
            recalc_dims = True
        if recalc_dims and ctx.interactive:
            self.barlist = []

        # Functions to convert to device units from user
        # units. Because the Y-axis is inverted between device and
        # user space, one must do relative Y units differently from
        # absolute Y units.
        def dx_from_ux(ux):     return ux * width
        def dy_rel_from_uy(uy): return uy * height
        def dy_from_uy(uy):     return height - (uy * height)


        y_origin = dy_from_uy(self.get_scaled_minimum('y'))

        for scaled, unscaled, metadata in self.get_all_points_and_metadata():
            (_x, _y, _bar_align, _bar_color, _border_color,
             _bar_width, _border_width, _bar_fixup) = scaled
            (_un_x, _un_y, _un_bar_align, _un_bar_color, _un_border_color,
             _un_bar_width, _un_border_width, _un_bar_fixup) = unscaled
            # y and bar height
            y = dy_from_uy(_y)
            bar_height = y_origin - y
            # x and bar width
            x_lo, x_hi = [dx_from_ux(_) for _ in _x]
            range_width = x_hi - x_lo
            if _bar_width == 1:
                # Leave the bar at full width. The border will overlap
                # with the next and previous bars, but that looks
                # better than having two borders butt up against each
                # other.
                bar_width = range_width
            else:
                # Reduce the width so the stroke is
                # included. Otherwise, the borders of consecutive
                # bars could touch, which is counterintuitive when
                # bar_width < 1.
                bar_width = ((range_width * _bar_width) -
                             (_border_width * 2))
            x_offset = (range_width - bar_width) * _bar_align

            x_start = x_lo + x_offset

            # Adjust bar base so stroke is included (remember that
            # y_origin may be above or below bar_height -- negative
            # numbers, etc.)
            if abs(bar_height) < (_border_width * .5):
                # The bar is so small that the stroke on the box is
                # larger than the box height. Philosophical problem:
                # the honorable thing here is probably to reduce the
                # stroke width and represent the data space accurately
                # (honor the data). The aesthetic thing is to keep all
                # the strokes the same (honor the stroke).
                #
                # I'm honoring the stroke for now; data that small
                # will be thought of as "small" most of the time;
                # their exact visual representation doesn't matter so
                # much.
                bar_height = 0

            bar_base = y_origin
            if bar_height >= 0:
                bar_base -= _border_width * .5
            else:
                bar_base -= _border_width * .5



            border_color = miscutils.parse_color(_border_color)
            bar_color = miscutils.parse_color(_bar_color)
            vertices = ctx.draw_rectangle(
                x_start, bar_base,
                bar_width, bar_height,
                _border_width, border_color,
                bar_color, fixup=_bar_fixup)


            if recalc_dims and ctx.interactive:
                self.barlist.append(Bar(vertices, _un_x, _un_y, metadata))

# Synonym for 'VerticalBarPlot' for backwards compatibility
BarPlot = VerticalBarPlot

class HorizontalBarPlot(DocumentedPlot):
    """
    .. figure:: images/plot-horizontalbarplot-example-1.*
        :align: center

        Example of a horizontal bar plot.


    Plots data similarly to the :ry:factory:`VerticalBarPlot` plot, but
    orients the bars horizontally, from top to bottom.

    Note that bar "width" in the :ry:factory:`HorizontalBarPlot` plot
    is the vertical dimension of the bar. The horizontal dimension is
    the bar "height."

    .. seealso:: :ry:factory:`VerticalBarPlot`

    .. tabularcolumns:: |p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|p{.20\\textwidth}|


    .. csv-table:: Axes
       :header: "Axis", "Value", "Rendering", "Default Scale"

       "``x``","A real number between 0 and 1","X (horizontal) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``y``","A real number between 0 and 1","Y (vertical)) spatial axis","A :ry:factory:`~rayon.scales.LinearScale` scale"
       "``bar_alignment``","A real number between 0 and 1. 0 == full left, 1 == full right.","Alignment of bar in allotted region.","``.5``"
       "``bar_color``","A color specification (see :ref:`specifying-colors`)","Interior color of the bar.","``#0000FFC0``"
       "``border_color``","A color specification (see :ref:`specifying-colors`)","The border color of the bar.","``#000000FF``"
       "``bar_width``","A real number between 0 and 1","The proportion of the range specified by ``u`` that should be filled with the bar. 0 == no width; 1 == full range.","``1``"
       "``border_width``","Integer size (in points/pixels)","Size of the border.","``0``"
       "``bar_fixup``","``True`` or ``False``","If ``True``, align bar edges on pixels","``True``"
    """
    axes = ('x', 'y', 'bar_alignment', 'bar_color', 'border_color', 'bar_width',
            'border_width', 'bar_fixup')
    default_scales = dict(x=scales.LinearScale, y=scales.CategoricalRangeScale,
                          bar_alignment=.5, bar_color=(0, 0, 255, 192),
                          border_color=(0,0,0), bar_width=1, border_width=0,
                          bar_fixup=True)
    def __init__(self, **kargs):
        Plot.__init__(self, **kargs)
        self.barlist = []
        self.last_width = None
        self.last_height = None
    def bars_at(self, x, y):
        barlist = filter(lambda b: b.overlaps((x, x, y, y)), self.barlist)
        return barlist
    def draw_(self, ctx, width, height):
        # Note: bar "width" refers to the vertical dimension of a bar
        # in this plot -- the bars are treated as identical to BarPlot
        # bars that have been rotated clockwise 90 degrees. Similarly,
        # bar "height" refers to the bar's horizontal dimension.
        #
        # All other references to width and height refer to horizontal
        # (x) width and vertical (y) height, as is customary
        # throughout this codebase.
        if width != self.last_width or height != self.last_height:
            self.width = width
            self.height = height
            recalc_dims = True
        if recalc_dims and ctx.interactive:
            self.barlist = []

        # Functions to convert to device units from user
        # units. Because the Y-axis is inverted between device and
        # user space, one must do relative Y units differently from
        # absolute Y units.
        def dx_from_ux(ux):     return ux * width
        def dy_rel_from_uy(uy): return uy * height
        def dy_from_uy(uy):     return height - (uy * height)

        x_origin = dx_from_ux(self.get_scaled_minimum('x'))

        for scaled, unscaled, metadata in self.get_all_points_and_metadata():
            (_x, _y, _bar_align, _bar_color, _border_color,
             _bar_width, _border_width, _bar_fixup) = scaled
            (_un_x, _un_y, _un_bar_align, _un_bar_color, _un_border_color,
             _un_bar_width, _un_border_width, _un_bar_fixup) = unscaled
            # x and bar "height"
            x = dx_from_ux(_x)
            bar_height = x - x_origin
            # y and bar "width" (Note: thanks to scale inversion, y_lo
            # > y_hi)
            y_lo, y_hi = [dy_from_uy(_) for _ in _y]
            range_height = y_lo - y_hi
            # This is the same trick we pull in BarPlot.draw -- read
            # there for details.
            if _bar_width == 1:
                bar_width = range_height
            else:
                bar_width = ((range_height * _bar_width) -
                             (_border_width * 2))
            y_offset = (range_height - bar_width) * _bar_align

            y_start = y_lo - y_offset

            # Adjust bar base in same way as BarPlot.draw
            bar_base = x_origin
            if abs(bar_height) < (_border_width * .5):
                bar_height = 0

            if bar_height >=0:
                bar_base += _border_width * .5
            else:
                bar_base -= _border_width * .5

            border_color = miscutils.parse_color(_border_color)
            bar_color = miscutils.parse_color(_bar_color)
            vertices = ctx.draw_rectangle(bar_base, y_start,
                                          bar_height, bar_width,
                                          _border_width, border_color,
                                          bar_color, fixup=_bar_fixup)
            if recalc_dims and ctx.interactive:
                self.barlist.append(Bar(vertices, _un_x, _un_y, metadata))





# Plot creation

_plot_classes = {
    'scatter'        : ScatterPlot,
    'labeledscatter' : LabeledScatterPlot,
    'line'           : LinePlot,
    'field'          : FieldPlot,
    'filledline'     : FilledLinePlot,
    'bar'            : VerticalBarPlot,
    'vbar'           : VerticalBarPlot,
    'hbar'           : HorizontalBarPlot,
}

def create(plot_style, **plot_params):
    try:
        return _plot_classes[plot_style](**plot_params)
    except KeyError, e:
        raise SyntaxError("No such plot %s" % str(e))


def list_plot_types():
    return tuple((k, v) for k, v in _plot_classes.items())
