#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
"""
Utility functions to make writing Rayon command-line tools easier.
"""
import os
import sys
import errno
import optparse
import ConfigParser
import struct
import socket

from pprint import pprint

from rayon.cliutils import mk_tickspec_parser
from rayon.tickspec_parse import parse_tickspec
import rayon.miscutils as miscutils


class ToolRuntimeError(Exception):
    """
    Indicates that the tool identified unacceptable runtime
    behavior. The error message is expected to be
    """
    pass

NO_DEFAULT = object()


def mk_optparse_cb(dst_val, converter):
    def parse_input(option, opt_str, value, parser):
        setattr(parser.values, dst_val, converter(value))
    return parse_input


class Option(object):
    """
    A class representing an application option. This option can be
    supplied either on the command line or in a configuration
    variable.

    *name* is the name of the new option, as it would appear on a
    command line. *name* should not contain leading dashes; they will
    be ignored if specified.

    *varname* is the name of the option as it would appear to the
    program (i.e., when the option is parsed, a legal python symbol
    name.) If no value is specified, the *name* will be used, with any
    dashes converted to underscores.

    *otype* is the datatype of the value. The following types are
    currently supported:

       ``args``
         A comma-separated list of subvalues, converted to a list.

       ``color``
         A textual color specification (see
         :ref:`specifying-colors`). This is converted to a quadruple
         of integers in the range 0-255, representing the RGBA value
         of the color.

       ``float``
         A number in floating-point representation, converted using
         Python's ``float`` constructor.

       ``int``
         A number in integer representation, converted using Python's
         ``int`` constructor.

       ``ip``
         An IP address in dotted quad notation, converted to an
         integer.

       ``str``
         A string. No conversion is performed

       ``ticks``
         A tick specfication (documented in the ``ryscatterplot(1)``
         man page and in the other command man pages that use it),
         which is converted into a series of functions that take the
         data and return a position in the data for the tick mark.
       

    If specified, *default_val* gives the default value the option
    has. If the option has no default value, it will be required on
    the command-line.
    """
    def __init__(self, name, varname=None, default_val=NO_DEFAULT,
                 hlp="", suppress_help=False):
        if name.startswith("--"):
            name = name[2:]
        self._name = name

        if varname is None:
            varname = self._name.replace("-", "_")
        self._varname = varname

        self._default_val = default_val
        self._help = hlp
        self._suppress_help = suppress_help

    metavar_ = "ARG"
    
    def get_optparse_args(self):
        odict = self.get_optparse_dict_()
        odict['dest'] = self._varname
        odict['metavar'] = self.metavar_
        if self._suppress_help:
            odict['help'] = optparse.SUPPRESS_HELP
        else:
            odict['help'] = self._help
        return ["--" + self._name], odict
        
    def get_varname(self)       : return self._varname
    def get_name(self)          : return self._name
    def has_default_value(self) : return self._default_val != NO_DEFAULT
    def get_default_value(self) : return self._default_val

    def convert_(self, value): raise NotImplementedError('convert_')
    def convert(self, value): return self.convert_(value)


class StringOption(Option):
    metavar_ = "STRING"
    def get_optparse_dict_(self):
        return dict()
    def convert_(self, value):
        # Optparse handles strings itself, so this is only ever
        # called to parse values from config files
        return str(value)

class PathOption(StringOption):
    metavar_ = "PATH"

class IntOption(Option):
    metavar_ = "INT"
    def get_optparse_dict_(self):
        return dict(type="int")
    def convert_(self, value):
        # Optparse handles ints itself, so this is only ever
        # called to parse values from config files
        return int(value)
    
class FloatOption(Option):
    metavar_ = "FLOAT"
    def get_optparse_dict_(self):
        return dict(type="float")
    def convert_(self, value):
        # Optparse handles floats itself, so this is only ever
        # called to parse values from config files
        return float(value)
    

class BooleanOption(Option):
    metavar = "BOOL"
    def get_optparse_dict_(self):
        return dict(action="store_true")
    def convert_(self, value):
        # Optparse handles booleans itself, so this is only ever
        # called to parse values from config files
        if value.lower() in ("false", "0"):
            return False
        return True

        
class ChoiceOption(Option):
    metavar_ = "LABEL"
    def __init__(self, name, varname=None, default_val=NO_DEFAULT,
                 hlp="", suppress_help=False, choices=[]):
        Option.__init__(self, name, varname, default_val, hlp, suppress_help)
        if choices is []:
            raise Exception("Option '%s' created with no choices" % name)
        self._choices = choices
        
    def get_optparse_dict_(self):
        return dict(type="choice", choices=self._choices)

    def convert_(self, value):
        # Optparse handles choices itself, so this is only ever
        # called to parse values from config files
        if value not in self._choices:
            raise ValueError("Invalid value '%s': choose from %s" %
                             (value, self._choices))
        return value
        
class CallbackOption(Option):
    def get_optparse_dict_(self):
        cb = mk_optparse_cb(self.get_varname(), self.convert_)
        return dict(action="callback", callback=cb, type="str")

# Deprecated: Use NewTickspecOption
class TickspecOption(CallbackOption):
    metavar_ = "TICKSPEC"
    def convert_(self, value): return mk_tickspec_parser(value)

class NewTickspecOption(CallbackOption):
    metavar_ = "TICKSPEC"
    def convert_(self, value):
        return parse_tickspec(self.get_name(), value)


class ColnameOption(CallbackOption):
    metavar_ = "COLNAME_OR_IDX"
    def convert_(self, value):
        try:
            return int(value)
        except ValueError:
            return str(value)

class ArgsOption(CallbackOption):
    metavar_ = "ARGS"
    def convert_(self, value):
        return dict((k, miscutils.infer_type_from_string(v)(v)) for k, v in
                    [sa.split("=") for sa in value.split(",")])
        
    
class ColorOption(CallbackOption):
    metavar_ = "COLOR"
    def convert_(self, value):
        try:
            return miscutils.parse_color_string(value)
        except ValueError, e:
            try:
                return miscutils.parse_color_string("#" + value)
            except:
                # Raise original error
                raise e
        

class IpOption(Option):
    metavar_ = "INT_OR_IP"
    def convert_(self, value):
        try:
            return int(value)
        except ValueError:
            return struct.unpack("!L", socket.inet_aton(value)[0])




# Default options

default_width  = 400
default_height = 400

# TODO: the sytem_options/common_options distinction may be
# superfluous. Eliminate?
#
# system_options == options controlling configuration parsing
#
# common_options == options unrelated to config parsing that all apps
# should have

system_options = [
    BooleanOption(
        "ignore-config", default_val=False,
        hlp="Ignore any configuration information from configuration files"),
    StringOption(
        "config-file", hlp="File to use for visualization configuration"),
    BooleanOption("print-options", default_val=False, suppress_help=True)
]

common_options = [
    PathOption(
        "input-path", default_val="-",
        hlp="Path to input data. (Default: STDIN)"),
    PathOption(
        "output-path", default_val=NO_DEFAULT,
        hlp="(REQUIRED) Path to output file"),
    IntOption(
        "width", default_val=default_width,
        hlp="image width (pixels for PNG, otherwise points) "
        "(Default: %d)" % default_width),
    IntOption(
        "height", default_val=default_height,
        hlp="image height (pixels for PNG, otherwise points) "
        "(Default: %d)" % default_height),
]

        

#
# Some helper functions        
#

def normpath(p):
    if p is None:
        return None
    return os.path.normpath(os.path.expanduser(p))

def get_sysconfig_filenames():
    cfg_locations = ["/etc/ryrc", "~/.ryrc"]

    ryrc = normpath(os.getenv("RYRC"))
    
    if ryrc is not None and os.path.exists(ryrc):
        cfg_locations.append(ryrc)

    return cfg_locations


def get_config_dictionary(fname, optspec, section_name, ignore_nofile):
    # Parse a section of a config file into a dict. If ignore_nofile
    # == False, error if the file does not exist. Otherwise, return an
    # empty dict.
    cfg = ConfigParser.SafeConfigParser()
    if fname == '-':
        cfg.readfp(sys.stdin)
    else:
        try:
            cfg.readfp(open(normpath(fname)), 'r')
        except IOError, e:
            if e.errno == errno.ENOENT:
                if ignore_nofile:
                    return dict()
                else:
                    raise

    # Add to cfg_dict, applying any data transformations that need to
    # be applied.
    cfg_dict = dict()

    def values_in_section(s):
        if cfg.has_section(s):
            for k in cfg.options(s):
                yield (k, cfg.get(s, k))

    def opt_by_name(n):
        for o in optspec:
            if o.get_varname() == n:
                return o
        return None
                
    for name, value in values_in_section('rydefault'):
        opt = opt_by_name(name)
        cfg_dict[name] = opt.convert(value)

    for name, value in values_in_section(section_name):
        opt = opt_by_name(name)
        if opt is None:
            # Check again with hyphens converted into underscores
            newname = name.replace('-', '_')
            opt = opt_by_name(newname)
            if opt is None:
                # Unrecognized option. Ignore
                continue
            else:
                name = newname
        cfg_dict[name] = opt.convert(value)
    
    return cfg_dict

def get_option_dictionary(options, defaults, desc, args=None):
    if args is None:
        args = sys.argv[1:]

    optparser = optparse.OptionParser(usage="%%prog [options] - %s" % desc)
    optparser.set_defaults(**defaults)
    for opt in options:
        o_args, o_kwargs = opt.get_optparse_args()
        optparser.add_option(*o_args, **o_kwargs)

    # Create our own Values instance to disable optparse's "helpful"
    # default handling
    vals = optparse.Values()
    parsed_options, leftovers = optparser.parse_args(args, values=vals)
    options_dict = vars(parsed_options)

    return options_dict, leftovers


def parse_configuration(command=None, args_list=None,
                        cmd_options=None, include_common_options=True,
                        cmd_desc=None):
    """
    Coalesces configuration information from configuration files, the
    command line and the command defaults. Returns a double; the first
    item is a dictionary of configuration options, the second is a
    list of any additional arguments that were not parsed on the
    command line.

    This function first tries to read the two system configuration
    files ``/etc/ryrc`` or ``~/.ryrc``. If the environment variable
    ``RYRC`` is set, it should point to a configuration file, which
    will be read next. Finally, the system will read any configuration
    file specified on the command line using the ``--config-file``
    switch. The value of this switch may be ``-``, indicating that the
    configuration file will be passed on ``STDIN``.

    If ``ignore_config`` is specified in a configuration file in the
    section given by *command*, configuration information in
    configuration files read previously will be ignored; this lets a
    configuration file reset the configuration to default settings if
    necessary. Similarly, ``--ignore-config`` can be passed as a
    command-line option and no configuration-file information will be
    used, only command defaults and options on the command line.

    *command* is the string name of the command. This string will be
    the name of the section in the configuration file(s) containing
    options for invocations of this command. If not supplied, only
    configuration information from the default sections of the
    configuration file(s) will be returned.

    *args_list* is the list of command-line arguments, formatted in
    the manner of :var:`sys.argv`. This can be used to generate a
    configuration using some option input other than the
    command-line. If not supplied, the value of ``sys.argv[1:]`` will
    be used.

    *cmd_options* is a list of :class:`Option` objects that define the
    options the application wwill accept. If *include_common_options*
    is ``True`` (the default), the following base options will be
    added to this list:

        * ``input-path``
        * ``output-path``
        * ``width``
        * ``height``

    The following options will be added regardless of the status of
    *include_common_options*:

        * ``ignore-config``
        * ``config-file``
        * ``help``

    To see the descriptions of these options, run any of the rytools
    with the ``--help`` switch.

    *cmd_desc* is a string description of the name. If not supplied,
    an empty string will be used.
    """
    # TODO: Extra directives in config files are currently silently
    # ignored. This code should support a --verbose or --lint or -Wall
    # or something that warns/errors when it sees config directives it
    # doesn't understand. (This already happens with command-line
    # options.)
    

    if cmd_options is None: cmd_options = []
    if cmd_desc    is None: cmd_desc    = ""

    # Merge and parse the options
    if include_common_options:
        options = list(set(system_options) |
                       set(common_options) |
                       set(cmd_options))
    else:
        options = cmd_options

    defaults = dict((x.get_varname(), x.get_default_value())
                    for x in options if x.has_default_value())
    
    option_values, leftovers = get_option_dictionary(
        options, defaults, cmd_desc, args_list)


    def update_config(old_values, new_values):
        # Update old_values with new_values. If 'ignore_config' is in
        # new-values, replace the old values instead (but make sure
        # defaults are honored)
        rc = dict()
        if new_values.get('ignore_config', False):
            rc = dict(defaults)
        else:
            rc = dict(old_values)
        rc.update(new_values)
        return rc

    final_values = dict(defaults)

    # Apply config files (or not, as a shortcut, if the options
    # contain "--ignore-config")
    if not option_values.get('ignore_config', False):
        # If these files don't exist, we don't care
        for sysconf_name in get_sysconfig_filenames():
            sysconf_values = get_config_dictionary(
                sysconf_name, options, command, True)
            final_values = update_config(final_values, sysconf_values)

        # If this file doesn't exit, we care
        if option_values.has_key('config_file'):
            userconf_values = get_config_dictionary(
                option_values['config_file'], options, command, False)
            final_values = update_config(final_values, userconf_values)
            
    
    # Apply options (no need to check "ignore_config" since we
    # short-circuit that above)
    final_values.update(option_values)

    return final_values, leftovers



def execute(cmd_name=None, cmd_func=None, cmd_args=None,
            cmd_options=None, include_common_options=True, cmd_desc=None):
    """
    Runs the main loop in an rytools application.

    *cmd_name* is the name of the tool. If nothing is provided,
    ``sys.argv[0]`` will be used.

    *cmd_func* is a function which takes a single argument -- a
    dictionary of command options. It optionally returns an integer
    that will be used as an exit code. (If nothing is returned,
    success is assumed.)

    *cmd_args* is a list to use as the argument list. If not supplied,
    ``sys.argv[1:] will be used.

    *cmd_options* is a list of :class:`Option` objects that define the
     options the application wwill accept. If *include_common_options*
     is ``True`` (the default), the following base options will be
     added to this list:

        * ``input-path``
        * ``output-path``
        * ``width``
        * ``height``

    The following options will be added regardless of the status of
    *include_common_options*:

       * ``ignore-config``
       * ``config-file``
       * ``help``

    To see the descriptions of these options, run any of the rytools
    with the ``--help`` switch.

    *cmd_desc* is a string description of the name. If not supplied,
    an empty string will be used.
    
    """
    if cmd_name is None:
        cmd_name = os.path.basename(sys.argv[0])

    if cmd_func is None:
        print "Not told to do anything. Exiting."
        sys.exit(0)

    # The rest of the args are handled by other functions

    def maybe_enter_debugger(exc_info):
        if os.getenv("RY_TRACE") is not None:
            import traceback
            traceback.print_exception(*exc_info)
        if os.getenv("RY_PDB") is not None:
            print str(e)
            import pdb
            pdb.post_mortem(exc_info[2])

    # Figure out what we're told
    try:
        # Ignoring extra args for now
        opts, _ = parse_configuration(cmd_name, args_list=cmd_args,
                                      cmd_options=cmd_options,
                                      cmd_desc=cmd_desc)
    except Exception, e:
        print "Configuration error: %s" % str(e)
        print "Type '%s --help' for more information on arguments" % cmd_name
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except KeyboardInterrupt:
        print "Application interrupted by user"
        sys.exit(1)


    # Do what we're told
    try:
        if opts.get("print_options", False):
            pprint(opts)
            return
        
        rc = cmd_func(opts)
        if rc is None:
            sys.exit(0)
        else:
            sys.exit(rc)
    except ToolRuntimeError, e:
        print "Error: %s" % str(e)
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except Exception, e:
        print "Error (%s): %s" % (type(e), str(e))
        maybe_enter_debugger(sys.exc_info())
        sys.exit(1)
    except KeyboardInterrupt:
        print "Application interrupted by user"
        sys.exit(1)
        

# Some general utility functions

def istream_from_param(param):
    if param == "-":
        return sys.stdin
    else:
        return open(param, 'r')




__all__ = """
Option
StringOption
PathOption
IntOption
FloatOption
BooleanOption
ChoiceOption
CallbackOption
TickspecOption
NewTickspecOption
ColnameOption
ArgsOption
ColorOption
IpOption
execute
istream_from_param
ToolRuntimeError
""".split()
