#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
from __future__ import division
import time
import math
import calendar
import copy

from rayon import RayonException
import data # To do type-checking for Datasets in update() methods
import miscutils

import netsa.data.nice
import netsa.data.times



#
# Helper classes/functions
#

class RayonScaleException(RayonException):
    """Base class of scale-related exceptions."""
    pass

class RayonLimitException(RayonScaleException):
    """
    Indicates that the user tried to generate an invalid scale, or to
    use the limits of a scale without properly specifying them. If a
    scale is not fixed and no data has been associated with it, a call
    to a scale's ``get_input_min`` or ``get_output_max`` methods will
    raise this exception.
    """
    pass

class RayonRangeException(RayonScaleException):
    """
    Indicates that a user has passed a value that cannot be mapped
    using this scale.
    """
    pass


def cover_range(a_lo, a_hi, b_lo, b_hi):
    # return the smallest possible range that covers both a and b
    # where a = (a_lo, a_hi) and b = (b_lo, b_hi). Any or all
    # arguments can be None, meaning that part of the interval is
    # as-yet undefined.
    #
    # If all arguments are None, this function returns (None, None)

    a = a_lo, a_hi
    b = b_lo, b_hi
    
    # The values of the elements of a and b can be None, so use an
    # alternative definition of nothing for an absent range/element.
    nil = object()
    
    def get_element_or_range(rnge):
        # assumption: rnge != (None, None)
        if rnge[0] is None:
            return rnge[1], nil
        elif rnge[1] is None:
            return rnge[0], nil
        else:
            return nil, rnge

    def rangify_elements(a_elem, b_elem):
        if b_elem >= a_elem:
            return (a_elem, b_elem)
        else:
            return (b_elem, a_elem)

    def rangify_range_and_element(rnge, elem):
        if elem >= rnge[0]:
            if elem <= rnge[1]:
                # |===|
                #   |
                return rnge
            else:
                # |===|
                #       |
                return (rnge[0], elem)
        else:
            #   |===|
            # |
            return (elem, rnge[1])

    def rangify_ranges(a, b):
        if a[0] > a[1]:
            raise RayonLimitException(
                "Couldn't update scale range: %s > %s" % (a[0], a[1]))
        if b[0] > b[1]:
            raise RayonLimitException(
                "Couldn't update scale range: %s > %s" % (b[0], b[1]))
        if b[0] >= a[1]:
            # a |===|
            # b       |---|
            return (a[0], b[1])
        elif a[0] >= b[1]:
            # a       |===|
            # b |---|
            return (b[0], a[1])
        else:
            if a[1] <= b[1]:
                if a[0] >= b[0]:
                    # a    |===|
                    # b |-------|
                    return b
                else:
                    return (a[0], b[1])
            else:
                if a[0] >= b[0]:
                    # a     |=======|
                    # b |-------|
                    return (b[0], a[1])
                else:
                    # a   |=======|
                    # b     |---|
                    return a

    # The main body of the function:

    if a == (None, None):
        return b

    if b == (None, None):
        return a

    a_elem, a_range = get_element_or_range(a)
    b_elem, b_range = get_element_or_range(b)

    if a_range is nil:
        if b_range is nil:
            return rangify_elements(a_elem, b_elem)
        else:
            return rangify_range_and_element(b_range, a_elem)
    else:
        if b_range is nil:
            return rangify_range_and_element(a_range, b_elem)
        else:
            return rangify_ranges(a_range, b_range)


        

class RangeLimits(object):
    # Helps continuous Scales keep track of the limits of their input
    # range.
    def __init__(self, lo=None, hi=None, start_at_zero=True):
        self.start_at_zero = start_at_zero
        self.lo, self.hi = cover_range(None, None, lo, hi)
        self.orig_lo = self.lo
        self.orig_hi = self.hi
    # TODO: Rename accessor methods to get_*
    def update(self, lo, hi):
        self.lo, self.hi = cover_range(self.lo, self.hi, lo, hi)
    def reset(self):
        self.lo, self.hi = self.orig_lo, self.orig_hi
    def limits(self):
        # Note: this method is about computing sensible data limits
        # for visualization. Notably, if the high end of the scale is
        # positive, it will limit the lower bound at zero. If this
        # isn't what you want, use high() and low() to do your own
        # thing.
        
        # Compute a sensible set of limits based on existing range
        # information. If no sensible set of limits can be
        # constructed, raises a RayonLimitException.
        if self.hi is None:
            raise RayonLimitException("No upper bound on scale")
        hi = self.hi
        if self.lo is None:
            if self.hi > 0:
                lo = 0
            else:
                raise RayonLimitException("No lower bound on scale with "
                                 "negative upper bound")
        else:
            lo = self.lo
        return lo, hi
    def spread(self):
        lo, hi = self.limits()
        return hi - lo
    def low(self):
        return self.lo
    def high(self):
        return self.hi
    def is_closed(self):
        # True if the scale has an upper and lower bound
        return self.lo is not None and self.hi is not None

    def __contains__(self, el):
        if self.lo is not None and el < self.lo:
            return False
        if self.hi is not None and el > self.hi:
            return False
        return True

#
# Scales
#

class Scale(object):
    """
    The base class of objects that serve as scales in Rayon. (See
    :ref:`scales` and :ref:`customizing-scale`.) This class should not
    be created directly; however subclasses should call this class's
    ``__init__`` method in their own ``__init__`` method.

    *fixed* should be ``True`` if the scale is fixed -- that is, its
    limits will not grow and/or shrink to accomodate additional
    data. Otherwise, *fixed* should be ``False``.

    If *check_input* is ``True`` the subclass's
    :meth:`~Scale.check_input_` method will be called prior to
    expanding the scale limits with new data.

    *input_min*, *input_max*, *output_min* and *output_max* represent
    the limits on the input and output ranges, respectively. If
    *fixed* is ``True``, these values may be omitted; Rayon will
    determine them automatically. If
    :meth:`~Scale.has_input_lower_bound_` is ``False``, *input_min* may
    be omitted, and if :meth:`~Scale.has_input_upper_bound_` is
    ``False``, *output_max* may be omitted. Similarly, If
    :meth:`~Scale.has_output_lower_bound_` is ``False``, *input_min*
    may be omitted, and if :meth:`~Scale.has_output_upper_bound_` is
    ``False``, *output_max* may be omitted. It is always okay to
    supply all four of these values.
    """
    @classmethod
    def ni_message(cls, meth):
        return "'%s' not implemented in %s" % (meth, cls.__name__)
    def __init__(self, 
                 fixed=False, check_input=False,
                 input_min=None, input_max=None,
                 output_min=None, output_max=None):
        if type(self) == Scale:
            raise RayonException(
                "Cannot instantiate Scale object, only subclasses")
        # The collection of datasets with which we are associated. We
        # will use these to compute our limits, when asked.
        self._datasets_ordered = []
        # If the set of datasets has been changed since the last time
        # we computed our limits, datasets_dirty will be set to
        # True. (This doesn't apply if the dataset is fixed, of
        # course.)
        if fixed:
            self._datasets_dirty = False
        else:
            self._datasets_dirty = True
        self._input_min = input_min
        self._input_max = input_max
        self._output_min = output_min
        self._output_max = output_max
        self._is_fixed = fixed
        self._check_input = check_input
        

    def attach(self, d, owner, label=None):
        # Used internally by Rayon.
        #
        # Called when an object needs to associate a scale with data
        # that will be used to calculate the limits of the scale. d is
        # the data, owner is the putative owner of that data with
        # respect to the scale. label is a label the owner may choose
        # to give the data if it may be associating more than one
        # dataset with the scale.
        if not self._is_fixed:
            if self._check_input:
                self.check_input_(d)
            key = (owner, label)
            for idx, (k, v) in enumerate(self._datasets_ordered):
                if k == key:
                    # Replace with this one
                    self._datasets_ordered[idx] = (key, d)
                    self._datasets_dirty = True
                    return
            # It's a new dataset
            self._datasets_ordered.append((key, d))
            self._datasets_dirty = True

    def detach(self, owner, label=None):
        # Used internally by Rayon.
        #
        # Dissociate a scale with a set of data referenced by owner
        # and label. See attach for more.
        if not self._is_fixed:
            key = (owner, label)
            for idx, (k, v) in enumerate(self._datasets_ordered):
                if k == key:
                    self._datasets_ordered.pop(idx)
            self._datasets_dirty = True

    def is_fixed(self):
        """
        Returns ``True`` if a :class:`Scale` object's acceptable input
        range can dynamically adjust to new data, otherwise ``False``.
        """
        return self._is_fixed

    def _compute_limits(self):
        if self._datasets_dirty:
            imax, imin, omax, omin = self.reset_()
            for k, d in self._datasets_ordered:
                imax, imin, omax, omin = self.update_(d)
            (self._input_min, self._input_max,
             self._output_min, self._output_max) = imax, imin, omax, omin
            self._datasets_dirty = False

    def get_input_min(self):
        """
        Returns the lowest value in the input range.
        """
        if not self.has_input_lower_bound_():
            raise RayonLimitException("Cannot call get_input_min on scale with "
                                      "no lower bound.")
        self._compute_limits()
        return self._input_min
        
    def get_input_max(self):
        """
        Returns the highest value in the input range.
        """
        if not self.has_input_upper_bound_():
            raise RayonLimitException("Cannot call get_input_max on scale with "
                                      "no upper bound.")
        self._compute_limits()
        return self._input_max
        
    def get_output_min(self):
        """
        Returns the lowest value in the output range.
        """
        if not self.has_output_lower_bound_():
            raise RayonLimitException("Cannot call get_output_min on scale "
                                      "with no lower bound.")
        self._compute_limits()
        return self._output_min

    def get_output_max(self):
        """
        Returns the highest value in the output range.
        """
        if not self.has_output_upper_bound_():
            raise RayonLimitException("Cannot call get_output_max on scale "
                                      "with no upper bound.")
        self._compute_limits()
        return self._output_max

    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over a sequence of aesthetically pleasing
        positions in the input range of the scale. These positions
        will be used for for placing tick marks.

        *num_ticks* is the requested number of tick positions. This is
        only a request; :meth:`get_nice_tick_positions` may choose to
        honor it or not as appropriate to the scale.

        If *inside* is ``True``, the lowest and highest points in the
        set of positions will fall inside the minimum and maximum
        values in the input range. If it is ``False``, the lowest and
        highest points in the set must fall on or outside the minimum
        and maximum values in the input range.
        """
        self._compute_limits()
        return self.get_nice_tick_positions_(num_ticks=5, inside=False)

    def __call__(self, a):
        self._compute_limits()
        return self.convert_(a)

    # Handled by the subclass...

    def update_(self, new_data):
        """
        Changes limits of the input range such that all data in
        *new_data* would fall within it. *new_data* is an iterable
        over the set of new data. *new_data* can be any sequence,
        including a :class:`~rayon.data.Column` object. If *new_data*
        is empty, this method should do no additional work.

        This function returns a 4-tuple ``(input_min, input_max,
        output_min, output_max)``, where ``input_min`` and
        ``input_max`` are the lower and upper bounds on input and
        ``output_min`` and ``output_max`` are the lower and upper
        bounds on output. These values can be arbitrary if they will
        be ignored (for example, ``input_max`` can be any value if
        :meth:`Scale.has_input_upper_bound_` returns ``False``.) If
        the :class:`Scale` object's limits have not yet been
        completely specified, ``input_min`` or ``input_max`` may
        be ``None``.
        """
        raise NotImplementedError(self.ni_message("update_"))

    def reset_(self):
        """
        Resets the :class:`Scale` object's limits to those it had when
        it was created.
        
        This function returns a 4-tuple ``(input_min, input_max,
        output_min, output_max)``, where ``input_min`` and
        ``input_max`` are the lower and upper bounds on input and
        ``output_min`` and ``output_max`` are the lower and upper
        bounds on output. These values can be arbitrary if they will
        be ignored (for example, ``input_max`` can be any value if
        :meth:`Scale.has_input_upper_bound_` returns ``False``.)  If
        the :class:`Scale` object's limits have not yet been
        completely specified, ``input_min`` or ``input_max`` may
        be ``None``.

        By default, this method calls :meth:`Scale.update_` with
        no data and returns the result.
        """
        return self.update_([])

    def convert_(self, a):
        """
        Transforms *a* into a value in the output domain. Returns the
        tranformed value. *a* should be an appropriate value for the
        input range; the exact type depends on the subclass.
        """
        raise NotImplementedError(self.ni_message("convert_"))

    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        """
        Implements the described behavior of
        :meth:`Scale.get_nice_tick_positions`.
        
        By default, this method will return an iterator over an empty
        list.
        """
        return iter([])

    def check_input_(self, in_data):
        """
        Checks that the data in *in_data* adheres to the standards set
        for input by the :class:`Scale` subclass. If problems are
        found with the data, this method should raise a
        :exc:`RayonScaleException`.

        By default, this method allows all data; subclasses may
        therefore choose not to implement it.
        """
        pass

    def has_input_upper_bound_(self):
        """
        Returns ``True`` if input to the :class:`Scale` object has an
        upper bound. If this method returns ``False``,
        :meth:`Scale.get_input_min` will raise a
        :exc:`RayonLimitException` when called.

        By default, this method returns ``True``.
        """
        return True

    def has_input_lower_bound_(self):
        """
        Returns ``True`` if input to the :class:`Scale` object has a
        lower bound. If this method returns ``False``,
        :meth:`Scale.get_input_min` will raise a
        :exc:`RayonLimitException` when called.

        By default, this method returns ``True``.
        """
        return True
    
    def has_output_upper_bound_(self):
        """
        Returns ``True`` if output of the :class:`Scale` object has an
        upper bound. If this method returns ``False``,
        :meth:`Scale.get_input_min` will raise a
        :exc:`RayonLimitException` when called.

        By default, this method returns ``True``.
        """
        return True

    def has_output_lower_bound_(self):
        """
        Returns ``True`` if input of the :class:`Scale` object has a
        lower bound. If this method returns ``False``,
        :meth:`Scale.get_input_min` will raise a
        :exc:`RayonLimitException` when called.

        By default, this method returns ``True``.
        """
        return True
    


class RangeScale(Scale):
    """
    Abstract base class for :class:`Scale` objects whose return values
    are ranges -- tuples of numbers representing the lower and upper
    boundaries of a continuous range.
    """
    pass


class CategoricalRangeScale(RangeScale):
    """
    Maps categories to contiguous, non-overlapping sub-ranges in a
    larger range. This scale is a ranged version of the
    :ry:factory:`CategoricalScale` scale. (See
    :ref:`ref-scales-range-scales`)

    *categories* is a sequence of input categories. The elements in
    the sequence may be of any type provided that they are unique and
    can be tested for equality. This parameter may be omitted; in that
    case, the categories will be determined by Rayon based on the data
    the scale is required to transform.

    *output_min* and *output_max* define the boundaries of a numeric
    range. The scale will map categories to contiguous subranges of
    equal size that cover the range. When used as a scale on a spatial
    axis in a plot or border, the default values for *output_min* and
    *output_max* will use all the space allocated for the element
    being visualized.

    If *fixed* is ``True``, the value of *categories* will be the
    final value for the scale; no categories may be added or
    removed. Otherwise, the category list may change if new data is
    associated with the scale.

    If *reverse* is ``False``, categories will be ordered as they were
    added to the scale. Additional categories beyond the initial set
    will be appended in the order they were inserted. If *reverse* is
    ``True``, this order will be reversed.
    """
    def __init__(self, categories=None, output_min=0, output_max=1,
                 fixed=True, reverse=False):
        if categories is None:
            categories = []
        if len(categories) == 0:
            # Fixed is only fixed if we provide categories on
            # initialization
            fixed = False
        self.categories = dict((cat, idx)
                               for idx, cat in enumerate(categories))
        self.ordered_categories = list(categories)
        self.orig_ordered_categories = list(self.ordered_categories)
        self.lo = output_min
        self.hi = output_max
        self.out_range = RangeLimits(output_min, output_max)
        self.reverse = reverse

        imin, imax, omin, omax = self._current_limits()

        RangeScale.__init__(self, fixed=fixed, check_input=False,
                            input_min=imin, input_max=imax,
                            output_min=omin, output_max=omax)

    def _range_from_point(self, p):
        return p, p + self.range_size()

    def _current_limits(self):
        if self.range_size() is None:
            # No points
            first_range = None
            last_range = None
            first_cat = None
            last_cat = None
        else:
            first_cat = self.ordered_categories[0]
            last_cat = self.ordered_categories[-1]
            first_point = self.out_range.low()
            last_point = self.out_range.high() - self.range_size()
            first_range = self._range_from_point(first_point)
            last_range = self._range_from_point(last_point)
        return (first_cat, last_cat, first_range, last_range)
        
    def update_(self, new_data):
        idx_offset = len(self.ordered_categories)

        new_cats = set()
        new_cats_ordered = []
        for idx, d in enumerate(new_data):
            # Check both new_cats and categories, since we don't want
            # to add either old categories or dups in the new ones.
            if d not in new_cats and d not in self.categories:
                new_cats.add((d, idx + idx_offset))
                new_cats_ordered.append(d)
        self.categories.update(new_cats)
        self.ordered_categories.extend(new_cats_ordered)
        return self._current_limits()

    def reset_(self):
        self.ordered_categories = list(self.orig_ordered_categories)
        self.categories = dict(
            (cat, idx) for idx, cat in enumerate(self.ordered_categories))
        return self._current_limits()
        
    def convert_(self, a):
        try:
            if self.reverse:
                index = len(self.categories) - 1 - self.categories[a]
            else:
                index = self.categories[a]
        except KeyError:
            raise RayonRangeException("Value '%s' not a category "
                                      "in this scale" % a)
        lo = (self.range_size() * index) + self.out_range.low()
        return self._range_from_point(lo)
            
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return (oc for oc in self.ordered_categories)

    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    def range_size(self):
        """
        Returns the size of the subranges allocated to each category,
        in terms of the output range. This value multiplied by the
        number of categories should equal the size of the output
        range.
        """
        if len(self.categories) == 0:
            return None
        else:
            return self.out_range.spread() / len(self.categories)

    # Overridden to provide documentation
    def get_input_max(self):
        """
        Returns the last category added to the scale. If a sequence of
        categories was supplied to the scale when it was created and
        no additional data has been associated with the object, the
        last category in that sequence is returned. If data has
        been associated with this scale since it was created, this
        method returns the last unique category in the most recent set
        of data currently associated with it.

        If the *reverse* parameter was passed to this scale when it
        was created, the order described above is reversed. The result
        of calling this method is then equivalent to calling
        :meth:`~CategoricalRangeScale.get_input_min` on a scale that
        has not been reversed.

        If no categories are associated with this scale, this method
        returns ``None``.
        """
        return RangeScale.get_input_max(self)

    def get_input_min(self):
        """
        Returns the first category added to the scale. If a sequence
        of categories was supplied to the scale when it was created,
        the first category in that sequence is returned. If the scale
        was created without categories and data has been associated
        with this scale, this method returns the first category in the
        oldest set of data currently associated with it.

        If the *reverse* parameter was passed to this scale when it
        was created, the order described above is reversed. The result
        of calling this method is then equivalent to calling
        :meth:`~CategoricalRangeScale.get_input_max` on a scale that
        has not been reversed.
        
        If no categories are associated with this scale, this method
        returns ``None``.
        """
        return RangeScale.get_input_min(self)

    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over all the categories currently
        associated with this scale. The *num_ticks* and *inside*
        parameters are ignored.

        If the *reverse* parameter was passed to this scale when it
        was created, this method is equivalent to calling
        ``reversed(get_nice_tick_positions)`` on a non-reversed
        version of this scale.
        """
        return RangeScale.get_nice_tick_positions(self, num_ticks, inside)
    
    def to_categorical_scale(self, alignment="beginning", fixed=True):
        """
        Returns a :class:`CategoricalScale` using the same categories
        and range bounds data as this scale.

        *alignment* is where in the range to put the reference point
        for the categories; see :meth:``CategoricalScale.__init__``.

        The resulting :class:`CategoricalScale` will be fixed (i.e.,
        categories cannot be added to it) unless *fixed* is ``False``.
        """
        self._compute_limits()
        if self.reverse:
            # Un-reverse the categories, but maintain reversed
            # orientation
            cats = tuple(reversed(self.ordered_categories))
        else:
            cats = tuple(self.ordered_categories)

        return CategoricalScale(
            categories=cats,
            input_min=self.lo,
            input_max=self.hi,
            alignment=alignment,
            fixed=fixed,
            reverse=self.reverse)


class CategoricalScale(Scale):
    """
    Maps categories to evenly-spaced points on a numeric range. These
    points are located inside non-overlapping subranges within the
    output range.

    *categories* is a sequence of input categories. The elements in
    the sequence may be of any type provided that they are unique and
    can be tested for equality. This parameter may be omitted; in that
    case, the categories will be determined by Rayon based on the data
    the scale is required to transform.

    *output_min* and *output_max* define the boundaries of a numeric
    range. The scale will map categories to points evenly distributed
    within this range. When used as a scale on a spatial axis in a
    plot or border, the default values will use all the space
    allocated for the element being visualized.

    *alignment* indicates where the output point will lie within the
    allotted subrange. Acceptable values are ``beginning``, ``center``
    or ``end``, indicating that the point should be at the beginning,
    midpoint or end of the subrange, respectively.
    
    If *fixed* is ``True``, the value of *categories* will be the
    final value for the scale; no categories may be added or
    removed. Otherwise, the category list may change if new data is
    associated with the scale.

    If *reverse* is ``False``, categories will be ordered as they were
    added to the scale. Additional categories beyond the initial set
    will be appended in the order they were inserted. If *reverse* is
    ``True``, this order will be reversed.
    """
    def __init__(self, categories=None, output_min=0, output_max=1,
                 alignment="beginning", fixed=True, reverse=False):
        if alignment not in ('beginning', 'center', 'end'):
            raise RayonException("Invalid alignment value: '%s'" % alignment)
        self.alignment = alignment

        if categories is None:
            categories = []
        if len(categories) == 0:
            # Fixed is only fixed if we provide categories on
            # initialization
            fixed = False
            imin = None
            imax = None
        else:
            imin = categories[0]
            imax = categories[-1]

        self.ranges = CategoricalRangeScale(categories,
                                            output_min, output_max,
                                            fixed, reverse)
        omin = self._point_from_range(self.ranges.get_output_min())
        omax = self._point_from_range(self.ranges.get_output_max())
            
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=imin, input_max=imax,
                       output_min=omin, output_max=omax)


    def _point_from_range(self, rangetuple):
        # If the scale is empty, the range tuple may be None
        if rangetuple is None:
            return None
        lo, hi = rangetuple
        if self.alignment == 'beginning':
            return lo
        elif self.alignment == 'center':
            return lo + ( self.ranges.range_size() / 2)
        elif self.alignment == 'end':

            return hi

    def update_(self, new_data):
        """
        Adds categories to the scale. The categories are added after
        any existing categories, in the order they appear in
        *new_data*. Elements of *new_data* that are duplicates of
        existing categories are ignored.

        If this scale was constructed with *fixed* equal to ``True``,
        this method will have no effect.
        """
        in_min, in_max, out_min, out_max = self.ranges.update_(new_data)
        return (in_min, in_max,
                self._point_from_range(out_min),
                self._point_from_range(out_max))

    def reset_(self):
        in_min, in_max, out_min, out_max = self.ranges.reset_()
        return (in_min, in_max,
                self._point_from_range(out_min),
                self._point_from_range(out_max))
    
    def convert_(self, a):
        return self._point_from_range(self.ranges.convert_(a))
    
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return self.ranges.get_nice_tick_positions_(num_ticks, inside)

    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    # Overridden to provide documentation
    def get_input_max(self):
        """
        Returns the last category added to the scale. If a sequence of
        categories was supplied to the scale when it was created and
        no additional data has been associated with the object, the
        last category in that sequence is returned. If data has
        been associated with this scale since it was created, this
        method returns the last unique category in the most recent set
        of data currently associated with it.

        If the *reverse* parameter was passed to this scale when it
        was created, the order described above is reversed. The result
        of calling this method is then equivalent to calling
        :meth:`~CategoricalScale.get_input_min` on a scale that has
        not been reversed.

        If no categories are associated with this scale, this method
        returns ``None``.
        """
        return Scale.get_input_max(self)

    def get_input_min(self):
        """
        Returns the first category added to the scale. If a sequence
        of categories was supplied to the scale when it was created,
        the first category in that sequence is returned. If the scale
        was created without categories and data has been associated
        with this scale, this method returns the first category in the
        oldest set of data currently associated with it.

        If the *reverse* parameter was passed to this scale when it
        was created, the order described above is reversed. The result
        of calling this method is then equivalent to calling
        :meth:`~CategoricalScale.get_input_max` on a scale that has not
        been reversed.
        
        If no categories are associated with this scale, this method
        returns ``None``.
        """
        return Scale.get_input_min(self)

    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over all the categories currently
        associated with this scale. The *num_ticks* and *inside*
        parameters are ignored.

        If the *reverse* parameter was passed to this scale when it
        was created, this method is equivalent to calling
        ``reversed(get_nice_tick_positions)`` on a non-reversed
        version of this scale.
        """
        return Scale.get_nice_tick_positions(self, num_ticks, inside)

    def to_categorical_range_scale(self, fixed=True):
        """
        Returns a :ry:factory:`CategoricalRangeScale` scale using the
        same categories and range bounds data as this scale.

        The resulting :ry:factory:`CategoricalRangeScale` scale will
        be fixed (i.e., categories cannot be added to it) unless
        *fixed* is ``False``.
        """
        self._compute_limits()
        new_scale = copy.deepcopy(self.ranges)
        # Prime evil! fixed is not part of the public
        # interface of CategoricalRangeScale!
        new_scale.fixed = True
        return new_scale

    
            
def _extract_lo_hi(new_data):
    # Extract the lo and high points from a set of data.
    #
    # If it's a Column, use the min function on it; the data may
    # already have been sorted. Otherwise, sort and return min and
    # max.
    if len(new_data) <= 0:
        return
    if type(new_data) == data.Column:
        lo, hi = new_data.min(), new_data.max()
    else:
        new_data = sorted(new_data)
        lo = new_data[0]
        hi = new_data[-1]
    return lo, hi
            
            
    
class LinearScale(Scale):
    """
    Maps values from one numeric range to another with linear scaling.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* specify the initial lower and upper
    bounds of the output range. 
    
    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """
    def __init__(self, input_min=None, input_max=None,
                 output_min=0, output_max=1,
                 fixed=True, start_at_zero=True,
                 a_lo=None, a_hi=None, b_lo=None, b_hi=None):
        # a_lo, a_hi, b_lo and b_hi are undocumented and should be
        # removed in Rayon 2.0. For backward compatibility, however,
        # their values supercede those of (in|out)put_(min|max)
        if a_lo is not None: input_min = a_lo
        if a_hi is not None: input_max = a_hi
        if b_lo is not None: output_min = b_lo
        if b_hi is not None: output_max = b_hi

        # TODO: start_at_zero -- should it exist? If so, document
        # it. If not, don't.
        
        # We're only a fixed scale if both boundaries have been
        # specified and fixed is set to True
        if input_min is None or input_max is None:
            fixed=False
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        self.in_range = RangeLimits(lo=input_min, hi=input_max,
                                   start_at_zero=start_at_zero)
        self.out_range = RangeLimits(lo=output_min, hi=output_max,
                                   start_at_zero=start_at_zero)
    def update_(self, new_data):
        lo, hi = _extract_lo_hi(new_data)
        self.in_range.update(lo, hi)
        return (self.in_range.low(), self.in_range.high(),
                self.out_range.low(), self.out_range.high())
    def reset_(self):
        self.in_range.reset()
        return (self.in_range.low(), self.in_range.high(),
                self.out_range.low(), self.out_range.high())
    def convert_(self, a):
        try:
            in_lo, in_hi = self.in_range.limits()
            out_lo, out_hi = self.out_range.limits()
        except RayonLimitException, e:
            raise RayonException(
                "Insufficient information to scale value: %s" % str(e))
        in_spread = self.in_range.spread()
        in_off = -in_lo
        if in_spread == 0:
            in_ratio = 0
        else:
            in_ratio = (a + in_off) / in_spread
        return (in_ratio * self.out_range.spread()) + out_lo
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return netsa.data.nice.nice_ticks(self.get_input_min(),
                                          self.get_input_max(),
                                          num_ticks, inside)[2]
    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True




class LogScale(Scale):
    """
    Maps values from one numeric range to another with log-transformed
    linear scaling.

    Valid input for this scale is any positive real number; 0 and
    negative numbers are invalid input and will raise a
    :class:`~rayon.RayonException` (a :class:`RayonLimitException` if
    invalid scale limits are specified, or a
    :class:`RayonRangeException` if the user attempts to convert an
    invalid value).

    If your data is integer data containing positive numbers and zero
    (such as most count data), consider using the
    :ry:factory:`CountingLogScale` scale, which is designed for this
    kind of data.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* specify the initial lower and upper
    bounds of the output range. 
    
    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """
    
    # Represents the result of taking the log of something unloggable
    # (or inadvisably loggable, in this case.)
    INVALID = object()
    
    @classmethod
    def lin2log(cls, x):
        if x is None: return None
        if x <= 0: return cls.INVALID
        return math.log(x, 10)
    def __init__(self, input_min=None, input_max=None,
                 output_min=0, output_max=1,
                 fixed=True, a_lo=None, a_hi=None, b_lo=None, b_hi=None):
        # a_lo, a_hi, b_lo and b_hi are undocumented and should be
        # removed in Rayon 2.0. For backward compatibility, however,
        # their values supercede those of (in|out)put_(min|max)
        if a_lo is not None: input_min = a_lo
        if a_hi is not None: input_max = a_hi
        if b_lo is not None: output_min = b_lo
        if b_hi is not None: output_max = b_hi

        # We're only a fixed scale if both boundaries have been
        # specified and fixed is set to True
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=True,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        log_input_min = self.lin2log(input_min)
        log_input_max = self.lin2log(input_max)
        # Check limits, if supplied
        if self.INVALID in (log_input_min, log_input_max):
            raise RayonLimitException("Limits of LogScale must be "
                                      "greater than 0")

        self.linscale = LinearScale(log_input_min, log_input_max,
                                    output_min, output_max, fixed)
    def _format_limits(self, limits):
        in_min, in_max, out_min, out_max = limits
        def exponentify(x):
            if x is None: return None
            return 10**x
        return exponentify(in_min), exponentify(in_max), out_min, out_max
    def update_(self, new_data):
        if len(new_data) <= 0:
            return
        raw_lo, raw_hi = _extract_lo_hi(new_data)
        # None should only be here if someone passed in bad data. But
        # that's not impossible....
        if None in (raw_lo, raw_hi):
            raise RayonLimitException("Limits of LogScale cannot be None")

        log_lo = self.lin2log(raw_lo)
        log_hi = self.lin2log(raw_hi)
        if self.INVALID in (log_lo, log_hi):
            raise RayonLimitException("Limits of LogScale must "
                                      "be greater than 0")
        return self._format_limits(self.linscale.update_([log_lo, log_hi]))
    def reset_(self):
        return self._format_limits(self.linscale.reset_())
    def convert_(self, a):
        if a is None:
            raise RayonRangeException("LogScale cannot plot None")
        log_a = self.lin2log(a)
        if log_a is self.INVALID:
            raise RayonRangeException("Invalid value for LogScale: %s" % a)
        return self.linscale.convert_(log_a)
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        if inside:
            hi = int(math.floor(math.log(self.get_input_max(), 10)))
            lo = int(math.ceil(math.log(self.get_input_min(), 10)))
        else:
            hi = int(math.ceil(math.log(self.get_input_max(), 10)))
            lo = int(math.floor(math.log(self.get_input_min(), 10)))
        if hi == lo:
            hi += 1
        return (10**i for i in xrange(lo, hi+1))

    def check_input_(self, in_data):
        for i in in_data:
            if i <= 0:
                raise RayonScaleException("Input contains data <= 0")

    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    # Overridden for documentation
    
    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over a set of tick positions evenly spaced
        on the log scale for this range of data. (E.g. 10, 100,
        1000,...)  The value of *num_ticks* is ignored; *inside* is
        treated as ``True`` regardless of its value, meaning the range
        of data associated with this scale will always fall inside the
        range emitted by this method.
        """
        return Scale.get_nice_tick_positions(self, num_ticks, inside)


class CountingLogScale(Scale):
    """
    Maps values from one numeric range to another using a modified
    log-transformed linear scale best suited to count data. 

    The input range for this scale can include zeroes (but not
    negative numbers), and is suitable for integer data. However, it
    distorts non-integer data relative to a simple log scale. If the
    data being scaled contains only positive values and no zeroes, the
    :ry:factory:`LogScale` scale is more appropriate.
    
    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* specify the initial lower and upper
    bounds of the output range. 
    
    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """
    def __init__(self, input_min=None, input_max=None,
                 output_min=0, output_max=1,
                 fixed=True, a_lo=None, a_hi=None, b_lo=None, b_hi=None):
        # a_lo, a_hi, b_lo and b_hi are undocumented and should be
        # removed in Rayon 2.0. For backward compatibility, however,
        # their values supercede those of (in|out)put_(min|max)
        if a_lo is not None: input_min = a_lo
        if a_hi is not None: input_max = a_hi
        if b_lo is not None: output_min = b_lo
        if b_hi is not None: output_max = b_hi

        # We're only a fixed scale if both boundaries have been
        # specified and fixed is set to True
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=True,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        
        if ((input_min is not None and input_min < 0)
            or (input_max is not None and input_max < 0)):
            raise RayonLimitException("Limits of CountingLogScale "
                                      "must be greater than or equal "
                                      "to zero.")
        if input_min is not None: input_min += 1
        if input_max is not None: input_max += 1
        self._logscale = LogScale(input_min, input_max,
                                  output_min, output_max, fixed)
    def _format_limits(self, limits):
        in_min, in_max, out_min, out_max = limits
        def munge(n):
            if n is None: return None
            else: return n - 1
        return munge(in_min), munge(in_max), out_min, out_max 
    def update_(self, new_data):
        raw_lo, raw_hi = _extract_lo_hi(new_data)
        try:
            return self._format_limits(self._logscale.update_(
                [raw_lo + 1, raw_hi + 1]))
        except RayonRangeException:
            # Re-raise with clearer message
            raise RayonRangeException(
                "Invalid value for counting log scale: %s" % a)
    def reset_(self):
        return self._format_limits(self._logscale.reset_())
    def convert_(self, a):
        try:
            return self._logscale.convert_(a + 1)
        except RayonRangeException:
            # Re-raise with clearer message
            raise RayonRangeException(
                "Invalid value for counting log scale: %s" % a)
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        if inside:
            hi = int(math.floor(math.log(self.get_input_max(), 10)))
            if self.get_input_min() == 0:
                lo = 0
            else:
                lo = int(math.ceil(math.log(self.get_input_min(), 10)))
        else:
            hi = int(math.ceil(math.log(self.get_input_max(), 10)))
            if self.get_input_min() == 0:
                lo = 0
            else:
                lo = int(math.floor(math.log(self.get_input_min(), 10)))
        if hi == lo:
            hi += 1
        if lo == 0:
            yield 0
        for i in xrange(lo, hi+1):
            yield 10**i

    def check_input_(self, in_data):
        for i in in_data:
            if in_data < 0:
                raise RayonScaleException("Input contains data < 0")

    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    # Overridden for documentation

    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over a set of tick positions evenly spaced
        on the log scale for this range of data. (E.g. 10, 100,
        1000,...)  The value of *num_ticks* is ignored; *inside* is
        treated as ``True`` regardless of its value, meaning the range
        of data associated with this scale will always fall inside the
        range emitted by this method.
        """
        return Scale.get_nice_tick_positions(self, num_ticks, inside)

class TimeScale(Scale):
    """
    Maps values from a time range linearly to values in a numeric
    range.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range, as :class:`datetime.datetime`
    objects. If one or both of these parameters is not supplied, Rayon
    will select a value that will minimally contain the data
    associated with the scale.

    *output_min* and *output_max* specify the initial lower and upper
    bounds of the output range.

    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """
    # Implementation: basically wraps around LinearScale, converting
    # inputs and outputs between dates and numbers.
    def __init__(self, input_min=None, input_max=None,
                 output_min=0, output_max=1,
                 fixed=True, begin=None, end=None, lo=None, hi=None):
        # begin, end, lo and hi are undocumented and should be removed
        # in Rayon 2.0. For backward compatibility, however, their
        # values supercede those of (in|out)put_(min|max)
        if begin is not None: input_min = begin
        if end   is not None: input_max = end
        if lo    is not None: output_min = lo
        if hi    is not None: output_max = hi

        # Yet again - we're only a fixed scale if both boundaries have
        # been specified and fixed is set to True
        #
        # This is starting to look like it should be in the
        # superclass, but I disagree -- other scales might not have such
        # clear limits, but may still be fixed/unfixed. (They might
        # have internal thresholds for low, medium and high that the
        # data can adjust, for instance.)
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        if input_min is not None:
            input_min = calendar.timegm(input_min.timetuple())
        if input_max is not None:
            input_max = calendar.timegm(input_max.timetuple())
        self._linscale = LinearScale(input_min, input_max,
                                     output_min, output_max, fixed)
    def convert_(self, a):
        """ Convert datetime to number in output range.
                             
            @type dt: datetime
            @param dt: date within this scale's time range

            @return b position in output scale """
        return self._linscale.convert_(calendar.timegm(a.timetuple()))
    def _format_limits(self, limits):
        in_min, in_max, out_min, out_max = limits
        def datetimeify(d):
            if d is None: return None
            else: return netsa.data.times.make_datetime(d)
        return (datetimeify(in_min), datetimeify(in_max),
                out_min, out_max)
    def update_(self, new_data):
        # Extract lo and hi from new_data, so we only convert them
        # from datetimes.
        if type(new_data) == data.Column:
            new_new_data = [new_data.min(), new_data.max()]
        else:
            if len(new_data) == 0:
                new_new_data = []
            else:
                new_data = sorted(new_data)
                new_new_data = [new_data[0], new_data[-1]]
        return self._format_limits(self._linscale.update_(
            [calendar.timegm(d.timetuple()) for d in new_new_data]))
    def reset_(self):
        return self._format_limits(self._linscale.reset_())
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return netsa.data.nice.nice_time_ticks(self.get_input_min(),
                                               self.get_input_max(),
                                               num_ticks, inside)[2]
    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True




class LinearGradientScale(Scale):
    """
    Maps a numeric range linearly to a range of colors along a
    gradient.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* are color specifications (see
    :ref:`layout-color-specs`) representing the beginning and end
    points of a color gradient. *numsteps* is the number of colors
    in the gradient, including the beginning and end points.

    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """

    def __init__(self, input_min=None, input_max=None,
                 output_min="#ffffff", output_max="#000000", numsteps=5,
                 fixed=True, num_lo=None, num_hi=None,
                 color_lo=None, color_hi=None):
        # num_lo and num_hi are undocumented and should be removed in
        # Rayon 2.0. For backward compatibility, however, use them if
        # they are provided
        if num_lo is not None: input_min = num_lo
        if num_hi is not None: input_max = num_hi
        # Ditto color_lo and color_hi
        if color_lo is not None: output_min = color_lo
        if color_hi is not None: output_max = color_hi
        
        
        # Yet again - we're only a fixed scale if both boundaries have
        # been specified and fixed is set to True
        #
        # This is starting to look like it should be in the
        # superclass, but I disagree -- other scales might not have such
        # clear limits, but may still be fixed/unfixed. (They might
        # have internal thresholds for low, medium and high that the
        # data can adjust, for instance.)
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=color_hi)
        
        self.in_range = RangeLimits(input_min, input_max, start_at_zero=False)
        self.color_lo = miscutils.parse_color(output_min)
        self.color_hi = miscutils.parse_color(output_max)
        self.color_spread = numsteps
        self.gradient = miscutils.gradient_colors(
            self.color_lo, self.color_hi, self.color_spread)
                                   
    def update_(self, new_data):
        self.in_range.update(new_data)
        return (self.in_range.low(), self.in_range.high(),
                self.color_lo, self.color_hi)
    def reset_(self):
        self.in_range.reset()
        return (self.in_range.low(), self.in_range.high(),
                self.color_lo, self.color_hi)
                
    def convert_(self, a):
        if a not in self.in_range:
            raise RayonException(
                "%s: Input value '%s' out of range" % (type(self), a))

        if self.in_range.spread() == 0:
            a_ratio = 0
        else:
            offset = self.in_range.low() * -1
            a_ratio = (a + offset) / self.in_range.spread()
        scaled_a = int(a_ratio * (self.color_spread - 1))
        return self.gradient[scaled_a]

    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return netsa.data.nice.nice_ticks(self.get_input_min(),
                                               self.get_input_max(),
                                               num_ticks, inside)[2]
    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True


    
class LogGradientScale(Scale):
    """
    Maps a numeric range to a range of colors along a gradient, using
    a log-transformed linear scale.

    This scale's relationship to the :ry:factory:`LinearGradientScale`
    scale is analogous to the :ry:factory:`LogScale` scale's
    relationship to the :ry:factory:`LinearScale` scale.  As with the
    :ry:factory:`LogScale` scale, valid input for this scale is any
    positive real number; zero and negative numbers are invalid input
    and will raise a :class:`~rayon.RayonException` (a
    :class:`RayonLimitException` if invalid scale limits are
    specified, or a :class:`RayonRangeException` if the user attempts
    to convert an invalid value).

    If your data is integer data containing positive numbers and zero
    (such as most count data), consider using the
    :ry:factory:`CountingLogGradientScale` scale, which is designed
    for this kind of data.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* are color specifications (see
    :ref:`layout-color-specs`) representing the beginning and end
    points of a color gradient. *numsteps* is the number of colors
    in the gradient, including the beginning and end points.
    
    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """

    def __init__(self, input_min=None, input_max=None,
                 output_min="#ffffff", output_max="#000000",
                 numsteps=5, fixed=True, num_lo=None, num_hi=None,
                 color_lo=None, color_hi=None):
        # num_lo and num_hi are undocumented and should be removed in
        # Rayon 2.0. For backward compatibility, however, use them if
        # they are provided
        if num_lo is not None: input_min = num_lo
        if num_hi is not None: input_max = num_hi
        # ditto color_lo and color_hi
        if color_lo is not None: output_min = color_lo
        if color_hi is not None: output_max = color_hi

        # Yet again - we're only a fixed scale if both boundaries have
        # been specified and fixed is set to True
        #
        # This is starting to look like it should be in the
        # superclass, but I disagree -- other scales might not have such
        # clear limits, but may still be fixed/unfixed. (They might
        # have internal thresholds for low, medium and high that the
        # data can adjust, for instance.)
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        
        self._logscale = LogScale(input_min, input_max, fixed=fixed)
        self._gradscale = LinearGradientScale(0, 1, output_min, output_max,
                                              numsteps)
    def update_(self, new_data):
        in_min, in_max, _, _ = self._logscale.update_(new_data)
        return (in_min,
                in_max,
                self._gradscale.get_output_min(),
                self._gradscale.get_output_max())
    def reset_(self):
        in_min, in_max, _, _ = self._logscale.reset_()
        return (in_min,
                in_max,
                self._gradscale.get_output_min(),
                self._gradscale.get_output_max())
    def convert_(self, a):
        log_scaled_a = self._logscale.convert_(a)
        return self._gradscale.convert_(log_scaled_a)
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return self._logscale.get_nice_tick_positions_(num_ticks, inside)
    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    # Overridden for documentation
    
    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over a set of tick positions evenly spaced
        on the log scale for this range of data. (E.g. 10, 100,
        1000,...)  The value of *num_ticks* is ignored; *inside* is
        treated as ``True`` regardless of its value, meaning the range
        of data associated with this scale will always fall inside the
        range emitted by this method.
        """
        return Scale.get_nice_tick_positions(self, num_ticks, inside)

    

class CountingLogGradientScale(Scale):
    """
    Maps a numeric range to a range of colors along a gradient, using
    a modified log-transformed linear scale suitable for count data.

    This scale is analogous to the :ry:factory:`CountingLogScale`
    scale. Its input range can include zeroes (but not negative
    numbers), and is suitable for integer data. However, it distorts
    non-integer data relative to a simple log scale.  If the data
    being scaled contains only positive values and no zeroes, the
    :ry:factory:`LogGradientScale` scale is more appropriate.

    *input_min* and *input_max* specify the initial lower and upper
    bounds of the input range. If one or both of these parameters is
    not supplied, Rayon will select a value that will minimally
    contain the data associated with the scale.

    *output_min* and *output_max* are color specifications (see
    :ref:`layout-color-specs`) representing the beginning and end
    points of a color gradient. *numsteps* is the number of colors
    in the gradient, including the beginning and end points.
    
    If *fixed* is ``True``, the values of *input_min* and *input_max*
    will be the final values for the scale; otherwise, Rayon may
    adjust the input range if new data is associated with the scale.
    """
    def __init__(self, input_min=None, input_max=None,
                 output_min=(255, 255, 255), output_max=(0,0,0),
                 numsteps=5, fixed=True, num_lo=None, num_hi=None,
                 color_lo=None, color_hi=None):
        # num_lo and num_hi are undocumented and should be removed in
        # Rayon 2.0. For backward compatibility, however, use them if
        # they are provided
        if num_lo is not None: input_min = num_lo
        if num_hi is not None: input_max = num_hi
        # ditto color_lo and color_hi
        if color_lo is not None: output_min = color_lo
        if color_hi is not None: output_max = color_hi

        # Yet again - we're only a fixed scale if both boundaries have
        # been specified and fixed is set to True
        #
        # This is starting to look like it should be in the
        # superclass, but I disagree -- other scales might not have such
        # clear limits, but may still be fixed/unfixed. (They might
        # have internal thresholds for low, medium and high that the
        # data can adjust, for instance.)
        if input_min is None or input_max is None:
            fixed = False
        Scale.__init__(self, fixed=fixed, check_input=False,
                       input_min=input_min, input_max=input_max,
                       output_min=output_min, output_max=output_max)
        self._logscale = CountingLogScale(input_min, input_max, fixed=fixed)
        self._gradscale = LinearGradientScale(0, 1, output_min, output_max,
                                              numsteps)
    def update_(self, new_data):
        in_min, in_max, _, _ = self._logscale.update_(new_data)
        return (in_min, in_max,
                self._gradscale.get_output_min(),
                self._gradscale.get_output_max())
    def reset_(self):
        in_min, in_max, _, _ = self._logscale.reset_(new_data)
        return (in_min, in_max,
                self._gradscale.get_output_min(),
                self._gradscale.get_output_max())
    def convert_(self, a):
        log_scaled_a = self._logscale.convert_(a)
        return self._gradscale.convert_(log_scaled_a)
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return self._logscale.get_nice_tick_positions_(num_ticks, inside)
    def has_input_upper_bound_(self): return True
    def has_input_lower_bound_(self): return True

    # Overridden for documentation
    
    def get_nice_tick_positions(self, num_ticks=5, inside=False):
        """
        Returns an iterator over a set of tick positions evenly spaced
        on the log scale for this range of data. (E.g. 10, 100,
        1000,...)  The value of *num_ticks* is ignored; *inside* is
        treated as ``True`` regardless of its value, meaning the range
        of data associated with this scale will always fall inside the
        range emitted by this method.
        """
        return Scale.get_nice_tick_positions(self, num_ticks, inside)




class DumbScale(Scale):
    "Identity scale; returns whatever it was passed with no scaling."
    def __init__(self): Scale.__init__(self, fixed=True, check_input=False)
    def has_input_upper_bound_(self): return False
    def has_input_lower_bound_(self): return False
    def has_output_upper_bound_(self): return False
    def has_output_lower_bound_(self): return False
    def update_(self, new_data): return None, None, None, None
    def convert_(self, a): return a
    def get_nice_tick_positions_(self, num_ticks=5, inside=False):
        return []
        


# Factory function

_scale_types = {
    "linear"      : LinearScale,
    "log"         : LogScale,
    "clog"        : CountingLogScale,
    "time"        : TimeScale,
    "categorical" : CategoricalScale,
    "catrange"    : CategoricalRangeScale,
    "gradient"    : LinearGradientScale,
    "loggrad"     : LogGradientScale,
    "cloggrad"    : CountingLogGradientScale,
}

def create(name, *args, **kargs):
    return _scale_types[name](*args, **kargs)

def list_scale_types():
    return tuple((k, v) for k, v in _scale_types.items())
