#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
This module concerns itself with the generation of ticks. Ticks are
points of interest in a data series. The presentation of ticks to
users is handled by classes in the L{markers} and L{tickmarkers}
modules.
"""

# TODO: Finish deprecating Ticks and RangeTicks names. New names are
# Tickset and RangeTickset.

import base
import miscutils
import markers
import data
from tickspec_parse import parse_tickspec

import datetime

from rayon import RayonException


# python 2.4 min() and max() don't take a key parameter, so we have to
# wrap them in compatibility functions
try:
    min([(0, 1), (1, 0)], key=lambda x: x[1])
    def min_with_key(seq, key):
        return min(seq, key=key)
    def max_with_key(seq, key):
        return max(seq, key=key)
except TypeError:
    def min_with_key(seq, key):
        to_search = [(key(i), i) for i in seq]
        return min(to_search)[1]
    def max_with_key(seq, key):
        to_search = [(key(i), i) for i in seq]
        return max(to_search)[1]



class TickSetBase(base.Stamp):
    # Abstract base class.
    def __init__(self, tick_tuples, scale, marker):
        if type(self) == TickSetBase:
            raise RayonException("TickSetBase is not instantiable")

        # Create limit data and attach it to the scale
        if len(tick_tuples) == 0:
            tickset_limits = (0, 0)
        else:
            tickset_limits = (min_with_key(tick_tuples, key=lambda x: x[0])[0],
                              max_with_key(tick_tuples, key=lambda x: x[0])[0])
        if hasattr(scale, "attach"):
            scale.attach(tickset_limits, self)
        
        self.tick_tuples = tick_tuples
        self.scale = scale
        self.marker = marker


    def get_scale(self):
        return self.scale

        
class TickSet(TickSetBase):

    def __iter__(self):
        """
        Returns an iterator of pairs over the positions and labels of
        ticks in this tick set. The first element of the pair is the
        tick position (as returned from :meth:`iter_tick_positions`)
        and the second is the label associated with that position (as
        returned from :meth:`iter_tick_labels`).
        """
        return miscutils.tick_data_iter(self.tick_tuples)
    def iter_tick_positions(self):
        """
        Returns an iterator over the positions of ticks in this tick
        set. Positions will be returned in the same order as the
        labels in the :meth:`iter_tick_labels` method.
        """
        return (_[0] for _ in self)
    def iter_tick_labels(self):
        """
        Returns an iterator over the tick labels in this tick
        set. Labels will be returned in the same order as the
        positions in the :meth:`iter_tick_positions` method.
        """
        return (_[1] for _ in self)

    def get_scale(self):
        """
        Returns the scale associated with this tick set.
        """
        return self.scale


    def draw_vertically(self, ctx, width, height, x=None):
        if x is None:
            x = width
        for tick, text in miscutils.tick_data_iter(self.tick_tuples):
            self.marker.draw(ctx, x, (1 - self.scale(tick)) * height,
                             width, height, text)

    def draw_horizontally(self, ctx, width, height, y=None):
        if y is None:
            y = 0
        for tick, text in miscutils.tick_data_iter(self.tick_tuples):
            self.marker.draw(ctx, self.scale(tick) * width, y,
                             width, height, text)


Ticks = TickSet

class RangeTickSet(TickSetBase):
    # Documented in classref-ticks.rst
    alignment_map = {
        "left"   : 0.0,
        "bottom" : 0.0,
        "center" : 0.5,
        "top"    : 1.0,
        "right"  : 1.0
    }
    def __init__(self, tick_tuples, scale, marker, align=.5):
        """
        Constructor

        @type tick_tuples: sequence of 2-tuples
        @param tick_tuples: Identical to I{tick_tuples} in L{Ticks}.

        @type scale: callable
        @param scale: scaling function or L{scales.Scale} object. See
            scales.py for interface. The scale is expected to return a
            range in the destination space. If the scale
            returns a single value instead, just use L{Ticks}.

        @type marker: markers.Marker
        @param marker: Marker object associated with this set of ticks.

        @type  align: float between [0, 1], or one of string literals
            C{left}, C{bottom}, C{center}, C{right}, or C{top}.
        @param align: How far to the left within the range to locate
            the tick, proportional to the range output by the
            scale. C{0} aligns the tick to the beginning (left,
            bottom) of the range; C{1} aligns the tick to the end
            (right, top) of the range; C{.5} aligns the tick to the
            center of the range. (C{left}/C{bottom} and
            C{right}/C{top} are synonyms; you can use a C{RangeTicks}
            specified with C{left} or C{right} to draw ticks on a
            vertical border, for example.)
        """
        TickSetBase.__init__(self, tick_tuples, scale, marker)
        if type(align) == str:
            self.align = self.alignment_map[align]
        else:
            self.align = align

    # Iterators
    def __iter__(self):
        """
        Returns an iterator of pairs over the positions and labels of
        ticks in this tick set. The first element of the pair is the
        tick position (as returned from :meth:`iter_tick_positions`)
        and the second is the label associated with that position (as
        returned from :meth:`iter_tick_labels`).
        """
        return miscutils.tick_data_iter(self.tick_tuples)
    def iter_tick_positions(self) :
        """
        Returns an iterator over the positions of ticks in this tick
        set. Positions will be returned in the same order as the
        labels in the :meth:`iter_tick_labels` method.
        """
        return (_[0] for _ in self)
    def iter_tick_labels(self)    :
        """
        Returns an iterator over the tick labels in this tick
        set. Labels will be returned in the same order as the
        positions in the :meth:`iter_tick_positions` method.
        """
        return (_[1] for _ in self)


    def draw_vertically(self, ctx, width, height, x=None):
        if x is None:
            x = width
        for tick, text in miscutils.tick_data_iter(self.tick_tuples):
            lo, hi = self.scale(tick)
            loc = lo + (self.scale.range_size() * self.align)
            self.marker.draw(ctx, x, (1 - loc) * height,
                             width, height, text)

    def draw_horizontally(self, ctx, width, height, y=None):
        if y is None:
            y = 0
        for tick, text in miscutils.tick_data_iter(self.tick_tuples):
            lo, hi = self.scale(tick)
            loc = lo + (self.scale.range_size() * self.align)
            self.marker.draw(ctx, loc * width, y,
                             width, height, text)

RangeTicks = RangeTickSet

def tuples_from_spec(col, spec, formatter):
    """
    Convert a string specification of where to put tickmarkers into
    series of ticks appropriate to the specification given data.

    @type  col: L{data.Column}
    @param col: input data from which to derive the ticks

    @type  spec: string
    @param spec: comma-separated list of tick specifications. Valid
        values are C{min}, C{median}, and C{max}, which put ticks at
        the minimum, median and maximum values in the data,
        respectively. Another valid value is C{pN}, where I{N} is a
        value from 1 to 99 representing a percentile (e.g., C{median}
        is equivalent to C{p50}). Finally, the value C{nN} will put a
        tickmark at the numeric position I{N} (e.g., C{n50} puts a
        tickmark at the value 50).

    @type  formatter: string, or function str -> str
    @param formatter: Specification for how to format the data in
        I{col} to make ticks. either a function, or a printf-style
        format string .
    """
    def donice(num_ticks):
        if isinstance(col[0], datetime.datetime):
            lo, hi, nice_ticks = miscutils.nice_time_ticks(
                col.min(), col.max(), num_ticks, inside=True)
        else:
            lo, hi, nice_ticks = miscutils.nice_ticks(
                col.min(), col.max(), num_ticks, inside=True)
        return list(nice_ticks)
    toks = tuple(s.strip() for s in spec.split(","))
    tickset = []
    for t in toks:
        if t.startswith("p"):
            tickset.append(col.percentile(int(t[1:])))
        elif t.startswith("n"):
            tickset.extend(donice(int(t[1:])))
        elif t.lower() == "min":
            tickset.append(col.min())
        elif t.lower() == "median":
            tickset.append(col.percentile(50))
        elif t.lower() == "max":
            tickset.append(col.max())
        else:
            raise ValueError("Unexpected tick specifier: '%s'" % t)
    def formatted_value(rawval):
        if isinstance(formatter, str):
            return formatter % rawval
        elif callable(formatter):
            return formatter(rawval)
        else:
            return None
    rc = [(tick, formatted_value(tick)) for tick in tickset]
    return rc


_tickset_types = {
    "default": TickSet,
    "ranged" : RangeTickSet
}

def create(tickset_type, data=None, spec=None, formatter=None, **kargs):
    # Create a tickset from tuples, or from an old-style (deprecated) spec
    if data is not None or spec is not None or formatter is not None:
        if None in (data, spec, formatter):
            raise RayonException("get_tickset needs none or all of "
                                 "data, spec and formatter")
        kargs['tick_tuples'] = ticks.tuples_from_spec(data, spec, formatter)
    return _tickset_types[tickset_type](**kargs)


def create_from_tuples(tickset_type, tick_tuples, scale, marker, alignment=.5):
    if tickset_type == "ranged":
        return RangeTickSet(tick_tuples, scale, marker, alignment)
    else:
        return TickSet(tick_tuples, scale, marker)

def create_from_new_spec(tickset_type, col, spec, formatter, scale, marker, alignment=.5):
    # Create a tickset from a new-style spec.
    col = data.Column(col)
    def formatted_value(rawval):
        if isinstance(formatter, str):
            return formatter % rawval
        elif callable(formatter):
            return formatter(rawval)
        elif formatter is None:
            return "%s" % rawval
        else:
            return None
    def flatten(seq):
        return reduce(lambda x,y: x + list(y), seq, [])
    functions = parse_tickspec("", spec)
    tick_vals = flatten([f([col], scale) for f in functions])
    tick_tuples = tuple((val, formatted_value(val)) for val in tick_vals)

    if tickset_type == "ranged":
        return RangeTickSet(tick_tuples, scale, marker, alignment)
    else:
        return TickSet(tick_tuples, scale, marker)

    

def list_tickset_types():
    return tuple((k, v) for k, v in _tickset_types.items())




__all__ = ["create",
           "Ticks", "TickSet",
           "RangeTicks", "RangeTickSet",
           "tuples_from_spec"]
