#  Copyright (C) 2011 by Carnegie Mellon University.
#  
#  @OPENSOURCE_HEADER_START@
#  Use of the Network Situational Awareness Portal and related source
#  code is subject to the terms of the following licenses:
#  
#  GNU Public License (GPL) Rights pursuant to Version 2, June 1991
#  Government Purpose License Rights (GPLR) pursuant to DFARS 252.225-7013
#  
#  NO WARRANTY
#  
#  ANY INFORMATION, MATERIALS, SERVICES, INTELLECTUAL PROPERTY OR OTHER 
#  PROPERTY OR RIGHTS GRANTED OR PROVIDED BY CARNEGIE MELLON UNIVERSITY 
#  PURSUANT TO THIS LICENSE (HEREINAFTER THE "DELIVERABLES") ARE ON AN 
#  "AS-IS" BASIS. CARNEGIE MELLON UNIVERSITY MAKES NO WARRANTIES OF ANY 
#  KIND, EITHER EXPRESS OR IMPLIED AS TO ANY MATTER INCLUDING, BUT NOT 
#  LIMITED TO, WARRANTY OF FITNESS FOR A PARTICULAR PURPOSE, 
#  MERCHANTABILITY, INFORMATIONAL CONTENT, NONINFRINGEMENT, OR ERROR-FREE 
#  OPERATION. CARNEGIE MELLON UNIVERSITY SHALL NOT BE LIABLE FOR INDIRECT, 
#  SPECIAL OR CONSEQUENTIAL DAMAGES, SUCH AS LOSS OF PROFITS OR INABILITY 
#  TO USE SAID INTELLECTUAL PROPERTY, UNDER THIS LICENSE, REGARDLESS OF 
#  WHETHER SUCH PARTY WAS AWARE OF THE POSSIBILITY OF SUCH DAMAGES. 
#  LICENSEE AGREES THAT IT WILL NOT MAKE ANY WARRANTY ON BEHALF OF 
#  CARNEGIE MELLON UNIVERSITY, EXPRESS OR IMPLIED, TO ANY PERSON 
#  CONCERNING THE APPLICATION OF OR THE RESULTS TO BE OBTAINED WITH THE 
#  DELIVERABLES UNDER THIS LICENSE.
#  
#  Licensee hereby agrees to defend, indemnify, and hold harmless Carnegie 
#  Mellon University, its trustees, officers, employees, and agents from 
#  all claims or demands made against them (and any related losses, 
#  expenses, or attorney's fees) arising out of, or relating to Licensee's 
#  and/or its sub licensees' negligent use or willful misuse of or 
#  negligent conduct or willful misconduct regarding the Software, 
#  facilities, or other rights or assistance granted by Carnegie Mellon 
#  University under this License, including, but not limited to, any 
#  claims of product liability, personal injury, death, damage to 
#  property, or violation of any laws or regulations.
#  
#  Carnegie Mellon University Software Engineering Institute authored 
#  documents are sponsored by the U.S. Department of Defense under 
#  Contract FA8721-05-C-0003. Carnegie Mellon University retains 
#  copyrights in all material produced under this contract. The U.S. 
#  Government retains a non-exclusive, royalty-free license to publish or 
#  reproduce these documents, or allow others to do so, for U.S. 
#  Government purposes only pursuant to the copyright license under the 
#  contract clause at 252.227.7013.
#  @OPENSOURCE_HEADER_END@
"""
Code for parsing a string tick specification into tick tuples.
"""
import re
from netsa.data.times import make_datetime
import datetime

from rayon import RayonException

# Rebooting the tickspec parsing

# A utility function used below
def boolean_from_string(param_name, s):
    if s.lower() in ("1", "y", "yes", "true"):
        return True
    elif s.lower() in ("0", "n", "no", "false"):
        return False
    else:
        raise RayonException(param_name + ": invalid value: " + s)



# Regex that matches "function calls" in the tickspec
#
# TICKFN := IDENTIFIER "(" PARAMLIST ")"
#
# IDENTIFIER := <A letter or underscore followed by
#                1+ alphanumerics or underscores>
#
# PARAMLIST := <A comma-separated list of positional parameters>
tickfn_re = re.compile(r'''([a-zA-Z_][a-zA-Z0-9_]*)(?:\((.*?)\))?(?:,|$)''')


# Each function <funcname> in the tickspec should map to one of the
# functions below named ts_<funcname> (ts == "tickspec", BTW)

def ts_percentile(param_name, args):
    try:
        num = int(args[0])
    except TypeError:
        raise RayonException(param_name + ": invalid value: " + args[0])
    except IndexError:
        raise RayonException(param_name + ": argument required")
    def p(cols, scale=None):
        return [cols[0].percentile(num)]
    return p

def ts_min(param_name, args):
    def min(cols, scale=None):
        return [cols[0].min()]
    return min

def ts_max(param_name, args):
    def max(cols, scale=None):
        return [cols[0].max()]
    return max

def ts_median(param_name, args):
    med = ts_percentile(param_name, ["50"])
    med.__name__ = "med"
    return med

def ts_scalemin(param_name, args):
    def smin(cols, scale):
        return [scale.get_input_min()]
    return smin

def ts_scalemax(param_name, args):
    def smax(cols, scale):
        return [scale.get_input_max()]
    return smax


def ts_nice(param_name, args):
    try:
        num_ticks = int(args[0])
    except TypeError:
        raise RayonException(param_name + ": invalid value: " + args[0])
    except IndexError:
        raise RayonException(param_name + ": argument required")

    if len(args) > 1:
        inside = boolean_from_string(param_name, args[1])
    else:
        inside = None
        
    def n(cols, scale):
        if inside is None:
            return scale.get_nice_tick_positions(num_ticks=num_ticks)
        else:
            return scale.get_nice_tick_positions(num_ticks=num_ticks,
                                                 inside=inside)
    return n

def get_rhs(param_name, args):
    try:
        return args[0]
    except IndexError:
        raise RayonException(param_name + ": argument required")

type_constructors = { datetime.datetime : make_datetime }

def coerce_type(param_name, col, rhs):
    try:
        type_class = type(col[0])
    except IndexError:
        return None

    # Look for converter in our lookup table. Failing that, try the
    # default constructor (i.e. the type itself).
    if type_class in type_constructors:
        type_constructor = type_constructors[type_class]
    else:
        type_constructor = type_class

    try:
        return type_constructor(rhs)
    except ValueError:
        raise RayonException("%s: can't coerce value to '%s' to %s" %
                             (param_name, rhs, type_class.__name__))


def ts_lessthan(param_name, args):
    untyped_rhs = get_rhs(param_name, args)
    def lt(cols, scale=None):
        rhs = coerce_type(param_name, cols[0], untyped_rhs)
        return (c for c in cols[0] if c < rhs)
    return lt
        
def ts_lessthan_equal(param_name, args):
    untyped_rhs = get_rhs(param_name, args)
    def le(cols, scale=None):
        rhs = coerce_type(param_name, cols[0], untyped_rhs)
        return (c for c in cols[0] if c <= rhs)
    return le
        
def ts_equal(param_name, args):
    untyped_rhs = get_rhs(param_name, args)
    def eq(cols, scale=None):
        rhs = coerce_type(param_name, cols[0], untyped_rhs)
        return (c for c in cols[0] if c == rhs)
    return eq

def ts_greaterthan_equal(param_name, args):
    untyped_rhs = get_rhs(param_name, args)
    def ge(cols, scale=None):
        rhs = coerce_type(param_name, cols[0], untyped_rhs)
        return (c for c in cols[0] if c >= rhs)
    return ge
        
def ts_greaterthan(param_name, args):
    untyped_rhs = get_rhs(param_name, args)
    def gt(cols, scale=None):
        rhs = coerce_type(param_name, cols[0], untyped_rhs)
        return (c for c in cols[0] if c > rhs)
    return gt

def ts_at(param_name, args):
    # Return ticks "at" the exact places specified in the args.
    def at(cols, scale=None):
        return  [coerce_type(param_name, cols[0], i)
                 for i in args]
    return at

def ts_none(param_name, args):
    def none(cols, scale=None):
        return []
    return none

function_map = {
    'at'   : ts_at,
    'ge'   : ts_greaterthan_equal,
    'gt'   : ts_greaterthan,
    'le'   : ts_lessthan_equal,
    'lt'   : ts_lessthan,
    'max'  : ts_max,
    'med'  : ts_median,
    'min'  : ts_min,
    'n'    : ts_nice,
    'none' : ts_none,
    'p'    : ts_percentile,
    'smax' : ts_scalemax,
    'smin' : ts_scalemin,
}

# For backward compatibility with the old way of doing things, support
# "n1".."n100" as tick specifiers
def mk_nice_ts(num):
    def ts_nice_n(param_name, args):
        return ts_nice(param_name, [str(num)])
    return ts_nice_n

        
for i in xrange(1,101):
    function_map["n" + str(i)] = mk_nice_ts(i)




def function_from_match(param_name, m, allowed, forbidden):
    fn_name, fn_args = m.groups()
    if fn_args is None:
        fn_args = []
    else:
        fn_args = fn_args.split(",")

    # Whitelist takes precedence over blacklist. Note that literals
    # can't be specified if a whitelist is specified.
    if allowed:
        if fn_name not in allowed:
            raise RayonException(param_name +
                                 ": Invalid tick specifier: "  +
                                 fn_name)
    elif forbidden:
        if fn_name in forbidden:
            raise RayonException(param_name +
                                 ": Invalid tick specifier: "  +
                                 fn_name)

    try:
        fn = function_map[fn_name]
    except KeyError:
        raise RayonException(param_name +
                             ": Unknown tick specifier: " +
                             fn_name)


    return fn(param_name, fn_args)
        

def parse_tickspec(param_name, value, allowed=None, forbidden=None):
    # Given a tick specification of the type:
    #
    #   p(50),le(25),...
    #
    # ...return a list of functions. Each function f has the signature:
    #
    #   f(col, scale=None) -> num
    #
    # ...where col is a Column object, scale is a scale, and num is a
    # number.
    #
    # param_name is a label applied to errors. For command-line tools,
    # it should be the parameter whose argument is the tick
    # specification.
    #
    # value is the tick specification
    #
    # allowed is a whitelist of allowable tick specifiers. If not
    # supplied, all tick specifiers are valid. If a whitelist is
    # specified, literal values are not allowed.
    #
    # forbidden is a blacklist of disallowed tick specifiers. If
    # whitelist is specified, blacklist will be ignored.

    # Convert param name to command-line parameter
    param_name = param_name.replace('_', '-')
    
    functions = []
    pos = 0
    end = len(value)
    while pos != end:
        m = tickfn_re.search(value, pos)
        if m is None:
            break
        functions.append(function_from_match(param_name, m, allowed, forbidden))
        pos = m.end()
    if pos != end:
        # Extra stuff at the end
        msg = [
            "Invalid tick spec:",
            "    %s" % value,
            (" " * (pos - 1) + "^"),
            "Unrecognized additional chracter(s)",
        ]
        raise TypeError("\n".join(msg))
    return functions

__all__ = ["parse_tickspec"]
